//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//**
//**  HARTDEVICES.PAS
//**
//**  HART Slave and HART Master communication components with serial and USB port
//**
//**  Autor : Georg  Hahn
//**  Datum :  13.11.96 / 15.08.05 / 12.04.06 / 06.09.06 / 24.10.08 / 05.03.10
//**
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
unit HartDevices;
{$M+}
{$UNSAFECODE ON}

{$DEFINE WithoutThread}

interface

uses
  Windows, Classes, Gauges, UsefulObjects, SerialComDevices, DynLoadFTD2XX,
  SerialFTDevices;

type

  PHartDeviceType = ^THartDeviceType;
  THartDeviceType = (dtSerialCom, dtFTDevice);

  THartCommand = record
    CommandNumber: byte;
    Data: string;
    end;

  THartAnswer = record
    CommandNumber,
    CommStatus,   // Communication or Command Summary Information
    DeviceStatus: byte;
    Data: string;
    end;

  TExecutionError = (eeNotOpened,eeTimeOut);
  TTransferError = (teParity,teHartFormat,teDifferentCommand);
  TDeviceError = (deWrongData,deDeviceException,deDeviceProblems,deDeviceFailure);

  TExecutionErrors = set of TExecutionError;
  TTransferErrors = set of TTransferError;
  TDeviceErrors = set of TDeviceError;

  PHartCommErrors = ^THartCommErrors;
  THartCommErrors = record
    ExecutionErrors: TExecutionErrors;
    TransferErrors: TTransferErrors;
    DeviceErrors: TDeviceErrors;
    end;

  THartCommParameters = record
    ReadTimeOut,
    WriteTimeOut,
    CharTransmitInterval,
    DelayParityByte: word;
    end;

  TOnReadAnswer = procedure (HartAnswer: THartAnswer; Errors: THartCommErrors) of object;
  TOnReadAnswerError = procedure (LastCommandNumber, ActualCommandNumber: byte; {HartAnswer: THartAnswer;} Errors: THartCommErrors) of object;
  TOnAnswerSimulationRequest = procedure (HartCommand: THartCommand; var HartAnswer: THartAnswer) of object;

  TOnReadCommand = procedure (HartCommand: THartCommand) of object;
  TOnReadCommandError = procedure (HartCommand: THartCommand; Errors: THartCommErrors) of object;
  TOnCommandSimulationRequest = procedure (var HartCommand: THartCommand) of object;

{$IFDEF WithThread}
  PHartReceiver = ^THartReceiver;
  THartReceiver = class(TThread)
  private
    { Private declarations }
    pHartDeviceType: PHartDeviceType;
    pComHandle: PHandle;
    pFTDHandle: PFT_HANDLE;
    pSimulationMode: pboolean;

    pReadTimeOut,
    pMaxStringLength: pword; // Attention! this was pdword !!!!
    pReceptionString,
    pSimulationString,
    pSynchronisationHeader: pstring;

    pHartCommErrors: PHartCommErrors;
    pParity: pbyte;

    fEnabled,
    fDisabled,
    fReceptionFinished,
    fReceptionTimeOut: boolean;
    fReceptionTimeOutLimit: dword;
    fThreadReceptionString: string;

    fOnThreadReceiveChar,
    fOnThreadReceptionFinished: TSimpleEvent;

    fOnThreadReadChar: TRxCharRead;
    fOnThreadReadString,
    fOnThreadReceptionTimeOut: TRxStringRead;
  protected
    { Protected-Deklarationen }
    function GetSimulatedChar(var Ch: char): boolean;
    procedure Execute; override;
  public
    { Public-Deklarationen}
    constructor Create;
    destructor Destroy; override;
    procedure Resume;
    procedure Suspend;
  published
    { Published-Deklarationen }
    property OnThreadReceiveChar: TSimpleEvent read fOnThreadReceiveChar write fOnThreadReceiveChar;
    property OnThreadReceptionFinished: TSimpleEvent read fOnThreadReceptionFinished write fOnThreadReceptionFinished;

    property OnThreadReadChar: TRxCharRead read fOnThreadReadChar write fOnThreadReadChar;
    property OnThreadReadString: TRxStringRead read fOnThreadReadString write fOnThreadReadString;
    property OnThreadReceptionTimeOut: TRxStringRead read fOnThreadReceptionTimeOut write fOnThreadReceptionTimeOut;
  end;
{$ENDIF}

  TCustomHartDevice = class(TComponent)
  private
    { Private-Deklarationen}
    fHartDeviceType: THartDeviceType;
    fCommHandle: THandle;
    fFTDHandle: FT_HANDLE;
    fCommNumber: integer;
    fFTDeviceName: TFTDeviceNameString;
    fSmallTimeOut,
    fLargeTimeOut: TTimeOutController;
{$IFDEF WithThread}
    fHartReceiver: THartReceiver;
{$ENDIF}
    fWriteTimeOut,
    fReadTimeOut,
    fCharTransmitInterval,
    fCharTransferDuration,  // Attention! this value is in microseconds.
    fDelayParityByte,
    fMaxStringLength: word;
{$IFDEF WithThread}
    fHartReceptionString,
{$ENDIF}
    fSynchronisationHeader,
    fHartSimulationString: string;

    fBaudRateValue: dword;
    fParity: byte;
    fFixedBaudRate,
    fSimulationMode,
    fDeviceOpen,
    fDefectErrorEnabled,
    fDisruptionErrorEnabled,
    fExceptionErrorEnabled: boolean;
    fHartCommErrors: THartCommErrors;
    fActualHartCommand: THartCommand;

    fOnTransmitChar,
    fOnTransmissionFinished,
    fOnReceiveChar,
    fOnReceptionFinished: TSimpleEvent;
{$IFDEF WithThread}
    fOnReadChar: TRxCharRead;
    fOnReadString,
    fOnReceptionTimeOut: TRxStringRead;
{$ENDIF}
    procedure SetHartDeviceType (Value: THartDeviceType);
    procedure SetWriteTimeOut (Value: word);
    procedure SetReadTimeOut (Value: word);
    procedure SetCharTransmitInterval (Value: word);
    procedure SetDelayParityByte (Value: word);
    procedure SetFixedBaudRate (Value: boolean);
    procedure SetBaudRate (Value: dword);
    procedure SetMaxStringLength (Value: word);
    procedure SetSynchronisationHeader (Value: string);

    procedure ClearErrors;
    function TimeOutDetected: boolean;
    function ExecutionErrorDetected: boolean;
    function TransferErrorDetected: boolean;
    function ConfigurationErrorDetected: boolean;
    function DeviceErrorDetected: boolean;
    function AnyErrorDetected: boolean;

    function PutString (Source: string): boolean;
    function TransmitMessage (Source: string): boolean;
    function SimulateTransmision (Source: string): boolean;
    procedure SimulateReception (Count: byte);

{$IFDEF WithThread}
    procedure AssignThreadEvents;
    procedure UnAssignThreadEvents;
{$ELSE}
    function CharsAvailable (Value: byte): boolean;
    function GetSimulatedChar(var Ch: char): boolean;
    function ReceiveMessage (var Target: string): boolean;
{$ENDIF}

  protected
    { Protected-Deklarationen }
    property SynchronisationHeader: string read fSynchronisationHeader write SetSynchronisationHeader;
    property HartDeviceType: THartDeviceType read fHartDeviceType write SetHartDeviceType default dtSerialCom;

    property OnTransmitChar: TSimpleEvent read fOnTransmitChar write fOnTransmitChar;
    property OnTransmissionFinished: TSimpleEvent read fOnTransmissionFinished write fOnTransmissionFinished;

    property OnReceiveChar: TSimpleEvent read fOnReceiveChar write fOnReceiveChar;
    property OnReceptionFinished: TSimpleEvent read fOnReceptionFinished write fOnReceptionFinished;

{$IFDEF WithThread}
    property OnReadChar: TRxCharRead read fOnReadChar write fOnReadChar;
    property OnReadString: TRxStringRead read fOnReadString write fOnReadString;
    property OnReceptionTimeOut: TRxStringRead read fOnReceptionTimeOut write fOnReceptionTimeOut;
{$ENDIF}
  public
    { Public-Deklarationen}
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;

    procedure SetHartCommParameters (Source: THartCommParameters); overload;
    procedure SetHartCommParameters (RTO, WTO, CTI, DPB: word); overload;

    function Open (CommNumber: byte): boolean; overload;
    function Open (FTDeviceName: TFTDeviceNameString): boolean; overload;

    procedure SuspendReception;
    procedure ResumeReception;

    procedure Close;

    procedure PurgeRead;
    procedure PurgeWrite;
    procedure PurgeReadWrite;
  published
    { Published-Deklarationen }
    property HartCommErrors: THartCommErrors read fHartCommErrors;
    property WriteTimeOut: word read fWriteTimeOut write SetWriteTimeOut;
    property ReadTimeOut: word read fReadTimeOut write SetReadTimeOut;
    property CharTransmitInterval: word read fCharTransmitInterval write SetCharTransmitInterval default 0;
    property DelayParityByte: word read fDelayParityByte write SetDelayParityByte default 0;
    property FixedBaudRate: boolean read fFixedBaudrate write SetFixedBaudRate default true;
    property BaudRateValue: dword read fBaudRateValue write SetBaudRate default CBR_1200;
    property MaxStringLength: word read fMaxStringLength write SetMaxStringLength;
    property TimeOutErrorOccur: boolean read TimeOutDetected;
    property ExecutionErrorOccur: boolean read ExecutionErrorDetected;
    property TransferErrorOccur: boolean read TransferErrorDetected;
    property ConfigurationErrorOccur: boolean read ConfigurationErrorDetected;
    property DeviceErrorOccur: boolean read DeviceErrorDetected;
    property AnyErrorOccur: boolean read AnyErrorDetected;
    property ParityResult: byte read fParity;
    property DefectErrorEnabled: boolean read fDefectErrorEnabled write fDefectErrorEnabled default true;
    property DisruptionErrorEnabled: boolean read fDisruptionErrorEnabled write fDisruptionErrorEnabled default true;
    property ExceptionErrorEnabled: boolean read fExceptionErrorEnabled write fExceptionErrorEnabled default false;
  end;

  TOnTalkStarted = procedure (Command: byte) of object;
  TOnListenFinished = procedure (HartAnswer: THartAnswer) of object;

  PHartMaster = ^THartMaster;
  THartMaster = class(TCustomHartDevice)
  private
    { Private-Deklarationen}
    fAnswerAvailable: boolean;
{$IFDEF WithThread}
    fHartAnswer: THartAnswer;
    fOnReadAnswer: TOnReadAnswer;
{$ENDIF}
    fOnReadAnswerError: TOnReadAnswerError;
    fOnAnswerSimulationRequest: TOnAnswerSimulationRequest;

    fOnTalkStarted: TOnTalkStarted;
    fOnListenFinished: TOnListenFinished;

    procedure EvaluateAnswerStatusIntegrity(Source: THartAnswer);
{$IFDEF WithThread}
    procedure AnswerReception (RxString: string);
    procedure AnswerError (RxString: string);
{$ENDIF}
  protected
    { Protected-Deklarationen }
  public
    { Public-Deklarationen}
    constructor Create (AOwner: TComponent); override;

    procedure Talk (Command: byte; Data: string); overload;
    procedure Talk (Source: THartCommand); overload;
    procedure PutByte (B: byte);
    function MasterDialog (Command: byte; Data: string; WaitTime: word): THartAnswer;
    function ForceReceiverSimulation(Value: string): boolean;
    property SynchronisationHeader;
  published
    { Published-Deklarationen }
    property HartDeviceType;

    property OnTransmitChar;
    property OnTransmissionFinished;

    property OnReceiveChar;
    property OnReceptionFinished;

{$IFDEF WithThread}
    property OnReadAnswer: TOnReadAnswer read fOnReadAnswer write fOnReadAnswer;
{$ENDIF}
    property OnReadAnswerError: TOnReadAnswerError read fOnReadAnswerError write fOnReadAnswerError;
    property OnAnswerSimulationRequest: TOnAnswerSimulationRequest read fOnAnswerSimulationRequest write fOnAnswerSimulationRequest;

    property OnTalkStarted: TOnTalkStarted read fOnTalkStarted write fOnTalkStarted;
    property OnListenFinished: TOnListenFinished read fOnListenFinished write fOnListenFinished;
  end;

  PHartSlave = ^THartSlave;
  THartSlave = class(TCustomHartDevice)
  private
    { Private-Deklarationen}
{$IFDEF WithThread}
    fOnReadCommand: TOnReadCommand;
{$ENDIF}
    fOnReadCommandError: TOnReadCommandError;
    fOnReadCommandSimulation: TOnCommandSimulationRequest;
{$IFDEF WithThread}
    procedure CommandReception (RxString: string);
    procedure CommandError (RxString: string);
{$ENDIF}
  protected
    { Protected-Deklarationen }
{$IFNDEF WithThread}
    function CheckCommandPending: boolean;
{$ENDIF}
  public
    { Public-Deklarationen}
    constructor Create (AOwner: TComponent); override;

    procedure Talk (CommStatus, DeviceStatus: byte; Data: string); overload;
    procedure Talk (Source: THartAnswer); overload;
{$IFNDEF WithThread}
    function Listen: THartCommand;
{$ENDIF}
    function ForceReceiverSimulation(Value: string): boolean;
{$IFNDEF WithThread}
    property CommandPending: boolean read CheckCommandPending;
{$ENDIF}
    property SynchronisationHeader;
  published
    { Published-Deklarationen }
    property HartDeviceType;

    property OnTransmitChar;
    property OnTransmissionFinished;

    property OnReceiveChar;
    property OnReceptionFinished;

{$IFDEF WithThread}
    property OnReadCommand: TOnReadCommand read fOnReadCommand write fOnReadCommand;
{$ENDIF}
    property OnReadCommandError: TOnReadCommandError read fOnReadCommandError write fOnReadCommandError;
    property OnReadCommandSimulation: TOnCommandSimulationRequest read fOnReadCommandSimulation write fOnReadCommandSimulation;
  end;

  procedure Register;

  function CreateMasterMessage(CommandNumber: byte; Data: string): string; overload;
  function CreateMasterMessage(Source: THartCommand): string; overload;

  function CreateSlaveMessage(CommandNumber, CommStatus, DeviceStatus: byte; Data: string): string; overload;
  function CreateSlaveMessage(Source: THartAnswer): string; overload;

  function EvaluateCommandMessage(Source,Header: string; var Target: THartCommand): boolean;
  function EvaluateAnswerMessage(Source,Header: string; var Target: THartAnswer): boolean;

type
  THartHeader = string[3];

const
  CInvalidCommand: byte   = 255;

  CDemoMode: string = 'Demo Mode';
  CTimeOut: string  = 'TimeOut';
  CClosed: string   = 'Closed';

  CEmptyCommand: THartCommand = (CommandNumber: 255; Data: '');
  CDemoCommand: THartCommand = (CommandNumber: 255; Data: 'Demo Mode');

  CEmptyAnswer: THartAnswer = (CommandNumber: 255; CommStatus: 0; DeviceStatus: 0; Data: '');
  CDemoAnswer: THartAnswer = (CommandNumber: 255; CommStatus: 0; DeviceStatus: 0; Data: 'Demo Mode');
  CTimeOutAnswer: THartAnswer = (CommandNumber: 255; CommStatus: 0; DeviceStatus: 0; Data: 'TimeOut');

  CDefaultHart: THartCommParameters = (ReadTimeOut: 1000;
                                       WriteTimeOut: 100;
                                       CharTransmitInterval : 0;
                                       DelayParityByte: 0);
  CMasterHartHeader: THartHeader = #$FF#$02#$00;
  CSlaveHartHeader: THartHeader  = #$FF#$06#$00;


implementation

uses ConversionUtils, ProcessUtils;

procedure Register;
begin
  RegisterComponents('Hardware', [THartMaster]);
  RegisterComponents('Hardware', [THartSlave]);
end;


function CreateMasterMessage(CommandNumber: byte; Data: string): string;
begin
  Result:=CMasterHartHeader+   // master/slave synchronisation header
          chr(CommandNumber)+
          chr(System.Length(Data))+
          Data;
end;

function CreateMasterMessage(Source: THartCommand): string;
begin
  Result:=CMasterHartHeader+   // master/slave synchronisation header
          chr(Source.CommandNumber)+
          chr(System.Length(Source.Data))+
          Source.Data;
end;

function CreateSlaveMessage(CommandNumber, CommStatus, DeviceStatus: byte; Data: string): string;
begin
  Result:=CSlaveHartHeader+   // master/slave synchronisation header
          chr(CommandNumber)+
          chr(System.Length(Data)+2)+
          chr(CommStatus)+
          chr(DeviceStatus)+
          Data;
end;

function CreateSlaveMessage(Source: THartAnswer): string;
begin
  Result:=CSlaveHartHeader+   // master/slave synchronisation header
          chr(Source.CommandNumber)+
          chr(System.Length(Source.Data)+2)+
          chr(Source.CommStatus)+
          chr(Source.DeviceStatus)+
          Source.Data;
end;

function EvaluateCommandMessage(Source,Header: string; var Target: THartCommand): boolean;
var NewHartCommand: THartCommand;
    StringLength: word;
begin
  Result:=false;

  NewHartCommand:=CEmptyCommand;
  StringLength:=System.Length(Source);

  if not(StringLength<6) and
     // effectively the ord(Source[5]) must not be negative
     // but 8-bit values > 127 will be even passed as < 0
     not(ord(Source[5])>127) and
     (StringLength=(6+ord(Source[5]))) and
     (System.Pos(Header,Source)=1) then
     begin
     with NewHartCommand do
          begin
          CommandNumber:=ord(Source[4]);
          Data:=System.Copy(Source,6,ord(Source[5]));
          end;
     Result:=true;
     end;
  Target:=NewHartCommand;
end;

function EvaluateAnswerMessage(Source,Header: string; var Target: THartAnswer): boolean;
var NewHartAnswer: THartAnswer;
    StringLength: word;
begin
  Result:=false;

  NewHartAnswer:=CEmptyAnswer;
  StringLength:=System.Length(Source);

  if not(StringLength<8) and
     not(ord(Source[5])<2) and
     (StringLength=(6+ord(Source[5]))) and
     (System.Pos(Header,Source)=1) then
     begin
     with NewHartAnswer do
          begin
          CommandNumber:=ord(Source[4]);
          CommStatus:=ord(Source[6]);
          DeviceStatus:=ord(Source[7]);
          Data:=System.Copy(Source,8,ord(Source[5])-2);
          end;
     Result:=true;
     end;
  Target:=NewHartAnswer;
end;

// THartReceiver

{$IFDEF WithThread}
constructor THartReceiver.Create;
begin
  inherited Create(true);

  fEnabled:=false;

  pHartDeviceType:=nil;
  pComHandle:=nil;
  pFTDHandle:=nil;
  pSimulationMode:=nil;

  pReadTimeOut:=nil;
  pMaxStringLength:=nil;
  pReceptionString:=nil;
  pSimulationString:=nil;
  pSynchronisationHeader:=nil;

  pHartCommErrors:=nil;
  pParity:=nil;

  fOnThreadReceiveChar:=nil;
  fOnThreadReceptionFinished:=nil;

  fOnThreadReadChar:=nil;
  fOnThreadReadString:=nil;
  fOnThreadReceptionTimeOut:=nil;
end;

destructor THartReceiver.Destroy;
begin

  inherited Destroy;
end;

procedure THartReceiver.Resume;
begin
  if not(fEnabled) then
     begin
     fReceptionFinished:=false;
     fReceptionTimeOut:=false;
     fThreadReceptionString:='';

     inherited Resume;
     fEnabled:=true;
     end;
end;

procedure THartReceiver.Suspend;
begin
  if fEnabled then
     begin
     fEnabled:=false;
     while not(fDisabled) do;
     inherited Suspend;
     end;
end;

function THartReceiver.GetSimulatedChar(var Ch: char): boolean;
begin
  if (pSimulationString<>nil) and
     (pSimulationString^<>'') then
     begin
     Ch:=pSimulationString^[1];
     System.Delete(pSimulationString^,1,1);
     Result:=true;
     end
  else Result:=false;
end;

procedure THartReceiver.Execute;
var NewCharReceived: boolean;
    StringLength,
    NumberBytes,
    fParity: byte;
    NewChar: char;
begin
  NewCharReceived:=false;
  NumberBytes:=0;

  while True do
        begin
        if fEnabled then
           begin
           if fDisabled then fDisabled:=false;
           if (pHartDeviceType<>nil) and
              (pSimulationMode<>nil) then
              begin

              if not(pSimulationMode^) then
                 begin
                 if (pHartDeviceType^=dtSerialCom) and
                    (pComHandle<>nil) and
                    (GetCOMCharsInQue(pComHandle^)>0) then
                    NewCharReceived:=GetCharFromCOM(pComHandle^,NewChar);
                 if (pHartDeviceType^=dtFTDevice) and
                    (pFTDHandle<>nil) and
                    (GetFTDCharsInQue(pFTDHandle^)>0) then
                    NewCharReceived:=GetCharFromFTD(pFTDHandle^,NewChar);
                 end
              else NewCharReceived:=GetSimulatedChar(NewChar);

              if NewCharReceived then
                 begin

                 if Assigned(fOnThreadReceiveChar) then fOnThreadReceiveChar;

                 if Assigned(fOnThreadReadChar) then fOnThreadReadChar(NewChar)
                 else // not(Assigned(fOnThreadReadChar))
                    begin

                    if (fThreadReceptionString='') then
                       begin
                       NumberBytes:=0;
                       fReceptionTimeOutLimit:=GetTickCount;
                       if (pReadTimeOut<>nil) then
                          fReceptionTimeOutLimit:=fReceptionTimeOutLimit+pReadTimeOut^
                       else fReceptionTimeOutLimit:=fReceptionTimeOutLimit+CDefaultReadTimeOut;
                       end;

                    fThreadReceptionString:=fThreadReceptionString+NewChar;
                    NewCharReceived:=false;

                    StringLength:=System.Length(fThreadReceptionString);

                    if not(StringLength<3) and
                       (pSynchronisationHeader<>nil) then
                       begin

                       if (System.Pos(pSynchronisationHeader^,fThreadReceptionString)=1) then
                          begin
                          if (StringLength=5) then
                             begin
                             NumberBytes:=ord(fThreadReceptionString[5]);

                             if (pReadTimeOut<>nil) then // wait more patiently when synchronisation received
                                fReceptionTimeOutLimit:=fReceptionTimeOutLimit+pReadTimeOut^
                             else fReceptionTimeOutLimit:=fReceptionTimeOutLimit+CDefaultReadTimeOut;

                             end;
                          if (StringLength>5) then
                             begin
                             fReceptionFinished:=(StringLength=(NumberBytes+6));
                             end;
                          end
                       else System.Delete(fThreadReceptionString,1,1);

                       end;
                    end;
                 end
              else // no any char available, so control the time out
                 begin
                 if (fThreadReceptionString<>'') and not(fReceptionTimeOut) then
                    fReceptionTimeOut:=not(fReceptionTimeOutLimit>GetTickCount);
                 end;

              if fReceptionFinished or fReceptionTimeOut then
                 begin

                 if (fThreadReceptionString<>'') and Assigned(fOnThreadReceptionFinished) then fOnThreadReceptionFinished;

                 if not(fReceptionTimeOut) then
                    begin
                    fParity:=ord(fThreadReceptionString[1]) xor GetStringParityByte(fThreadReceptionString);
                    if (pHartCommErrors<>nil) and (fParity<>0) then
                       pHartCommErrors^.TransferErrors:=pHartCommErrors^.TransferErrors+[teParity];
                    if (pParity<>nil) then
                        pParity^:=fParity;
                    if Assigned(fOnThreadReadString) then
                       fOnThreadReadString(fThreadReceptionString)
                    else if (pReceptionString<>nil) then
                            begin
                            while (pReceptionString^<>'') do; // wait ages until serviced
                            pReceptionString^:=fThreadReceptionString;
                            end;
                    end
                 else // ReceptionTimeOut
                    begin
                    if (pHartCommErrors<>nil) then
                       pHartCommErrors^.ExecutionErrors:=pHartCommErrors^.ExecutionErrors+[eeTimeOut];
                    if Assigned(fOnThreadReceptionTimeOut) then
                       fOnThreadReceptionTimeOut(fThreadReceptionString)
                    else if (pReceptionString<>nil) then
                            begin
                            while (pReceptionString^<>'') do; // wait ages until serviced
                            pReceptionString^:=fThreadReceptionString;
                            end;
                    end;

                 fReceptionFinished:=false;
                 fReceptionTimeOut:=false;
                 fThreadReceptionString:='';
                 end;
              end;
           end
        else // not(fEnabled)
           if not(fDisabled) then fDisabled:=true;
        end;
end;
{$ENDIF}

// TCustomHartDevice

procedure TCustomHartDevice.SetHartDeviceType (Value: THartDeviceType);
begin
// Hart Device Type modification is allowed only for closed devices
  if not(fDeviceOpen) and
     (Value<>fHartDeviceType) then
     begin
     // FTD2XX (USB) Mode activation is allowed only if the FTD2XX Library is available
     if (Value=dtFTDevice) and FTD2XXLibraryAvailable then
        fHartDeviceType:=dtFTDevice
     else fHartDeviceType:=dtSerialCom;
     end;
end;

procedure TCustomHartDevice.SetWriteTimeOut (Value: word);
begin
  // Don't accept Time Out > 5000 (which is 5s)
  if (fWriteTimeOut<>Value) and not(fWriteTimeOut>5000) then
     fWriteTimeOut:=Value;
end;

procedure TCustomHartDevice.SetReadTimeOut (Value: word);
begin
  // Don't accept Time Out > 60000 (which is 60s = 1min)
  if (fReadTimeOut<>Value) and not(fReadTimeOut>60000) then
     fReadTimeOut:=Value;
end;

procedure TCustomHartDevice.SetCharTransmitInterval (Value: word);
begin
  // Don't accept Char Transmit Interval > 1000
  if (fCharTransmitInterval<>Value) and not(fCharTransmitInterval>1000) then
     begin
     fCharTransmitInterval:=Value;
     end;
end;

procedure TCustomHartDevice.SetDelayParityByte (Value: word);
begin
  // Don't accept Delay Parity Byte > 1000 (which is 1s)
  if (Value<>fDelayParityByte) and not(Value>1000) then fDelayParityByte:=Value;
end;

procedure TCustomHartDevice.SetFixedBaudRate (Value: boolean);
begin
// Fixed Baud Rate Mode modification is allowed only for closed devices
  if not(fDeviceOpen) and
    (Value<>fFixedBaudRate) then
     begin
     fFixedBaudRate:=Value;
     if fFixedBaudRate then
        begin
        if (fHartDeviceType=dtSerialCom) then
           fBaudRateValue:=CBR_1200
        else // (fHartDeviceType=dtFTDevice)
           fBaudRateValue:=FT_BAUD_1200;
        end;
     // HartComm has always fByteSize=8 and fStopBits=1
     fCharTransferDuration:=(1000000*(8+1+1)) DIV fBaudRateValue;
     end;
end;

procedure TCustomHartDevice.SetBaudRate (Value: dword);
begin
// Baud Rate Value modification is allowed only for closed devices
  if not(fDeviceOpen) and
     not(fFixedBaudRate) and // and for disabled fixed rate
     (Value<>fBaudRateValue) then
     begin
     fBaudRateValue:=Value;
     // HartComm has always fByteSize=8 and fStopBits=1
     fCharTransferDuration:=(1000000*(8+1+1)) DIV fBaudRateValue;
     end;
end;

procedure TCustomHartDevice.SetMaxStringLength(Value: word);
begin
  if (Value<>fMaxStringLength) and not(Value>255) then
     fMaxStringLength:=Value;
end;

procedure TCustomHartDevice.SetSynchronisationHeader(Value: string);
begin
  // Hart Synchronisation Header consists 3 characters
  if (Value<>fSynchronisationHeader) and (System.Length(Value)=3)
     then fSynchronisationHeader:=Value;
end;

procedure TCustomHartDevice.ClearErrors;
begin
  with fHartCommErrors do
       begin
       ExecutionErrors:=([]);
       TransferErrors:=([]);
       DeviceErrors:=([]);
       end;
end;

function TCustomHartDevice.TimeOutDetected: boolean;
begin
  Result:=(eeTimeOut in fHartCommErrors.ExecutionErrors);
end;

function TCustomHartDevice.ExecutionErrorDetected: boolean;
begin
  Result:=(fHartCommErrors.ExecutionErrors <> ([]));
end;

function TCustomHartDevice.TransferErrorDetected: boolean;
begin
  Result:=(fHartCommErrors.TransferErrors <> ([]));
end;

function TCustomHartDevice.ConfigurationErrorDetected: boolean;
begin
  Result:=(deWrongData in fHartCommErrors.DeviceErrors);
end;

function TCustomHartDevice.DeviceErrorDetected: boolean;
begin
  Result:=((deDeviceException in fHartCommErrors.DeviceErrors) or
           (deDeviceProblems in fHartCommErrors.DeviceErrors) or
           (deDeviceFailure in fHartCommErrors.DeviceErrors));
end;

function TCustomHartDevice.AnyErrorDetected: boolean;
begin
  Result:=(fHartCommErrors.ExecutionErrors <> ([])) or
          (fHartCommErrors.TransferErrors <> ([])) or
          (fHartCommErrors.DeviceErrors <> ([]));
end;

function TCustomHartDevice.PutString (Source: string): boolean;
var fMessageTransmissionTime, fLength: word;
    Success: boolean;
begin
  fLength:=System.Length(Source);
  if (fLength>0) then
     fMessageTransmissionTime:=System.Round(fLength*fCharTransferDuration/1000)
  else fMessageTransmissionTime:=0;
  if (fHartDeviceType=dtSerialCom) then
     begin
     if (fWriteTimeOut=0) then
        Success:=PutStringToCOM(fCommHandle,Source)
     else Success:=PutStringToCOM(fCommHandle,Source,fSmallTimeOut,fMessageTransmissionTime+fWriteTimeOut);
     end
  else // (fHartDeviceType=dtFTDevice)
     begin
     if (fWriteTimeOut=0) then
        Success:=PutStringToFTD(fFTDHandle,Source)
     else Success:=PutStringToFTD(fFTDHandle,Source,fSmallTimeOut,fMessageTransmissionTime+fWriteTimeOut);
     end;

  Result:=Success;
end;

function TCustomHartDevice.TransmitMessage (Source: string): boolean;
var NumberBytes,I: integer;
    Success: boolean;
    tParity: byte;
begin
  NumberBytes:=System.Length(Source);
  tParity:=ord(Source[1]) xor GetStringParityByte(Source);
  Success:=false;
  
  if (fDelayParityByte=0) then
     begin
     Source:=Source+chr(tParity);
     if (fCharTransmitInterval > 0) or
        Assigned(fOnTransmitChar) then
        begin
        for I:=1 to NumberBytes+1 do
            begin
            Success:=PutString(Source[I]);
            if Success then
               begin
               if (fCharTransmitInterval > 0) then
                  WaitMicroseconds(fCharTransferDuration+fCharTransmitInterval*1000);
               if Assigned(fOnTransmitChar) then fOnTransmitChar;
               end
            else break;
            end;
        end
     else Success:=PutString(Source);
     end
  else // fDelayParityByte>0
     begin
     if (fCharTransmitInterval > 0) or
        Assigned(fOnTransmitChar) then
        begin
        for I:=1 to NumberBytes do
            begin
            Success:=PutString(Source[I]);
            if Success then
               begin
               if (fCharTransmitInterval > 0) then
                  WaitMicroseconds(fCharTransferDuration+fCharTransmitInterval*1000);
               if Assigned(fOnTransmitChar) then fOnTransmitChar;
               end
            else break;
            end;
        end
     else Success:=PutString(Source);
     WaitMilliseconds(fDelayParityByte);
     if Success then
        Success:=PutString(string(char(tParity)));
     if (fCharTransmitInterval > 0) then
        WaitMicroseconds(fCharTransferDuration+fCharTransmitInterval*1000);
     if Assigned(fOnTransmitChar) then fOnTransmitChar;
     end;

  if Success and
     Assigned(fOnTransmissionFinished) then fOnTransmissionFinished;

  Result:=Success;
end;

function TCustomHartDevice.SimulateTransmision (Source: string): boolean;
var I, NumberBytes: byte;
begin
  NumberBytes:=System.Length(Source);

  if Assigned(fOnTransmitChar) then
     begin
     for I:=1 to NumberBytes+1 do  // Don't forget the parity byte !
         begin
         WaitMicroseconds(fCharTransferDuration+fCharTransmitInterval*1000);
         fOnTransmitChar;
         end;
     end
  else WaitMilliseconds((fCharTransferDuration*(NumberBytes+1)) DIV 1000);
  if Assigned(fOnTransmissionFinished) then fOnTransmissionFinished;
  Result:=true;
end;

procedure TCustomHartDevice.SimulateReception(Count: byte);
var I: byte;
begin
  if Assigned(fOnReceiveChar) then
     begin
     for I:=1 to Count do
         begin
         WaitMicroseconds(fCharTransferDuration+fCharTransmitInterval*1000);
         fOnReceiveChar;
         end;
     end
  else WaitMilliseconds((fCharTransferDuration*(Count+1)) DIV 1000);
  if (Count>0) and Assigned(fOnReceptionFinished) then fOnReceptionFinished;
end;

{$IFDEF WithThread}
procedure TCustomHartDevice.AssignThreadEvents;
begin
  with fHartReceiver do
       begin
       if Assigned(fOnReceiveChar) then
          OnThreadReceiveChar:=OnReceiveChar;
       if Assigned(fOnReceptionFinished) then
          OnThreadReceptionFinished:=OnReceptionFinished;

       if Assigned(fOnReadChar) then
          OnThreadReadChar:=OnReadChar;
       if Assigned(fOnReadString) then
          OnThreadReadString:=OnReadString;
       if Assigned(fOnReceptionTimeOut) then
          OnThreadReceptionTimeOut:=OnReceptionTimeOut;
       end;
end;

procedure TCustomHartDevice.UnAssignThreadEvents;
begin
  with fHartReceiver do
       begin
       OnThreadReceiveChar:=nil;
       OnThreadReceptionFinished:=nil;

       OnThreadReadChar:=nil;
       OnThreadReadString:=nil;
       OnThreadReceptionTimeOut:=nil;
       end;
end;
{$ELSE}

function TCustomHartDevice.CharsAvailable (Value: byte): boolean;
begin
  if not(fSimulationMode) then
     begin
     if (fHartDeviceType=dtSerialCom) then
        Result:=not(GetCOMCharsInQue(fCommHandle)<Value)
     else // (fHartDeviceType=dtFTDevice)
        Result:=not(GetFTDCharsInQue(fFTDHandle)<Value);
     end
  else // fSimulationMode
     Result:=not(System.Length(fHartSimulationString)<Value);
end;

function TCustomHartDevice.GetSimulatedChar(var Ch: char): boolean;
begin
  if (fHartSimulationString<>'') then
     begin
     Ch:=fHartSimulationString[1];
     System.Delete(fHartSimulationString,1,1);
     Result:=true;
     end
  else Result:=false;
end;

function TCustomHartDevice.ReceiveMessage(var Target: string): boolean;
var ReceptionFinished,
    ReceptionTimeOut,
    NewCharReceived: boolean;
    fReceptionString: string;
    StringLength,
    NumberBytes: byte;
    NewChar: char;
begin
  Result:=false;

  fReceptionString:='';
  ReceptionFinished:=false;
  ReceptionTimeOut:=false;
  NewCharReceived:=false;
  NumberBytes:=0;

  fLargeTimeOut.Initialize(fReadTimeOut);

  repeat
     if not(fSimulationMode) then
        begin
        if (fHartDeviceType=dtSerialCom) and
           (GetCOMCharsInQue(fCommHandle)>0) then
           NewCharReceived:=GetCharFromCOM(fCommHandle,NewChar);
        if (fHartDeviceType=dtFTDevice) and
           (GetFTDCharsInQue(fFTDHandle)>0) then
           NewCharReceived:=GetCharFromFTD(fFTDHandle,NewChar);
        end
     else NewCharReceived:=GetSimulatedChar(NewChar);

     if NewCharReceived then
        begin
        if Assigned(fOnReceiveChar) then fOnReceiveChar;

        if (fReceptionString='') then
           begin
           NumberBytes:=0;
           fLargeTimeOut.Extend(fReadTimeOut);
           end;

        fReceptionString:=fReceptionString+NewChar;
        NewCharReceived:=false;

        StringLength:=System.Length(fReceptionString);

        if not(StringLength<3) then
           begin
           if (System.Pos(fSynchronisationHeader,fReceptionString)=1) then
              begin
              if (StringLength=5) then
                 begin
                 NumberBytes:=ord(fReceptionString[5]);
                 fLargeTimeOut.Extend(fReadTimeOut);
                 end;
              if (StringLength>5) then
                 begin
                 ReceptionFinished:=(StringLength=(NumberBytes+6));
                 end;
              end
           else System.Delete(fReceptionString,1,1);
           end;
        end
     else // no any char available, so control the time out
        begin
        if not(ReceptionTimeOut) then
           ReceptionTimeOut:=fLargeTimeOut.Elapsed;
        end;
  until ReceptionFinished or ReceptionTimeOut;

  if (fReceptionString<>'') and Assigned(fOnReceptionFinished) then fOnReceptionFinished;

  if not(ReceptionTimeOut) then
     begin
     fParity:=ord(fReceptionString[1]) xor GetStringParityByte(fReceptionString);
     if (fParity<>0) then
        fHartCommErrors.TransferErrors:=fHartCommErrors.TransferErrors+[teParity]
     else Result:=true;
     end
  else // ReceptionTimeOut
     begin
     fHartCommErrors.ExecutionErrors:=fHartCommErrors.ExecutionErrors+[eeTimeOut];
     end;
  Target:=fReceptionString;
end;
{$ENDIF}

constructor TCustomHartDevice.Create (AOwner: TComponent);
begin
  inherited Create (AOwner);

  fHartDeviceType:=dtSerialCom;

  fCommHandle:=INVALID_HANDLE_VALUE;  // set Com closed
  fFTDHandle:=FT_INVALID_HANDLE;      // set FTD closed

  fSmallTimeOut := TTimeOutController.Create;
  fLargeTimeOut := TTimeOutController.Create;

  fHartDeviceType:=dtSerialCom;

  fCommNumber:=-1;
  fFTDeviceName:='';

  fDeviceOpen:=false;
  fFixedBaudrate:=true;
  fBaudRateValue:=CBR_1200;
  // HartComm has always fByteSize=8 and fStopBits=1
  fCharTransferDuration:=(1000000*(8+1+1)) DIV fBaudRateValue;

  fSimulationMode:=false;
  fHartSimulationString:='';

  fReadTimeOut:=CDefaultHart.ReadTimeOut;
  fWriteTimeOut:=CDefaultHart.WriteTimeOut;

  fCharTransmitInterval:=CDefaultHart.CharTransmitInterval;
  fDelayParityByte:=CDefaultHart.DelayParityByte;

  fDefectErrorEnabled:=true;
  fDisruptionErrorEnabled:=true;
  fExceptionErrorEnabled:=false;

  ClearErrors;
  fHartCommErrors.ExecutionErrors:=([eeNotOpened]);

{$IFDEF WithThread}
  fHartReceiver:=THartReceiver.Create;
  with fHartReceiver do
       begin
       FreeOnTerminate:=true;
       Priority:=tpLower;

       pHartDeviceType:=@fHartDeviceType;
       pComHandle:=@fCommHandle;
       pFTDHandle:=@fFTDHandle;
       pSimulationMode:=@fSimulationMode;

       pReadTimeOut:=@fReadTimeOut;
       pMaxStringLength:=@fMaxStringLength;

       pReceptionString:=@fHartReceptionString;
       pSimulationString:=@fHartSimulationString;
       pSynchronisationHeader:=@fSynchronisationHeader;

       pHartCommErrors:=@fHartCommErrors;
       pParity:=@fParity;

       UnAssignThreadEvents;
       end;
  fHartReceptionString:='';
{$ENDIF}

  fActualHartCommand:=CEmptyCommand;
end;

destructor TCustomHartDevice.Destroy;
begin
  CloseSerialPort(fCommHandle);
  CloseFTDIDevice(fFTDHandle);
  fDeviceOpen:=false;

  fSmallTimeOut.Free;
  fLargeTimeOut.Free;

  inherited Destroy;
end;

procedure TCustomHartDevice.SetHartCommParameters (Source: THartCommParameters);
begin
  fReadTimeOut:=Source.ReadTimeOut;
  fWriteTimeOut:=Source.WriteTimeOut;
  fCharTransmitInterval:=Source.CharTransmitInterval;
  fDelayParityByte:=Source.DelayParityByte;
end;

procedure TCustomHartDevice.SetHartCommParameters (RTO, WTO, CTI, DPB: word);
begin
  fReadTimeOut:=RTO;
  fWriteTimeOut:=WTO;
  fCharTransmitInterval:=CTI;
  fDelayParityByte:=DPB;
end;

function TCustomHartDevice.Open (CommNumber: byte): boolean;
var lComSettings :TFComSettings;
begin
  fDeviceOpen:=false;
  if ((fHartDeviceType=dtSerialCom) and (CommNumber in [1..8])) then
     begin
     with lComSettings do
          begin
          Number:=CommNumber;
          BaudRate:=fBaudRateValue;
          ByteSize:=8;
          StopBits:=ONESTOPBIT;
          Parity:=ODDPARITY;
          // the EOSChar, XonChar and XoffChar are irrelevat
          Flags:=dcb_RtsControlEnable and not(dcb_DtrControlEnable); // dcb_Binary is set by default
          end;

     fCommHandle:=OpenSerialPort(lComSettings);

     if SerialPortOpened(fCommHandle) then
        begin
        fSimulationMode:=false;
        fDeviceOpen:=true;
        fCommNumber:=CommNumber;
        ClearErrors;
        PurgeReadWrite;
        end;
     end
  else if (CommNumber=0) then // this is for simulation mode
          begin
          fSimulationMode:=true;
          fDeviceOpen:=true;
          fCommNumber:=0;
          ClearErrors;
          end;

{$IFDEF WithThread}
  if fDeviceOpen then
     begin
     with fHartReceiver do
          begin
          AssignThreadEvents;

          {if Assigned(fOnReceiveChar) then
             OnThreadReceiveChar:=OnReceiveChar;
          if Assigned(fOnReceptionFinished) then
             OnThreadReceptionFinished:=OnReceptionFinished;

          if Assigned(fOnReadChar) then
             OnThreadReadChar:=OnReadChar;
          if Assigned(fOnReadString) then
             OnThreadReadString:=OnReadString;
          if Assigned(fOnReceptionTimeOut) then
             OnThreadReceptionTimeOut:=OnReceptionTimeOut;}

          pComHandle:=@fCommHandle;
          Resume;
          end;
     end;
{$ENDIF}

  Result:=fDeviceOpen;
end;

function TCustomHartDevice.Open (FTDeviceName: TFTDeviceNameString): boolean;
var lTDSettings: TFTDSettings;
begin
  fDeviceOpen:=false;
  if (fHartDeviceType=dtFTDevice) and (FTDeviceName<>'') then
     begin
     with lTDSettings do
          begin
          FTDOpenMode:=omByDescription;
          DeviceName:=FTDeviceName;
          SerialNumber:='';
          BaudRate:=fBaudRateValue;
          ByteSize:=FT_DATA_BITS_8;
          StopBits:=FT_STOP_BITS_1;
          Parity:=FT_PARITY_ODD;
          // the EOSChar, XonChar and XoffChar are irrelevat
          Flags:=0; // FT_dcb_Binary is set by default
          end;

     fFTDHandle:=OpenFTDIDevice(lTDSettings);

     if FTDIDeviceOpened(fFTDHandle) then
        begin
        fSimulationMode:=false;
        fDeviceOpen:=true;
        fFTDeviceName:=FTDeviceName;
        ClearErrors;
        PurgeReadWrite;
        end;
     end
  else if (FTDeviceName='') then // this is for simulation mode
          begin
          fSimulationMode:=true;
          fDeviceOpen:=true;
          fFTDeviceName:='';
          ClearErrors;
          end;

{$IFDEF WithThread}
  if fDeviceOpen then
     begin
     with fHartReceiver do
          begin
          AssignThreadEvents;

          {if Assigned(fOnReceiveChar) then
             OnThreadReceiveChar:=OnReceiveChar;
          if Assigned(fOnReceptionFinished) then
             OnThreadReceptionFinished:=OnReceptionFinished;

          if Assigned(fOnReadChar) then
             OnThreadReadChar:=OnReadChar;
          if Assigned(fOnReadString) then
             OnThreadReadString:=OnReadString;
          if Assigned(fOnReceptionTimeOut) then
             OnThreadReceptionTimeOut:=OnReceptionTimeOut;}

          pFTDHandle:=@fFTDHandle;
          Resume;
          end;
     end;
{$ENDIF}

  Result:=fDeviceOpen;
end;

procedure TCustomHartDevice.SuspendReception;
begin
{$IFDEF WithThread}
  fHartReceiver.Suspend;
{$ENDIF}
  fSmallTimeOut.Terminate;
  fLargeTimeOut.Terminate;
end;

procedure TCustomHartDevice.ResumeReception;
begin
  PurgeRead;
{$IFDEF WithThread}
  AssignThreadEvents;
  fHartReceiver.Resume;
{$ENDIF}
end;

procedure TCustomHartDevice.Close;
begin
  SuspendReception;

  if (fHartDeviceType=dtSerialCom) then
     begin
     CloseSerialPort(fCommHandle);
     fCommNumber:=-1;
     end
  else // (fHartDeviceType=dtFTDevice)
     begin
     CloseFTDIDevice(fFTDHandle);
     fFTDeviceName:='';
     end;
  fDeviceOpen:=false;
  fSimulationMode:=false;
  fHartCommErrors.ExecutionErrors:=([eeNotOpened]);
end;

procedure TCustomHartDevice.PurgeRead;
begin
  if (fHartDeviceType=dtSerialCom) then
     PurgeCOMReadBuffer(fCommHandle)
  else // (fHartDeviceType=dtFTDevice)
     PurgeFTDReadBuffer(fFTDHandle);
end;

procedure TCustomHartDevice.PurgeWrite;
begin
  if (fHartDeviceType=dtSerialCom) then
     PurgeCOMWriteBuffer(fCommHandle)
  else // (fHartDeviceType=dtFTDevice)
     PurgeFTDWriteBuffer(fFTDHandle);
end;

procedure TCustomHartDevice.PurgeReadWrite;
begin
  if (fHartDeviceType=dtSerialCom) then
     PurgeCOMReadWriteBuffer(fCommHandle)
  else // (fHartDeviceType=dtFTDevice)
     PurgeFTDReadWriteBuffer(fFTDHandle);
end;

// THartMaster

procedure THartMaster.EvaluateAnswerStatusIntegrity(Source: THartAnswer);
begin
  with Source do
       begin
       if (fActualHartCommand.CommandNumber <> CommandNumber) then
          fHartCommErrors.TransferErrors:=fHartCommErrors.TransferErrors+[teDifferentCommand];

       if (CommStatus <> 0) then
          begin
          if ((CommStatus and $80) <> 0) then fHartCommErrors.TransferErrors:=fHartCommErrors.TransferErrors+[teHartFormat]
          else fHartCommErrors.DeviceErrors:=fHartCommErrors.DeviceErrors+[deWrongData];
          end;

       if fDefectErrorEnabled and ((DeviceStatus and $80) <> 0) then fHartCommErrors.DeviceErrors:=fHartCommErrors.DeviceErrors+[deDeviceFailure];
       if fDisruptionErrorEnabled and ((DeviceStatus and $0F) <> 0) then fHartCommErrors.DeviceErrors:=fHartCommErrors.DeviceErrors+[deDeviceProblems];
       if fExceptionErrorEnabled and ((DeviceStatus and $70) <> 0) then fHartCommErrors.DeviceErrors:=fHartCommErrors.DeviceErrors+[deDeviceException];
       // fActualHartCommand.CommandNumber:=CommandNumber;
       end;
end;

{$IFDEF WithThread}
procedure THartMaster.AnswerReception (RxString: string);
var NewHartAnswer: THartAnswer;
begin
  if EvaluateAnswerMessage(RxString,fSynchronisationHeader,NewHartAnswer) and
     not(ExecutionErrorDetected) and
     not(teParity in fHartCommErrors.TransferErrors) then
     begin
     EvaluateAnswerStatusIntegrity(NewHartAnswer);
     if Assigned(fOnReadAnswer) then
        fOnReadAnswer(NewHartAnswer,fHartCommErrors)
     else
        begin
        fHartAnswer:=NewHartAnswer;
        fAnswerAvailable:=true;
        end;
     end
  else  // the answer is anyway incorrect
     begin
     if Assigned(fOnReadAnswerError) then
        begin
        fOnReadAnswerError(fActualHartCommand.CommandNumber,NewHartAnswer.CommandNumber,fHartCommErrors);
        ClearErrors;  // ???
        end
     else
        begin
        // fHartAnswer:=NewHartAnswer;
        fAnswerAvailable:=true;
        end;
     end;
end;

procedure THartMaster.AnswerError (RxString: string);
var NewHartAnswer: THartAnswer;
begin
  EvaluateAnswerMessage(RxString,fSynchronisationHeader,NewHartAnswer);
  EvaluateAnswerStatusIntegrity(NewHartAnswer);
  if Assigned(fOnReadAnswerError) then
     begin
     fOnReadAnswerError(fActualHartCommand.CommandNumber,NewHartAnswer.CommandNumber,fHartCommErrors);
     ClearErrors;  // ???
     end
  else
     begin
     // fHartAnswer:=NewHartAnswer;
     fAnswerAvailable:=true;
     end;
end;
{$ENDIF}

constructor THartMaster.Create(AOwner: TComponent);
begin
  inherited;
  SynchronisationHeader:=CSlaveHartHeader;

  fAnswerAvailable:=false;

{$IFDEF WithThread}
  with fHartReceiver do
       begin
       // OnThreadReadChar:=nil;
       OnThreadReadString:=AnswerReception;
       OnThreadReceptionTimeOut:=AnswerError;
       end;
{$ENDIF}
end;

procedure THartMaster.PutByte(B: byte);
begin
  ClearErrors;

  if fDeviceOpen then
     begin
     if not(fSimulationMode) then
        begin
        PutString(string(chr(B)));
        WaitMicroseconds(fCharTransferDuration);
        end
     else SimulateTransmision(char(B));  // simulation mode
     end
  else // not(fDeviceOpen)
       fHartCommErrors.ExecutionErrors:=([eeNotOpened]);
end;

procedure THartMaster.Talk(Command: byte; Data: string);
var fTransmitString: string;
begin
  ClearErrors;
  PurgeRead;

  if fDeviceOpen then
     begin
     fActualHartCommand.CommandNumber:=Command;
     fActualHartCommand.Data:=Data;
     fTransmitString:=CreateMasterMessage(Command,Data);
     if not(fSimulationMode) then
        TransmitMessage(fTransmitString)
     else // simulation mode
          SimulateTransmision(fTransmitString);
     end
  else // not(fDeviceOpen)
       fHartCommErrors.ExecutionErrors:=([eeNotOpened]);
end;

procedure THartMaster.Talk (Source: THartCommand);
var fTransmitString: string;
begin
  ClearErrors;
  PurgeRead;

  if fDeviceOpen then
     begin
     fActualHartCommand:=Source;
     fTransmitString:=CreateMasterMessage(Source);
     if not(fSimulationMode) then
        TransmitMessage(fTransmitString)
     else // simulation mode
        SimulateTransmision(fTransmitString);
     end
  else // not(fDeviceOpen)
       fHartCommErrors.ExecutionErrors:=([eeNotOpened]);
end;

function THartMaster.MasterDialog(Command: byte; Data: string; WaitTime: word): THartAnswer;
var NewHartAnswer: THartAnswer;
{$IFNDEF WithThread}
    RxString: string;
{$ENDIF}
begin
  Result:=CEmptyAnswer;
  fAnswerAvailable:=false;

  if Assigned(fOnTalkStarted) then fOnTalkStarted(Command);
  
  Talk (Command, Data);

  WaitMilliseconds(500);

  if (WaitTime>0)and
{$IFDEF WithThread}
     not(Assigned(fOnReadAnswer)) and
{$ENDIF}
     not(AnyErrorDetected) then
     begin

     if not(fSimulationMode) then
        begin
        fLargeTimeOut.Initialize(WaitTime+fReadTimeOut);

        while not(fAnswerAvailable) and not(fLargeTimeOut.Elapsed) do
              begin
{$IFNDEF WithThread}
              fAnswerAvailable:=CharsAvailable(6);
{$ENDIF}
              end;

        if fAnswerAvailable then
           begin
{$IFDEF WithThread}
           Result:=fHartAnswer;
{$ELSE}
           if ReceiveMessage(RxString) and
              EvaluateAnswerMessage(RxString,fSynchronisationHeader,NewHartAnswer) then
              begin
              EvaluateAnswerStatusIntegrity(NewHartAnswer);
              Result:=NewHartAnswer;
              end;
{$ENDIF}
           fAnswerAvailable:=false;
           end
        else
           begin
           fHartCommErrors.ExecutionErrors:=([eeTimeOut]);
           Result:=CTimeOutAnswer;
           end;
        end
     else // SimulationMode
        begin
        if Assigned(fOnAnswerSimulationRequest) then
           begin
           fOnAnswerSimulationRequest(fActualHartCommand,NewHartAnswer);
           SimulateReception(System.Length(NewHartAnswer.Data)+8);
           Result:=NewHartAnswer;
           end
        else // not(Assigned(fOnAnswerSimulationRequest))
           begin
           fLargeTimeOut.Initialize(WaitTime+fReadTimeOut);

           while not(fAnswerAvailable) and not(fLargeTimeOut.Elapsed) do
                 begin
{$IFNDEF WithThread}
                 fAnswerAvailable:=fHartSimulationString<>'';
{$ENDIF}
                 end;
           if fAnswerAvailable then
               begin
{$IFDEF WithThread}
              Result:=fHartAnswer;
{$ELSE}
              if (fHartSimulationString<>'') and
                 EvaluateAnswerMessage(fHartSimulationString,fSynchronisationHeader,NewHartAnswer) then
                 begin
                 SimulateReception(System.Length(fHartSimulationString));
                 EvaluateAnswerStatusIntegrity(NewHartAnswer);
                 fHartSimulationString:='';
                 Result:=NewHartAnswer;
                 end;
{$ENDIF}
              fAnswerAvailable:=false;
              end;
           end;
        end;
     end
  else Result.CommandNumber:=Command;

  if {fAnswerAvailable and} Assigned(fOnListenFinished) then fOnListenFinished(Result);
  
end;

function THartMaster.ForceReceiverSimulation(Value: string): boolean;
begin
  // Attention! don't use the function "ForceReceiverSimulation"
  // and assign Event "OnAnswerSimulationRequest" in the same compilation
  if fSimulationMode and
     not(Assigned(fOnAnswerSimulationRequest)) then
     begin
     fHartSimulationString:=Value;
     Result:=true;
     end
  else Result:=false;
end;

// THartSlave

{$IFDEF WithThread}
procedure THartSlave.CommandReception (RxString: string);
var NewHartCommand: THartCommand;
begin
  if EvaluateCommandMessage(RxString,fSynchronisationHeader,NewHartCommand) and
     not(ExecutionErrorDetected) and
     not(teParity in fHartCommErrors.TransferErrors) then
     begin
     fActualHartCommand.CommandNumber:=NewHartCommand.CommandNumber;
     if Assigned(fOnReadCommand) then fOnReadCommand(NewHartCommand);
     end
  else if Assigned(fOnReadCommandError) then
          begin
          fOnReadCommandError(NewHartCommand,fHartCommErrors);
          ClearErrors;  // ???
          end;
end;

procedure THartSlave.CommandError (RxString: string);
var NewHartCommand: THartCommand;
begin
  EvaluateCommandMessage(RxString,fSynchronisationHeader,NewHartCommand);
  if Assigned(fOnReadCommandError) then
     begin
     fOnReadCommandError(NewHartCommand,fHartCommErrors);
     ClearErrors;  // ???
     end;
end;
{$ENDIF}

{$IFNDEF WithThread}
function THartSlave.CheckCommandPending: boolean;
begin
  Result:=CharsAvailable(6);
end;
{$ENDIF}

constructor THartSlave.Create(AOwner: TComponent);
begin
  inherited;
  SynchronisationHeader:=CMasterHartHeader;

{$IFDEF WithThread}
  with fHartReceiver do
       begin
       // OnThreadReadChar:=nil;
       OnThreadReadString:=CommandReception;
       OnThreadReceptionTimeOut:=CommandError;
       end;
{$ENDIF}
end;

procedure THartSlave.Talk(CommStatus, DeviceStatus: byte; Data: string);
var fTransmitString: string;
begin
  ClearErrors;

  if fDeviceOpen then
     begin
     fTransmitString:=CreateSlaveMessage(fActualHartCommand.CommandNumber,CommStatus,DeviceStatus,Data);
     if not(fSimulationMode) then
        TransmitMessage(fTransmitString)
     else // SimulationMode
        SimulateTransmision(fTransmitString);
     fActualHartCommand.CommandNumber:=CInvalidCommand;
     end
  else // not(fDeviceOpen)
       fHartCommErrors.ExecutionErrors:=([eeNotOpened]);
end;

procedure THartSlave.Talk (Source: THartAnswer);
var fTransmitString: string;
begin
  ClearErrors;

  if fDeviceOpen then
     begin
     fTransmitString:=CreateSlaveMessage(fActualHartCommand.CommandNumber,Source.CommStatus,Source.DeviceStatus,Source.Data);
     if not(fSimulationMode) then
        TransmitMessage(fTransmitString)
     else // SimulationMode
        SimulateTransmision(fTransmitString);
     fActualHartCommand.CommandNumber:=CInvalidCommand;
     end
  else // not(fDeviceOpen)
       fHartCommErrors.ExecutionErrors:=([eeNotOpened]);
end;

{$IFNDEF WithThread}
function THartSlave.Listen: THartCommand;
var NewHartCommand: THartCommand;
    NewMessage: string;
begin
  Result:=CEmptyCommand;

  if not(fSimulationMode) then
     begin
     if (ReceiveMessage(NewMessage)) and
        (EvaluateCommandMessage(NewMessage,fSynchronisationHeader,NewHartCommand)) and
        not(ExecutionErrorDetected) and
        not(teParity in fHartCommErrors.TransferErrors) then
        begin
        fActualHartCommand.CommandNumber:=NewHartCommand.CommandNumber;
        Result:=NewHartCommand;
        end
     else  // the command is anyway incorrect
        begin
        if Assigned(fOnReadCommandError) then
           begin
           fOnReadCommandError(NewHartCommand,fHartCommErrors);
           ClearErrors;
           end;
        end;
     end
  else // SimulationMode
     // Note: For proper operation in the Simulation Mode, the CommandReception must be realised
     // by call of the "ForceReceiverSimulation" or by assigning of the "OnReadCommandSimulation"
     // When used with "ForceReceiverSimulation", the operation is identical as with physical device
     begin
     if Assigned(fOnReadCommandSimulation) then
        begin
        fOnReadCommandSimulation(NewHartCommand);
        // there is actually no error evaluation in demo mode
        SimulateReception(System.Length(fHartSimulationString)+6);
        fActualHartCommand.CommandNumber:=NewHartCommand.CommandNumber;
        Result:=NewHartCommand;
        end
     else // not(Assigned(fOnReadCommandSimulation))
        begin
        if (fHartSimulationString<>'') and
           (EvaluateCommandMessage(fHartSimulationString,fSynchronisationHeader,NewHartCommand)) then
           begin
           // there is actually no error evaluation in demo mode
           fActualHartCommand.CommandNumber:=NewHartCommand.CommandNumber;
           SimulateReception(System.Length(fHartSimulationString));
           fHartSimulationString:='';
           Result:=NewHartCommand;
           end;
        end;
     end;
end;
{$ENDIF}

function THartSlave.ForceReceiverSimulation (Value: string): boolean;
begin
  // Attention! don't use the function "ForceReceiverSimulation"
  // and assign Event "OnReadCommandSimulation" in the same compilation
  if fSimulationMode and
     not(Assigned(fOnReadCommandSimulation)) then
     begin
     fHartSimulationString:=Value;
     Result:=true;
     end
  else Result:=false;
end;

end.
