//******************************************************************************
//******************************************************************************
//******************************************************************************
//**
//**  PROCESSUTILS.PAS
//**
//**  Different helpful functions and procedures for process and system access
//**  Derived on 22.07.2008 from "HelpfulUtils.Pas"
//**
//**  Autor : Georg  Hahn
//**  Datum :  14.08.96 / 22.07.08 / 31.03.09 / 22.09.09 / 20.05.10 / 02.05/11
//**           28.09.13
//******************************************************************************
//******************************************************************************
//******************************************************************************
unit ProcessUtils;

interface

uses Types, SysUtils, Forms, ExtCtrls, ArraysAndGrids;

type
  TFileVersionInfos = record
    CompanyName, // Company
    FileDescription, // Description
    FileVersion, // File Version
    InternalName, // Internal Name
    LegalCopyRight, // Copyright
    LegalTradeMarks, // ...
    OriginalFileName, // Original File name
    ProductName,  // Product Name
    ProductVersion, // Product Version
    Comments: string;
  end;

  function DriveExists(DriveByte: byte): boolean; overload;
  function DriveExists(DriveLetter: char): boolean; overload;

  function IsValidFileName(FileName : TFileName) : boolean;

  function GetApplicationFullFileName: string;

  function ExtractFileBaseName(FileName: TFileName): string;
  function AppendFileBaseNameSuffix(OriginalName: TFileName; Suffix: string): TFileName;

  function StripPathString (Source: TFileName; L: integer): string;

  function GetFileSize(FileName: TFileName): longint;
  function GetFolderSize(PathName: TFileName; IncludeSubfolder: Boolean): int64;
  function GetMatchingFiles(FileName: TFileName): TStringDynArray;
  function GetFolderContents(PathName: TFileName; Attr: integer): TStringDynGrid;
  function GetAllSubfolderFiles(PathName,WildCards: TFileName; MaxSearchTime: dword): TStringDynGrid;

  function GetFolderTreeInfoString(PathName: TFileName; IncludeSubfolder: boolean; MaxSearchTime: dword): string;
  function GetFolderTreeInfoArray(PathName: TFileName; IncludeSubfolder: boolean; MaxSearchTime: dword): TStringDynGrid;

  function RunningProcessCount(ProcessName: TFileName): integer;
  function IsProcessRunning(ProcessName: TFileName): boolean;

  function CallDOSApplication(AppsName: TFileName; CommandLine: string; delay: dword; var Reply: string): boolean;

  function ExecuteApplication(AppsName, DefaultDir: TFileName; Params: string; ShowCmd: Integer): THandle;
  // procedure TerminateApplication;

  function IsWindows64: Boolean;

  function GetTotalPhysicalMemory: cardinal;
  function GetAvailablePhysicalMemory: cardinal;

  function GetWindowsFolder: string;
  function GetSystemFolder: string;
  function GetProgramFilesFolder : string;
  function GetMyDocumentsFolder : string;
  function GetPublicDocumentsFolder : string;

  function GetSystemLocaleInfoString: string;
  function GetUserLocaleInfoString: string;

  function GetEnvironmentVariableValue (VarName: string): string;
  function SetEnvironmentVariableValue (VarName,VarValue: string): integer;
  function DeleteEnvironmentVariable (VarName: string): integer;

  function GetTimeZoneOffset: TDateTime;
  function GetFileDifferenceTime(FileName: TFileName): TDateTime;

  function GetFileCreationDateTime(Source: TFileName): TDateTime;
  function GetFileLastWriteDateTime(Source: TFileName): TDateTime;
  function GetFileZoneDateTime(FileName: TFileName): TDateTime;
  function SetFileDateTime(Target: TFileName; CreateStamp, LastWriteStamp, Offset: TDateTime): boolean;

  function VerifyFileAttribute(FileName: TFileName; Attr: word): boolean;

  function GetEXEFileVersionInfos(EXEFileName: TFileName): TFileVersionInfos;

  function CopyFilesToFolder(FileName, FolderName: TFileName): boolean;
  function CopyFolderContents(Source, Destination: TFileName; Owner: TForm): boolean;

  function RandomIntegerValue (MeanValue, DeviationValue: integer): integer; overload;
  function RandomIntegerValue (MeanValue: integer; DeviationRatio: real): integer; overload;
  function RandomRealValue (MeanValue, DeviationRatio: real): real;

  function GetSystemResolutionInMilliseconds: real;

  function GetMicroTickCount: int64;

  procedure WaitMicroseconds (delay: dword);
  procedure SetDefaultProcessMessagesInterval;
  procedure SetProcessMessagesInterval (Interval: dword);
  function WaitMilliseconds(delay: dword): boolean;

  procedure LoadJPEGBitmapFile(FileName: TFileName; var Image: TImage);

  function GetClipboardStrings: TStringDynArray;
  procedure SaveStringsToClipboard(Source: TStringDynArray); overload;
  procedure SaveStringsToClipboard(Source: TStringDynGrid); overload;

const
  CHour: extended = 1/24;
  CMinute: extended = 1/(24*60);
  CSecond: extended = 1/(24*60*60);

var
  TerminateWaitMillisecondsLoop: boolean;

implementation

uses Windows, Classes, ShellAPI, SHFolder, ClipBrd, TlHelp32, Graphics, jpeg, StringUtils;

type
  EInvalidDest = class(EStreamError);
  EFCantMove = class(EStreamError);

const
  CDefaultProcessMessagesInterval: dword =  250; // default interval between ProcessMessages call in ms
  CMinimalProcessMessagesInterval: dword =   50;
  CMaximalProcessMessagesInterval: dword = 2000;

  SInvalidDest = 'Destination %s does not exist';
  SFCantMove = 'Cannot move file %s';

var
  lSystemResolutionInMilliseconds,
  lSystemCountForOneMicrosecond: real;
  lProcessMessagesInterval: dword;
  lCoarseTickCount: word;

function DriveExists(DriveByte: byte): boolean;
begin
  // 00=current, 01='A', 02='B', 04='C', 08='D' etc.
  Result:=GetLogicalDrives and (1 shl DriveByte) <> 0;
end;

function DriveExists(DriveLetter: char): boolean;
begin
  Result:=GetLogicalDrives and (1 shl 1+ord(UpCase(DriveLetter))-ord('A')) <> 0;
end;

function IsValidFileName(FileName : TFileName) : boolean;
const
  InvalidCharacters : set of char = ['\', '/', ':', '*', '?', '"', '<', '>', '|'];
var
  C : char;
begin
  Result:=(FileName<>'');

  if Result then
     begin
     for C in FileName do
         begin
         Result := not(C in InvalidCharacters);
         if not(Result) then
            break;
         end;
     end;
end;

function GetApplicationFullFileName: string;
var PFileName: array [0..260] of char;
    P: PChar;
begin
  GetModuleFileName(HInstance, PFileName, SizeOf(PFileName));
  P:=PChar(@PFileName) + lstrlen(PFileName);
  while (P^ <> '.') and (P <> @PFileName) do Dec(P);
  Result:=string(PFileName);
end;

function ExtractFileBaseName(FileName: TFileName): string;
var InterResult: string;
begin
  if (System.Pos('.',FileName)>0) then
     begin
     InterResult:=ExtractFileName(FileName);
     while (InterResult<>'') and (InterResult[System.Length(InterResult)]<>'.') do
           System.Delete(InterResult,System.Length(InterResult),1);
     if (InterResult[System.Length(InterResult)]='.') then
        System.Delete(InterResult,System.Length(InterResult),1);
     Result:=InterResult;
     end
  else Result:=FileName;
end;

function AppendFileBaseNameSuffix(OriginalName: TFileName; Suffix: string): TFileName;
begin
  Result:=ExtractFilePath(OriginalName)+ExtractFileBaseName(OriginalName)+Suffix+ExtractFileExt(OriginalName);
end;

function StripPathString (Source: TFileName; L: integer): string;
var InternalResult: string;
begin
  if (System.Length(Source) > (L-2)) then
     begin
     InternalResult:=System.Copy(Source,1,3)+'...\';
     System.Delete (Source, 1, 3);
     while ((System.Length (Source) > (L-9)) and (System.Pos('\',Source) > 0)) do
           System.Delete (Source, 1, System.Pos('\', Source));
     end
  else InternalResult:='';
  Result:=InternalResult+Source;
end;

function GetFileSize(FileName: TFileName): longint;
var SearchRec: TSearchRec;
begin
  try
    if FindFirst(ExpandFileName(FileName), faAnyFile, SearchRec) = 0 then
      Result := SearchRec.Size
    else Result := -1;
  finally
    SysUtils.FindClose(SearchRec);
  end;
end;

function GetFolderSize(PathName: TFileName; IncludeSubfolder: Boolean): int64;
var SearchRec: TSearchRec;
    NextFound: integer;
begin
  Result := 0;
  if PathName[System.Length(PathName)] <> '\' then PathName := PathName + '\';
  NextFound := SysUtils.FindFirst(PathName + '*.*', faAnyFile, SearchRec);
  while NextFound = 0 do
        begin
        System.Inc(Result, SearchRec.Size);
        if (SearchRec.Attr and faDirectory > 0) and (SearchRec.Name[1] <> '.') and (IncludeSubfolder = True) then
           System.Inc(Result, GetFolderSize(PathName + SearchRec.Name, True));
        NextFound := SysUtils.FindNext(SearchRec);
        end;
  SysUtils.FindClose(SearchRec);
end;

function GetMatchingFiles(FileName: TFileName): TStringDynArray;
var NewMatchingFiles: TStringDynArray;
    SearchRec: TSearchRec;
    NextFound: Boolean;
begin
  SetLength(NewMatchingFiles,0);

  if SysUtils.FindFirst(FileName, faAnyFile and not(faDirectory), SearchRec)=0 then
     begin
     NextFound:=true;
     while NextFound do
           begin
           if (SearchRec.Name<>'.') and (SearchRec.Name<>'..') then
              AppendElementToArray(ExtractFilePath(FileName)+SearchRec.Name,NewMatchingFiles);
           NextFound:=(SysUtils.FindNext(SearchRec)=0);
           end;
     SysUtils.FindClose(SearchRec);
     end;
  Result:=NewMatchingFiles;
end;

function GetFolderContents (PathName: TFileName; Attr: integer): TStringDynGrid;
var NewFolderContents: TStringDynGrid;
    NewItemInfo: TStringDynArray;
    SearchRec: TSearchRec;
    NextFound: Boolean;
begin
  SetLength(NewFolderContents,0);
  SetLength(NewItemInfo,5);

  if SysUtils.FindFirst(PathName, Attr, SearchRec)=0 then
     begin
     NextFound:=true;
     while NextFound do
           begin
           if (SearchRec.Name<>'.') and (SearchRec.Name<>'..') then
              begin
              NewItemInfo[0]:=SearchRec.Name;
              NewItemInfo[1]:=IntToStr(SearchRec.Size);
              NewItemInfo[2]:=DateToStr(FileDateToDateTime(SearchRec.Time));
              NewItemInfo[3]:=TimeToStr(FileDateToDateTime(SearchRec.Time));
              NewItemInfo[4]:=IntToStr(SearchRec.Attr);
              AppendArrayToGrid(NewItemInfo,NewFolderContents);
              end;
           NextFound:=(SysUtils.FindNext(SearchRec)=0);
           end;
     SysUtils.FindClose(SearchRec);
     end;
  Result:=NewFolderContents;
end;

function GetAllSubfolderFiles(PathName,WildCards: TFileName; MaxSearchTime: dword): TStringDynGrid;
var NewFolderContents: TStringDynGrid;
    NewFolderInfo, NewLineInfo, NewFileInfo: TStringDynArray;
    Command, Answer, NewLine: string;
    // Date, Time: string;
    I, L, LL, PD: integer;
const
   CDirOf = 'Directory of '; // CDirOf = 'Verzeichnis von ';
begin
  SetLength(NewFolderContents,0);
  SetLength(NewFileInfo,5);

  Command:='/c dir "'+PathName+'\'+WildCards+'" /s';
  CallDOSApplication('C:\Windows\System32\cmd.exe',Command,MaxSearchTime,Answer);   // MaxTimeSearch in ms

  if (Answer<>'') then
     begin
     NewFolderInfo:=SplitStringToStringDynArray(Answer,CLF);
     L:=System.Length(NewFolderInfo);
     if (L>0) then
        begin
        for I:=0 to L-1 do
            begin
            NewLine:=DeleteTrailingChars(NewFolderInfo[I],CCR);
            LL:=System.Length(NewLine);
            if (LL>0) and
               (System.Pos('Volume',NewLine)=0) and
               (System.Pos('    ',NewLine)<>1) and
               (System.Pos('<DIR>',NewLine)=0) then
               begin
               PD:=System.Pos(CDirOf,NewLine);
               if (PD>0) then
                  NewFileInfo[0]:=System.Copy(NewLine,PD+System.Length(CDirOf),LL)
               else
                  begin
                  NewLineInfo:=SplitStringToStringDynArray(NewLine,' ',true);
                  if (System.Length(NewLineInfo)=4) then
                     begin
                     NewFileInfo[1]:=NewLineInfo[3];
                     NewFileInfo[2]:=NewLineInfo[2];
                     NewFileInfo[3]:=NewLineInfo[0];
                     NewFileInfo[4]:=NewLineInfo[1];
                     AppendArrayToGrid(NewFileInfo,NewFolderContents);
                     end;
                  end;
               end;
            end;
        end;
     end;
  Result:=NewFolderContents;
end;

function GetFolderTreeInfoString(PathName: TFileName; IncludeSubfolder: boolean; MaxSearchTime: dword): string;
var Command,Answer: string;
begin
  Command:='"'+PathName+'" /a';
  if IncludeSubfolder then Command:=Command+' /f';

  CallDOSApplication('C:\Windows\System32\tree.com',Command,MaxSearchTime,Answer); // MaxTimeSearch in ms

  Result:=Answer;
end;

function GetTreeSubfolderDepth(Source: string): integer;
begin
  Result:=System.Pos('+',Source);
  if (Result=0) then
     Result:=System.Pos('\',Source);
end;

function GetTreeSubfolderName(Source: string): string;
var L,P: integer;
begin
  Result:='';
  L:=System.Length(Source);
  if (L>4) then
     begin
     P:=GetTreeSubfolderDepth(Source);
     if (P>0)and
        not(L<P+4) then
        Result:=System.Copy(Source,P+4,L);
     end;
end;

function GetTreeFileName(Source: string): string;
begin
  Result:=DeleteLeadingChars(Source,['|',' ']);
end;

function GetFolderTreeInfoArray(PathName: TFileName; IncludeSubfolder: boolean; MaxSearchTime: dword): TStringDynGrid;
var FolderInfoArray, PathArray, EntryDetailsArray: TStringDynArray;
    FolderDetailsInfoGrid: TStringDynGrid;
    I, J, L, LL, TSD, PathDepth: integer;
    MainPath, NextEntry: string;
begin
  FolderInfoArray:=SplitStringToStringDynArray(GetFolderTreeInfoString(PathName,IncludeSubfolder,MaxSearchTime),[#$A,#$D],true);
  SetLength(PathArray,0);
  SetLength(EntryDetailsArray,0);
  SetLength(FolderDetailsInfoGrid,0);

  L:=System.Length(FolderInfoArray);
  if (L>2) then
     begin
     AppendElementToArray(FolderInfoArray[2],EntryDetailsArray);
     AppendArrayToGrid(EntryDetailsArray,FolderDetailsInfoGrid);

     PathDepth:=0;
     for I:=3 to L-1 do
         begin
         TSD:=GetTreeSubfolderDepth(FolderInfoArray[I]);
         if (TSD>0) then // Subfolder Information detected
            begin
            NextEntry:=GetTreeSubfolderName(FolderInfoArray[I]);
            PathDepth:=(TSD DIV 4);
            LL:=System.Length(PathArray);
            if (LL>PathDepth+1) then
               SetLength(PathArray,PathDepth+1);
            if (LL<PathDepth+1) then
               AppendElementToArray(NextEntry,PathArray)
            else
               PathArray[PathDepth]:=NextEntry;
            end
         else // possibly File Name (or empty line)
            begin
            NextEntry:=GetTreeFileName(FolderInfoArray[I]);
            if (NextEntry<>'') then
               begin
               SetLength(EntryDetailsArray,0);
               AppendElementToArray(NextEntry,EntryDetailsArray);
               NextEntry:='\';
               for J:=0 to System.Length(PathArray)-1 do
                   NextEntry:=NextEntry+PathArray[J]+'\';

               AppendElementToArray(NextEntry,EntryDetailsArray);
               AppendArrayToGrid(EntryDetailsArray,FolderDetailsInfoGrid);
               end;
            end;
         end;
     end;
  Result:=FolderDetailsInfoGrid;
end;

function RunningProcessCount(ProcessName: TFileName): integer;
var aProcessEntry32: TProcessEntry32;
    aSnapshotHandle: THandle;
    PC: integer;
begin
  PC:=0;

  aSnapshotHandle := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if (aSnapshotHandle<>INVALID_HANDLE_VALUE) then
     begin
     aProcessEntry32.dwSize := Sizeof(aProcessEntry32);
     if (Process32First(aSnapshotHandle, aProcessEntry32)) then
        begin
        if System.Pos(AnsiLowerCase(aProcessEntry32.szExeFile), AnsiLowerCase(ExtractFileName(ProcessName))) > 0 then System.Inc(PC);
        while Process32Next(aSnapshotHandle, aProcessEntry32) do
              if System.Pos(AnsiLowerCase(aProcessEntry32.szExeFile), AnsiLowerCase(ExtractFileName(ProcessName))) > 0 then System.Inc(PC);
        end;
     CloseHandle(aSnapshotHandle);
     end;

  Result:=PC;
end;

function IsProcessRunning(ProcessName: TFileName): boolean;
var aProcessEntry32: TProcessEntry32;
    aSnapshotHandle: THandle;
    PC: integer;
begin
  Result:=false; PC:=0;

  aSnapshotHandle := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if (aSnapshotHandle<>INVALID_HANDLE_VALUE) then
     begin
     aProcessEntry32.dwSize := Sizeof(aProcessEntry32);
     if (Process32First(aSnapshotHandle, aProcessEntry32)) then
        begin
        if System.Pos(AnsiLowerCase(aProcessEntry32.szExeFile), AnsiLowerCase(ExtractFileName(ProcessName))) > 0 then System.Inc(PC);
        while Process32Next(aSnapshotHandle, aProcessEntry32) do
              begin
              if System.Pos(AnsiLowerCase(aProcessEntry32.szExeFile), AnsiLowerCase(ExtractFileName(ProcessName))) > 0 then System.Inc(PC);
              if (PC>1) then
                 begin
                 Result:=true;
                 break;
                 end;
              end;
        end;
     CloseHandle(aSnapshotHandle);
     end;
end;

function CallDOSApplication(AppsName: TFileName; CommandLine: string; delay: dword; var Reply: string): boolean;
const
  ReadBuffer = 2000000; // 2MB
var
  Security: TSecurityAttributes;
  CTOuts: TCommTimeOuts;

  OutReadPipe, OutWritePipe :THandle;
  InReadPipe, InWritePipe :THandle;
  ErrorReadPipe,ErrorWritePipe :THandle;
  Buffer, ErrBuffer :PChar;

  BytesRead, ErrBytesRead, BufferBegin :dword;
  BytesToRead, ProcessStatus :dword;
  Wait_ms :integer;

  StartUpInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  CmdLine: string;

  PipeEmpty, ErrPipeEmpty, ReadPipeOK, ReadErrPipeOK: boolean;

begin
  with Security do
       begin
       nLength:=SizeOf(TSecurityAttributes);
       lpSecurityDescriptor:=nil;
       bInheritHandle:=true;
       end;

  with CTOuts do
       begin
       ReadIntervalTimeout :=0;
       ReadTotalTimeoutMultiplier :=0;
       ReadTotalTimeoutConstant :=0;
       WriteTotalTimeoutMultiplier :=0;
       WriteTotalTimeoutConstant :=0;
       end;

  CreatePipe(InReadPipe, InWritePipe, @Security,0);
  CreatePipe(OutReadPipe, OutWritePipe, @Security, ReadBuffer + 1);

  SetCommTimeouts(InReadPipe,CTOuts);
  SetCommTimeouts(InWritePipe,CTOuts);
  SetCommTimeouts(OutReadPipe,CTOuts);
  SetCommTimeouts(OutWritePipe,CTOuts);
  SetCommTimeouts(ErrorReadPipe,CTOuts);
  SetCommTimeouts(ErrorWritePipe,CTOuts);

  FillChar(StartUpInfo, SizeOf(StartUpInfo), #0);
  with StartUpInfo do
       begin
       cb := SizeOf(StartUpInfo);
       dwFlags := STARTF_USESTDHANDLES or STARTF_USESHOWWINDOW or STARTF_FORCEONFEEDBACK;
       wShowWindow := SW_HIDE;
       hStdInput := InReadPipe;
       hStdOutput := OutWritePipe;
       hStdError := ErrorWritePipe;
       end;

  Buffer := AllocMem(ReadBuffer + 1);
  ErrBuffer := AllocMem(ReadBuffer + 1);

  // Enclose filename in quotes to take care of long filenames with spaces.
  CmdLine := '"' + AppsName + '" ' + CommandLine;

  Result := CreateProcess(nil, PChar(CmdLine), @Security, nil, true,
                          // CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
                          IDLE_PRIORITY_CLASS,
                          NIL, PChar(ExtractFilePath(AppsName)),
                          StartUpInfo, ProcessInfo);

  PipeEmpty:=true;
  ErrPipeEmpty :=True;

  // Wait for it to finish.
  if Result then
     begin
     bufferbegin := 0;
     BytesRead :=0;
     ErrBytesRead :=0;
     BytesToRead:=ReadBuffer;
     Wait_ms:=delay;

     repeat
       if (((BytesRead=BytesToRead) or (ErrBytesRead=ReadBuffer)) and (Wait_ms>1)) then Wait_ms:=Wait_ms div 2;
       BytesRead :=0;
       ErrBytesRead :=0;
       ProcessStatus :=0;
       BytesToRead :=ReadBuffer - bufferbegin;

       WaitForSingleObject(ProcessInfo.hProcess, Wait_ms);

       ReadPipeOK:=PeekNamedPipe(OutReadPipe,nil,ReadBuffer,nil,@BytesRead,nil);
       if (ReadPipeOK and (BytesRead>0)) then
          ReadPipeOK := ReadFile(OutReadPipe, Buffer[bufferbegin],BytesToRead, BytesRead, nil)
       else BytesRead:=0;

       ReadErrPipeOK:=PeekNamedPipe(ErrorReadPipe,nil,ReadBuffer,nil,@ErrBytesRead,nil);
       if (ReadErrPipeOK and (ErrBytesRead>0)) then
          ReadErrPipeOK:= ReadFile(ErrorReadPipe, ErrBuffer[0],ReadBuffer, ErrBytesRead, nil)
       else ErrBytesRead:=0;

       GetExitCodeProcess(ProcessInfo.hProcess,ProcessStatus);
     until (ProcessStatus <> STILL_ACTIVE) and PipeEmpty and ErrPipeEmpty;

    Reply:=String(Buffer);
    end;

  FreeMem(Buffer);
  CloseHandle(InReadPipe);
  CloseHandle(InWritePipe);
  CloseHandle(OutReadPipe);
  CloseHandle(OutWritePipe);
end;

function ExecuteApplication(AppsName, DefaultDir: TFileName; Params: string; ShowCmd: Integer): THandle;
var
  zFileName, zParams, zDir: array[0..79] of Char;
begin
    Result := ShellExecute(Application.MainForm.Handle, nil,  // open,
    StrPCopy(zFileName, AppsName), StrPCopy(zParams, Params),
    StrPCopy(zDir, DefaultDir), ShowCmd);
end;

// procedure TerminateApplication;

function IsWindows64: Boolean;
type
  TIsWow64Process = function(AHandle:THandle; var AIsWow64: BOOL): BOOL; stdcall;
var
  vKernel32Handle: DWORD;
  vIsWow64Process: TIsWow64Process;
  vIsWow64       : BOOL;
begin
  // 1) assume that we are not running under Windows 64 bit
  Result := False;

  // 2) Load kernel32.dll library
  vKernel32Handle := LoadLibrary('kernel32.dll');
  if (vKernel32Handle = 0) then Exit; // Loading kernel32.dll was failed, just return

  try

    // 3) Load windows api IsWow64Process
    @vIsWow64Process := GetProcAddress(vKernel32Handle, 'IsWow64Process');
    if not Assigned(vIsWow64Process) then Exit; // Loading IsWow64Process was failed, just return

    // 4) Execute IsWow64Process against our own process
    vIsWow64 := False;
    if (vIsWow64Process(GetCurrentProcess, vIsWow64)) then
      Result := vIsWow64;   // use the returned value

  finally
    FreeLibrary(vKernel32Handle);  // unload the library
  end;
end;

function GetTotalPhysicalMemory: cardinal;
var MS: TMemoryStatus;
begin
  MS.dwLength := SizeOf(TMemoryStatus);
  GlobalMemoryStatus(MS);
  Result:=MS.dwTotalPhys;
end;

function GetAvailablePhysicalMemory: cardinal;
var MS: TMemoryStatus;
begin
  MS.dwLength := SizeOf(TMemoryStatus);
  GlobalMemoryStatus(MS);
  Result:=MS.dwAvailPhys;
end;

function GetSpecialFolderInfo(Selector: integer) : string;
const
   SHGFP_TYPE_CURRENT = 0;
   SHGFP_TYPE_DEFAULT = 1;
var FolderString: array [0..MaxChar] of char;
begin
  SHGetFolderPath(0,Selector,0,SHGFP_TYPE_CURRENT,@FolderString[0]) ;
  Result:= FolderString;
end;

function GetWindowsFolder: string;
begin
  Result:=GetSpecialFolderInfo(CSIDL_WINDOWS);
end;

function GetSystemFolder: string;
begin
  Result:=GetSpecialFolderInfo(CSIDL_SYSTEM);
end;

function GetProgramFilesFolder : string;
begin
  Result:=GetSpecialFolderInfo(CSIDL_PROGRAM_FILES);
end;

function GetMyDocumentsFolder : string;
begin
  Result:=GetSpecialFolderInfo(CSIDL_PERSONAL);
end;

function GetPublicDocumentsFolder : string;
begin
  Result:=GetSpecialFolderInfo(CSIDL_COMMON_DOCUMENTS);
end;

function GetSystemLocaleInfoString: string;
var BufSize: Integer;
begin
  BufSize := GetLocaleInfo(GetSystemDefaultLCID, LOCALE_SLANGUAGE, nil, 0);
  SetLength(Result, BufSize);
  GetLocaleinfo(GetUserDefaultLCID, LOCALE_SLANGUAGE, PChar(Result), BufSize);
  SetLength(Result, BufSize - 1);
end;

function GetUserLocaleInfoString: string;
var BufSize: Integer;
begin
  BufSize := GetLocaleInfo(GetUserDefaultLCID, LOCALE_SLANGUAGE, nil, 0);
  SetLength(Result, BufSize);
  GetLocaleinfo(GetUserDefaultLCID, LOCALE_SLANGUAGE, PChar(Result), BufSize);
  SetLength(Result, BufSize - 1);
end;

// The three followed function comes from www.delphiDabbler.com
function GetEnvironmentVariableValue (VarName: string): string;
var BufSize: Integer;  // buffer size required for value
begin
  // Get required buffer size (inc. terminal #0)
  BufSize := GetEnvironmentVariable(PChar(VarName), nil, 0);
  if BufSize > 0 then
     begin
     // Read environment variable value into result string
     SetLength(Result, BufSize - 1);
     GetEnvironmentVariable(PChar(VarName),PChar(Result), BufSize);
     end
  else // No such environment variable
     Result := '';
end;

function SetEnvironmentVariableValue (VarName,VarValue: string): integer;
begin
  // Simply call API function
  if SetEnvironmentVariable(PChar(VarName),PChar(VarValue)) then
    Result := 0
  else Result := GetLastError;
end;

function DeleteEnvironmentVariable (VarName: string): integer;
begin
  if SetEnvironmentVariable(PChar(VarName), nil) then
    Result := 0
  else Result := GetLastError;
end;

function GetTimeZoneOffset: TDateTime;
var TZ: TTimeZoneInformation;
begin
  if (GetTimeZoneInformation(TZ)<>TIME_ZONE_ID_UNKNOWN)
     then Result:=TZ.Bias/(24*60)
  else Result:=0.0;
end;

function GetFileDifferenceTime(FileName: TFileName): TDateTime;
var FileHandle,FD: integer;
    CT,LAT,LWT: FILETIME;
    ST: SYSTEMTIME;
begin
  Result:=0.0;
  FileHandle:=FileOpen(FileName, fmOpenRead or fmShareDenyNone);
  if (FileHandle>0) then
     begin
     FD:=FileGetDate(FileHandle);
     if (FD<>-1) and
        (GetFileTime(FileHandle,@CT,@LAT,@LWT)) and
        (FileTimeToSystemTime(LWT,ST)) then
        Result:=FileDateToDateTime(FD)-SystemTimeToDateTime(ST);
     end;
end;

function GetFileCreationDateTime(Source: TFileName): TDateTime;
var CT,LAT,LWT: FILETIME;
    FileHandle: integer;
    ST: SYSTEMTIME;
begin
  // The result is always absolute dater and time UCT + 0:00
  // independent from time zone and summer/winter time
  Result:=-1.0;
  FileHandle:=FileOpen(Source, fmOpenRead or fmShareDenyNone);
  if (FileHandle > 0) and
     (GetFileTime(FileHandle,@CT,@LAT,@LWT)) and
     (FileTimeToSystemTime(CT,ST)) then
      Result:=SystemTimeToDateTime(ST);
  FileClose(FileHandle);
end;

function GetFileLastWriteDateTime(Source: TFileName): TDateTime;
var CT,LAT,LWT: FILETIME;
    FileHandle: integer;
    ST: SYSTEMTIME;
begin
  // The result is always absolute dater and time UCT + 0:00
  // independent from time zone and summer/winter time
  Result:=-1.0;
  FileHandle:=FileOpen(Source, fmOpenRead or fmShareDenyNone);
  if (FileHandle > 0) and
     (GetFileTime(FileHandle,@CT,@LAT,@LWT)) and
     (FileTimeToSystemTime(LWT,ST)) then
      Result:=SystemTimeToDateTime(ST);
  FileClose(FileHandle);
end;

function GetFileZoneDateTime(FileName: TFileName): TDateTime;
begin
  // The result is actual time with considered zone and summer/winter
  Result := FileDateToDateTime(FileAge(FileName));
end;

function SetFileDateTime(Target: TFileName; CreateStamp, LastWriteStamp, Offset: TDateTime): boolean;
var CT,LAT,LWT: FILETIME;
    FileHandle: integer;
    ST: SYSTEMTIME;
begin
  Result:=false;

  DateTimeToSystemTime(CreateStamp,ST);
  SystemTimeToFileTime(ST,CT);
  DateTimeToSystemTime(LastWriteStamp+Offset,ST);
  SystemTimeToFileTime(ST,LWT);
  LAT:=LWT;

  FileHandle:=FileOpen(Target, fmOpenWrite or fmShareDenyNone);
  if (FileHandle > 0) and
     (SetFileTime(FileHandle,@CT,@LAT,@LWT)) then
     Result:=(FileSetDate(FileHandle,DateTimeToFileDate(LastWriteStamp+Offset))=0);
  FileClose(FileHandle);
end;

function VerifyFileAttribute(FileName: TFileName; Attr: word): boolean;
var FileAttr: Integer;
begin
  FileAttr:=FileGetAttr(FileName);
  if (FileAttr = -1) then Result:=false
  else
     Result:=((FileAttr and Attr) = Attr);
end;

// This function was written using example from http://www.gtdelphicomponents.gr
function GetEXEFileVersionInfos(EXEFileName: TFileName): TFileVersionInfos;
var NewFileVersionInfos: TFileVersionInfos;
    VerHandle, VerBuffLen: Cardinal;
    VerInfoSize: dword;
    VerBufValue: pointer;
    VerBuff: PChar;
    VerKey: string;
  function GetInfo(const aKey: string): string;
  begin
    Result := '';
    VerKey := Format('\StringFileInfo\%.4x%.4x\%s',
                      [LoWord(Integer(VerBufValue^)),HiWord(Integer(VerBufValue^)), aKey]);
    if VerQueryValue(VerBuff, PChar(VerKey), VerBufValue, VerBuffLen) then
       Result := StrPas(VerBufValue);
  end;
  function QueryValue(S: string): string;
  begin
    Result := '';
    if GetFileVersionInfo(PChar(EXEFileName), VerHandle, VerInfoSize, VerBuff) and
        VerQueryValue(VerBuff, '\VarFileInfo\Translation', VerBufValue, VerBuffLen) then
     Result := GetInfo(S);
  end;
begin
  with NewFileVersionInfos do
       begin
       CompanyName:='';
       FileDescription:='';
       FileVersion:='';
       InternalName:='';
       LegalCopyRight:='';
       LegalTradeMarks:='';
       OriginalFileName:='';
       ProductName:='';
       ProductVersion:='';
       Comments:='';
       end;

  if FileExists(EXEFileName) and
    ((LowerCase(ExtractFileExt(EXEFileName))='.exe') or
     (LowerCase(ExtractFileExt(EXEFileName))='.dll')) then
     begin
     VerInfoSize:=GetFileVersionInfoSize(PChar(EXEFileName),VerHandle);
     if (VerInfoSize>0) then
        begin
          try
            VerBuff:=AllocMem(VerInfoSize);
            with NewFileVersionInfos do
                 begin
                 CompanyName:=QueryValue('CompanyName');
                 FileDescription:=QueryValue('FileDescription');
                 FileVersion:=QueryValue('FileVersion');
                 InternalName:=QueryValue('InternalName');
                 LegalCopyRight:=QueryValue('LegalCopyRight');
                 LegalTradeMarks:=QueryValue('LegalTradeMarks');
                 OriginalFileName:=QueryValue('OriginalFileName');
                 ProductName:=QueryValue('ProductName');
                 ProductVersion:=QueryValue('ProductVersion');
                 Comments:=QueryValue('Comments');
                 end;
          finally
            FreeMem(VerBuff);
          end;
        end;
     end;

  Result:=NewFileVersionInfos;
end;

function CopyFilesToFolder(FileName, FolderName: TFileName): boolean;
const
  CMaxStringBufferLength   = 100;
var fBuffer: array [0..CMaxStringBufferLength] of char;
    SourceStream, TargetStream: TFileStream;
    SourceFiles: TStringDynArray;
    Destination: TFileName;
    I, L, M, N: integer;
begin
  if (System.Pos('?',FileName)>0) or
     (System.Pos('*',FileName)>0) then
     SourceFiles:=GetMatchingFiles(FileName)
  else
     begin
     SetLength(SourceFiles,0);
     AppendElementToArray(FileName,SourceFiles);
     end;
  L:=System.Length(SourceFiles);

  if (L>0) and DirectoryExists(FolderName) then
     begin
     M:=0;
     for I:=0 to L-1 do
         begin
         if FileExists(SourceFiles[I]) then
            begin

            if (FolderName[System.Length(FolderName)]='\') then
               Destination:=FolderName+ExtractFileName(SourceFiles[I])
            else Destination:=FolderName+'\'+ExtractFileName(SourceFiles[I]);

            try
              SourceStream:=TFileStream.Create(SourceFiles[I],fmOpenRead);
              try
                TargetStream:=TFileStream.Create(Destination,fmCreate);
                repeat
                  N:=SourceStream.Read(fBuffer,CMaxStringBufferLength);
                  TargetStream.Write(fBuffer,N);
                until (SourceStream.Position=SourceStream.Size);
                System.Inc(M);
              finally
                TargetStream.Free;
              end;

            finally
              SourceStream.Free;
            end;

            end;
         end;
     Result:=(M=L);
     end
  else Result:=false;
end;

function CopyFolderContents(Source, Destination: TFileName; Owner: TForm): boolean;
var FO :TSHFileOpStruct;
begin
   with FO do
        begin
        Wnd    := Owner.Handle;
        wFunc  := FO_COPY;
        pFrom  := PChar(Source);
        pTo    := PChar(Destination);
        fFlags := FOF_NoConfirmMkDir;
        end;
   SHFileOperation(FO);
   Result:=true;
end;


function RandomIntegerValue (MeanValue, DeviationValue: integer): integer;
begin
  Result := MeanValue+System.Random(2*DeviationValue)-DeviationValue;
end;

function RandomIntegerValue (MeanValue: integer; DeviationRatio: real): integer;
begin
  Result := MeanValue+System.Trunc(MeanValue*DeviationRatio*(2.0*System.Random-1.0));
end;

function RandomRealValue (MeanValue, DeviationRatio: real): real;
begin
  Result := MeanValue*(1.0+DeviationRatio*(2.0*System.Random-1.0));
end;

// Timer utilities

function GetSystemResolutionInMilliseconds: real;
var TA,TI: dword;
    TAD: longbool;
begin
  // Note: if the standard (low performance) timer is available then evaluate
  // and return the system resolution result in milliseconds (as real value),
  // otherwise return real zero (0.0) as result.
  // Please note that the TA and TI value are given in 100ns steps
 if GetSystemTimeAdjustment(TA,TI,TAD) then
    Result:=TI/10000.0
 else Result:=0.0;
end;

function GetSystemCountForOneMicrosecond: real;
var fCounterFrequency, fProcessInitialisationTime: int64;
begin
  // Note: if high performance timer is available on the host system
  // then evaluate the count number per microsecond (as real value).
  // This number value will be then used to create required delay loops
  // by the functions "WaitMicrosecond" and "WaitMilliseconds".
  if QueryPerformanceFrequency(fCounterFrequency) and
     QueryPerformanceCounter(fProcessInitialisationTime) then
     Result:=fCounterFrequency/1000000.0
 else Result:=0.0;
end;

function GetMicroTickCount: int64;
var fActualCounterValue: int64;
begin
  if QueryPerformanceCounter(fActualCounterValue) then
     Result:=fActualCounterValue
 else Result:=0;
end;

function GetInternCoarseTickCount: word;
var CoarseTickCount,DummyTickCounter: word;
    MeasureStart,MeasureEnd: dword;
const
    NTBigLoop    = 65000;
    NTSmallLoop  = 50;
    NEmpiric     = 0.022;
begin
  MeasureStart:=GetTickCount;
  for CoarseTickCount:=0 to NTBigLoop do
      begin
      for DummyTickCounter:=0 to NTSmallLoop do;
      end;
  MeasureEnd:=GetTickCount-MeasureStart;
  if MeasureEnd<1 then MeasureEnd:=1;
  Result:=System.Round(NEmpiric*NTBigLoop*NTSmallLoop/MeasureEnd);
end;

procedure WaitMicroseconds (delay: dword);
var StartTime,FinishTime: int64;
    TensMicros,I,K: dword;
begin
  // Warning! don't use this procedure for very long values (typ. max. 100000), becaue it will
  // inevitable "freeze" the system and prevent any other control operations.
  // For long delay loops use rather the "WaitMilliseconds" function.

  // If the high performance timer is available on this system then the QueryPerformanceCounter
  // function will be used for time measurement. Otherwise the chunky approach with estimated
  // lCoarseTickCount value will be used.
  // The procedure will break without delay, when no reliable time estimation is possible.
  if (delay>0) then
     if (lSystemCountForOneMicrosecond>0.001) then
        begin
        if QueryPerformanceCounter(StartTime) then
           begin
           FinishTime:=StartTime+System.Round(delay*lSystemCountForOneMicrosecond);
           while QueryPerformanceCounter(StartTime) and (FinishTime>StartTime) do;
           end;
        end
     else // if no high performance timer available on this systen
        begin
        if (lCoarseTickCount>0) then
           begin
           if delay>10 then TensMicros:=delay DIV 10 else TensMicros:=1;
           for I:=0 to TensMicros do
               begin
               for K:=0 to lCoarseTickCount do;
               end;
           end;
        end;
end;

procedure SetDefaultProcessMessagesInterval;
begin
  lProcessMessagesInterval:=CDefaultProcessMessagesInterval;
end;

procedure SetProcessMessagesInterval (Interval: dword);
begin
  if (Interval<>lProcessMessagesInterval) and
     not((Interval<CMinimalProcessMessagesInterval) or (Interval>CMaximalProcessMessagesInterval)) then
     lProcessMessagesInterval:=Interval;
end;

function WaitMilliseconds(delay: dword): boolean;
var StartTime, NextTime, FinishTime: int64;
begin
  // Attention! don't use this function for very small values (typ. min. 50), because it may
  // cause very imprecise time measurement, dependent from the host system properties.
  // For short and precise delay loops use rather the "WaitMicroseconds" procedure.

  // If the high performance timer is available on this system then the QueryPerformanceCounter
  // function will be used for time measurement. Otherwise the worse resolution standard timer
  // will be used (the resolution on the development system was 15.625ms).
  // The function will break without delay, when no reliable time estimation is possible.
  // The function result is true only if the required delay elapsed successful.
  TerminateWaitMillisecondsLoop:=false;
  if (delay>0) then // wait until delay elapsed
     if (lSystemCountForOneMicrosecond>0.001) then
        begin
        if QueryPerformanceCounter(StartTime) then
           begin
           FinishTime:=StartTime+System.Round(delay*1000*lSystemCountForOneMicrosecond);
           NextTime:=StartTime+lProcessMessagesInterval*1000;
           while not(TerminateWaitMillisecondsLoop) and QueryPerformanceCounter(StartTime) and (FinishTime>StartTime) do
                 if (StartTime>NextTime) then
                    begin
                    NextTime:=NextTime+lProcessMessagesInterval*1000;
                    Application.ProcessMessages;
                    end;
           end;
        end
     else // if no high performance timer available on this system
        begin
        if (lSystemResolutionInMilliseconds>0.001) and
           (delay>System.Round(lSystemResolutionInMilliseconds)) then
           begin
           StartTime:=GetTickCount; FinishTime:=StartTime+delay;
           NextTime:=StartTime+lProcessMessagesInterval;
           while not(TerminateWaitMillisecondsLoop) and (FinishTime>GetTickCount) do
                 if (GetTickCount>NextTime) then
                    begin
                    NextTime:=NextTime+lProcessMessagesInterval;
                    Application.ProcessMessages;
                    end;
           end;
        end;
  Result:=not(TerminateWaitMillisecondsLoop);
end;

procedure LoadJPEGBitmapFile(FileName: TFileName; var Image: TImage);
begin
  if (FileExists(FileName)) then
     begin
     try Image.Picture.LoadFromFile(Filename);
     except
       on EInvalidGraphic do Image.Picture.Graphic := nil;
       end;
     end
  else Image.Picture.Graphic := nil;
end;

function GetClipboardStrings: TStringDynArray;
var ClipboardData: array [0..1023] of char;
    NewClipboardStrings: TStringDynArray;
    I: integer;
begin
  ClearDynArray(NewClipboardStrings);
  if Clipboard.HasFormat(CF_TEXT) then
     begin
     I:=Clipboard.GetTextBuf(ClipboardData,1023);
     if (I>0) and (I<1023) then
        NewClipboardStrings:=SplitStringToStringDynArray(CopyPChar(ClipboardData),[CLF,CCR,CTAB],true);
     end;
  Result:=NewClipboardStrings;
end;

procedure SaveStringsToClipboard(Source: TStringDynArray);
var I, L: integer;
    S: string;
begin
  L:=System.Length(Source);
  if (L>0) then
     begin
     S:='';
     for I:=0 to L-1 do
         S:=S+Source[I]+CCR+CLF;
     if  (System.Length(S)>0) then
         begin
         Clipboard.Clear;
         Clipboard.SetTextBuf(PChar(S));
         end;
     end;
end;

procedure SaveStringsToClipboard(Source: TStringDynGrid);
var I, J, K, L: integer;
    S: string;
begin
  L:=System.Length(Source);
  if (L>0) then
     begin
     S:='';
     for I:=0 to L-1 do
         begin
         K:=System.Length(Source[I]);
         if (K>0) then
            for J:=0 to K-1 do
                S:=S+Source[I,J]+CTAB;
         S:=S+CCR+CLF;
         end;
     if  (System.Length(S)>0) then
         begin
         Clipboard.Clear;
         Clipboard.SetTextBuf(PChar(S));
         end;
     end;
end;

initialization
  System.Randomize;

  lCoarseTickCount:=GetInternCoarseTickCount;
  lSystemCountForOneMicrosecond:=GetSystemCountForOneMicrosecond;
  lSystemResolutionInMilliseconds:=GetSystemResolutionInMilliseconds;

  SetDefaultProcessMessagesInterval;
end.




