//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//**
//**  SERIALCOMDEVICES.PAS // INCLUDES ALSO PREVIOUS FILE  //
//**
//**  Helpful constants, functions and components for COM access
//**  Derived from  "SERIALCOMUTILS.PAS"
//**
//**  Autor : Georg  Hahn
//**  Datum :  13.04.02 / 20.01.03 / 17.02.05 / 12.04.06 / 23.10.08 / 02.12.09
//**           13.01.10 / 08.08.12 / 19.12.12
//**
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
 unit SerialComDevices;
{$M+}
{$UNSAFECODE ON}

// Be careful using compilation directive "WithTread". The component behaviour
// will be in fact "smarter", but thread may operate incorrect on some machines
// and in such cases the serial communication will fail at all.

{$DEFINE WithThread}

interface

uses
  Windows, Classes, StringUtils, UsefulObjects;

const
  dcb_Binary                = $00000001;
  dcb_Parity                = $00000002;
  dcb_OutxCtsFlow           = $00000004;
  dcb_OutxDsrFlow           = $00000008;
  dcb_DtrControlEnable      = $00000010;
  dcb_DtrControlHandshake   = $00000020;
  dcb_DsrSensitivity        = $00000040;
  dcb_TXContinueOnXoff      = $00000080;
  dcb_OutX                  = $00000100;
  dcb_InX                   = $00000200;
  dcb_ErrorChar             = $00000400;
  dcb_Null                  = $00000800;
  dcb_RtsControlEnable      = $00001000;
  dcb_RtsControlHandshake   = $00002000;
  dcb_RtsControlToggle      = $00003000;
  dcb_AbortOnError          = $00004000;

  CPurgeRead        = PURGE_RXABORT + PURGE_RXCLEAR;
  CPurgeWrite       = PURGE_TXABORT + PURGE_TXCLEAR;
  CPurgeReadWrite   = CPurgeRead + CPurgeWrite;

  CDefaultComBaudRate     = CBR_9600;
  CDefaultReadTimeOut     = 1000;
  CDefaultMaxStringLength =  100;

type

  TComPresenceInfo = record
    Exists,
    Available: boolean;
    end;

  TComPresenceInfoArray = array [1..99] of TComPresenceInfo;

  TFComSettings = record
    Number: integer;
    BaudRate: dword;
    Flags: LongInt;
    ByteSize,
    StopBits,
    Parity: byte;
    EOSChar,
    XonChar,
    XoffChar: char;
    end;

  TOutputControlMode = (ocNone, ocCtsOnly, ocDsrOnly, ocCtsDsr);
  TXonXoffControlMode = (xcNone, xcInputOnly, xcOutputOnly, xcInputOutput);

  TModemStatus = record
    CTSLine,
    DSRLine: boolean;
    end;

  // function GetStringParityByte (Source: string): byte;

  function GetNextComBaudRateValue(Value: dword; Ascending: boolean): dword;
  function GetComCharTransmissionTime(ComSettings: TFComSettings): word;

  function GetMaxSerialPortNumber: dword;
  function SerialPortExists(ComNumber: word): boolean;
  function SerialPortAvailable(ComNumber: word): boolean;
  function GetSerialPortPresenceInfo(MaxComNumber: word): TComPresenceInfoArray;
  function ModemOnSerialPort(ComNumber: word): boolean;

  function OpenSerialPort(ComSettings: TFComSettings): THandle;
  function SerialPortOpened(fComHandle: THandle): boolean;
  procedure CloseSerialPort(var fComHandle: THandle);

  function SetTXDLine(fComHandle: THandle; TXD: boolean): boolean;
  function SetDTRLine(fComHandle: THandle; DTR: boolean): boolean;
  function SetRTSLine(fComHandle: THandle; RTS: boolean): boolean;
  function SetDTRandRTSlines(fComHandle: THandle; DTR,RTS: boolean): boolean;
  function GetCTSLine(fComHandle: THandle): boolean;
  function GetDSRLine(fComHandle: THandle): boolean;
  function GetCTSandDSRLines(fComHandle: THandle): TModemStatus;

  procedure PurgeCOMReadBuffer(fComHandle: THandle);
  procedure PurgeCOMWriteBuffer(fComHandle: THandle);
  procedure PurgeCOMReadWriteBuffer(fComHandle: THandle);

  function GetCOMCharsInQue(fComHandle: THandle): dword;
  function GetCOMCharsOutQue(fComHandle: THandle): dword;

  function SetCOMMonitoredEvents(fComHandle: THandle; EvtMask: dword): boolean;
  function GetCOMMonitoredEventsStatus(fComHandle: THandle; EvtMask: dword): boolean;

  function PutStringToCOM(fComHandle: THandle; Str: string): boolean; overload;
  function PutStringToCOM(fComHandle: THandle; Str: string; TOC: TTimeOutController; TimeOut: dword): boolean; overload;

  function GetCharFromCOM(fComHandle: THandle; var Ch: char): boolean; overload;
  function GetCharFromCOM(fComHandle: THandle; var Ch: char; TOC: TTimeOutController; TimeOut: dword): boolean; overload;

  function GetStringFromCOM(fComHandle: THandle; CharCount: byte; BreakReadOnEOS: boolean; EOSChar: char): string; overload;
  function GetStringFromCOM(fComHandle: THandle; CharCount: byte; BreakReadOnEOS: boolean; EOSChar: char; TOC: TTimeOutController; TimeOut: dword): string; overload;

type

  TSimpleEvent  = procedure of object;

  TRxCharRead  = procedure(RxChar: char) of object;
  TRxStringRead = procedure(RxString: string) of object;
  TRxStringRequest = procedure(var RxString: string) of object;
  // TRxReadReview = procedure(RxString: string; var Terminate: boolean) of object;

  TRxReadTermination = procedure(RxString: string; Duration: word; var Terminate: boolean) of object;


{$IFDEF WithThread}
  PComReceiver = ^TComReceiver;
  TComReceiver = class(TThread)
  private
    { Private declarations }
    pComHandle: PHandle;
    pBreakReadOnEOS,
    pRemoveEOSOnRead,
    pSimulationMode: pboolean;
    pEOSchar: pchar;
    pReadTimeOut,
    pMaxStringLength: pword;
    pReceptionString,
    pSimulationString: pstring;

    fEnabled,
    fDisabled,
    fReceptionFinished,
    fReceptionTimeOut: boolean;
    fReceptionBegin,
    fReceptionTimeOutLimit: dword;
    fThreadReceptionString: string;

    fOnThreadReceiveChar,
    fOnThreadReceptionFinished: TSimpleEvent;

    fOnThreadReadChar: TRxCharRead;
    fOnThreadReadString,
    fOnThreadReceptionTimeOut: TRxStringRead;
    // fOnThreadReceptionReview: TRxReadReview;
    fOnThreadReadTermination: TRxReadTermination;
  protected
    { Protected-Deklarationen }
    function GetSimulatedChar(var Ch: char): boolean;
    procedure Execute; override;
  public
    { Public-Deklarationen}
    constructor Create;
    destructor Destroy; override;
    procedure Resume;
    procedure Suspend;
  published
    { Published-Deklarationen }
    property OnThreadReceiveChar: TSimpleEvent read fOnThreadReceiveChar write fOnThreadReceiveChar;
    property OnThreadReceptionFinished: TSimpleEvent read fOnThreadReceptionFinished write fOnThreadReceptionFinished;

    property OnThreadReadChar: TRxCharRead read fOnThreadReadChar write fOnThreadReadChar;
    property OnThreadReadString: TRxStringRead read fOnThreadReadString write fOnThreadReadString;
    property OnThreadReceptionTimeOut: TRxStringRead read fOnThreadReceptionTimeOut write fOnThreadReceptionTimeOut;
    // property OnThreadReceptionReview: TRxReadReview read fOnThreadReceptionReview write fOnThreadReceptionReview;
    property OnThreadReadTermination: TRxReadTermination read fOnThreadReadTermination write fOnThreadReadTermination;
  end;
{$ENDIF}

  TCustomComDevice = class(TComponent)
  private
    { Private declarations }
    fComHandle: THandle;
    fSmallTimeOut: TTimeOutController;
    fLargeTimeOut: TTimeOutController;
{$IFDEF WithThread}
    fComReceiver: TComReceiver;
{$ENDIF}
    fInputDSRSensivity,
    fAppendEOSonWrite,
    fBreakReadOnEOS,
    fRemoveEOSOnRead,
    fValidateBaudRate,
    fSimulationMode,
    fThreadActive: boolean;
    fComNumber: integer;
    fBaudRate: dword;
    fByteSize,
    fStopBits,
    fParityMode,
    fDTRControl,
    fRTSControl: byte;
    fEOSChar,
    fXonChar,
    fXoffChar: char;
    fOutputControlMode: TOutputControlMode;
    fXonXoffControlMode: TXonXoffControlMode;
    fCharTransmitInterval,
    fCharTransmitDuration,
    fWriteTimeOut,
    fReadTimeOut,
    fMaxStringLength: word;
{$IFDEF WithThread}
    fComReceptionString,
{$ENDIF}
    fComSimulationString: string;
    fOnTransmitChar,
    fOnTransmissionFinished,
    fOnReceiveChar,
    fOnReceptionFinished: TSimpleEvent;
{$IFDEF WithThread}
    fOnReadChar: TRxCharRead;
    fOnReadString,
    fOnReceptionTimeOut: TRxStringRead;
    // fOnReceptionReview: TRxReadReview;
    fOnReadTermination: TRxReadTermination;
{$ENDIF}
    fOnReadStringSimulation: TRxStringRequest;
  protected
    { Protected declarations }
    function GetComOpenStatus: boolean;
    procedure SetComNumber(Value: integer);
    procedure SetBaudRate(Value: dword);
    procedure SetByteSize(Value: byte);
    procedure SetStopBits(Value: byte);
    procedure SetParityMode(Value: byte);
    procedure SetEOSChar(Value: char);
    procedure SetXonChar(Value: char);
    procedure SetXoffChar(Value: char);
    procedure SetDTRControl(Value: byte);
    procedure SetRTSControl(Value: byte);
    procedure SetOutputControlMode(Value: TOutputControlMode);
    procedure SetXonXoffControlMode(Value: TXonXoffControlMode);
    procedure SetInputDSRSensivity(Value: boolean);
    procedure SetMaxStringLength (Value: word);
    procedure SetValidateBaudRate(Value: boolean);
    procedure SetCharTransmitInterval(Value: word);
    procedure SetWriteTimeOut(Value: word);
    procedure SetReadTimeOut(Value: word);
    function GetCTSLineStatus: boolean;
    function GetDSRLineStatus: boolean;

    function PutString(Source: string): boolean;
    function TransmitString(Source: string): boolean;
    function SimulateTransmission (Source: string): boolean;

    function CharsInQue: word;
    function CharsOutQue: word;
    function GetChar(var Target: char): boolean;
    function GetString(CharCount: byte): string;
    function EvaluateComSettings: TFComSettings;

    property ComOpen: boolean read GetComOpenStatus;
    property ComNumber: integer read fComNumber write SetComNumber default -1;
    property BaudRate: dword read fBaudRate write SetBaudRate default CDefaultComBaudRate;
    property ByteSize: byte read fByteSize write SetByteSize default DATABITS_8;
    property StopBit: byte read fStopBits write SetStopBits default ONESTOPBIT;
    property ParityMode: byte read fParityMode write SetParityMode default NOPARITY;
    property EOSChar: char read fEOSChar write SetEOSChar default CLF;
    property XonChar: char read fXonChar write SetXonChar default CXon;
    property XoffChar: char read fXoffChar write SetXoffChar default CXoff;
    property DTRControl: byte read fDTRControl write SetDTRControl default DTR_CONTROL_DISABLE;
    property RTSControl: byte read fRTSControl write SetRTSControl default RTS_CONTROL_DISABLE;
    property OutputControlMode: TOutputControlMode read fOutputControlMode write SetOutputControlMode default ocNone;
    property XonXoffControlMode: TXonXoffControlMode read fXonXoffControlMode write SetXonXoffControlMode default xcNone;
    property InputDSRSensivity: boolean read fInputDSRSensivity write SetInputDSRSensivity default false;
    property AppendEOSonWrite: boolean read fAppendEOSonWrite write fAppendEOSonWrite default false;
    property BreadReadOnEOS: boolean read fBreakReadOnEOS write fBreakReadOnEOS default false;
    property RemoveEOSonRead: boolean read fRemoveEOSonRead write fRemoveEOSonRead default false;
    property MaxStringLength: word read fMaxStringLength write SetMaxStringLength;
    property ValidateBaudRate: boolean read fValidateBaudRate write SetValidateBaudRate default false;
    property CharTransmitInterval: word read fCharTransmitInterval write SetCharTransmitInterval default 0;
    property WriteTimeOut: word read fWriteTimeOut write SetWriteTimeOut default 0;
    property ReadTimeOut: word read fReadTimeOut write SetReadTimeOut default 0;
    property ThreadActive: boolean read fThreadActive;

    property CTSLineStatus: boolean read GetCTSLineStatus;
    property DSRLineStatus: boolean read GetDSRLineStatus;

    property OnTransmitChar: TSimpleEvent read fOnTransmitChar write fOnTransmitChar;
    property OnTransmissionFinished: TSimpleEvent read fOnTransmissionFinished write fOnTransmissionFinished;
    property OnReceiveChar: TSimpleEvent read fOnReceiveChar write fOnReceiveChar;
    property OnReceptionFinished: TSimpleEvent read fOnReceptionFinished write fOnReceptionFinished;
{$IFDEF WithThread}
    property OnReadChar: TRxCharRead read fOnReadChar write fOnReadChar;
    property OnReadString: TRxStringRead read fOnReadString write fOnReadString;
    property OnReceptionTimeOut: TRxStringRead read fOnReceptionTimeOut write fOnReceptionTimeOut;
    // property OnReceptionReview: TRxReadReview read fOnReceptionReview write fOnReceptionReview;
    property OnReadTermination: TRxReadTermination read fOnReadTermination write fOnReadTermination;
{$ENDIF}
    property OnReadStringSimulation: TRxStringRequest read fOnReadStringSimulation write fOnReadStringSimulation;
  public
    { Public declarations }
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;

    function Open: boolean; virtual;
    procedure Close; virtual;

    procedure PurgeReadBuffer;
    procedure PurgeWriteBuffer;
    procedure PurgeReadWriteBuffer;

    function GetCTSandDSRLineStatus: TModemStatus;
    function WriteString(Source: string): boolean; virtual;
    function ReadString(var Target: string; CharCount: byte): boolean; virtual;
    function QueryString(Question: string; BreakTime: integer; var Answer: string; CharCount: byte): boolean; virtual;
    function ForceReceiverSimulation(Value: string): boolean;
  published
    { Published declarations }
  end;

  TSimpleComDevice = class(TCustomComDevice)
  private
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
  published
    { Published declarations }
    property ComOpen;
    property ComNumber;
    property BaudRate;
    property ByteSize;
    property StopBit;
    property ParityMode;
    property EOSChar;
    property XonChar;
    property XoffChar;
    property DTRControl;
    property RTSControl;
    property OutputControlMode;
    property XonXoffControlMode;
    property InputDSRSensivity;
    property AppendEOSonWrite;
    property BreadReadOnEOS;
    property RemoveEOSonRead;
    property MaxStringLength;
    property ValidateBaudRate;
    property CharTransmitInterval;
    property WriteTimeOut;
    property ReadTimeOut;
    property ThreadActive;
    property CTSLineStatus;
    property DSRLineStatus;

    property OnTransmitChar;
    property OnTransmissionFinished;
    property OnReceiveChar;
    property OnReceptionFinished;
{$IFDEF WithThread}
    property OnReadChar;
    property OnReadString;
    property OnReceptionTimeOut;
    property OnReadTermination;
{$ELSE}
    property OnReadStringSimulation;
{$ENDIF}
  end;

  procedure Register;

var
  TerminateWaitSerialComLoop: boolean;

implementation

uses SysUtils, Forms, RegUtils, ProcessUtils, ConversionUtils;

  // COM   - function and pin assignment
  // RxD   - input     -  Pin 2 on DB9  -  Pin 3 on DB25
  // TxD   - output    -  Pin 3 on DB9  -  Pin 2 on DB25
  // DTR   - output    -  Pin 4 on DB9  -  Pin
  // DSR   - input     -  Pin 6 on DB9  -  Pin
  // RTS   - output    -  Pin 7 on DB9  -  Pin
  // CTS   - input     -  Pin 8 on DB9  -  Pin
  // GND   -           -  Pin 5 on DB9  -  Pin 7 on DB25

const
  CSerialCommInfo  = 'Hardware\DeviceMap\SerialComm';
  CModemDeviceInfo = 'System\CurrentControlSet\Services\Class\Modem';

  CBaudRatesTable: array [0..14] of dword = (CBR_110, CBR_300, CBR_600, CBR_1200, CBR_2400, CBR_4800, CBR_9600, CBR_14400,
                                             CBR_19200, CBR_38400, CBR_56000, CBR_57600, CBR_115200, CBR_128000, CBR_256000);


procedure Register;
begin
  RegisterComponents('Hardware', [TSimpleComDevice]);
end;

{function GetStringParityByte (Source: string): byte;
var I, L: integer;
    PB: byte;
begin
  L:=System.Length(Source); PB:=0;
  for I:=1 to L do
      PB:=PB xor ord(Source[I]);
  Result:=PB;
end;
}
function AcceptableComBaudRate(Value: dword): boolean;
var I: byte;
begin
  Result:=false;
  for I:=0 to System.Length(CBaudRatesTable)-1 do
      if (Value=CBaudRatesTable[I]) then
         begin
         Result:=true;
         break;
         end;
end;

function GetNextComBaudRateValue(Value: dword; Ascending: boolean): dword;
var I,L: byte;
begin
  L:=System.Length(CBaudRatesTable);
  for I:=L-1 downto 0 do
      if (Value=CBaudRatesTable[I]) then break;
  if Ascending then
     begin
     if (I<L-2) then System.Inc(I);
     end
  else  // not(Ascending)
     begin
     if (I>1) then System.Dec(I);
     end;
  Result:=CBaudRatesTable[I];
end;

// The below character transmission time is in micro seconds and gives acceptable
// result accuracy when the Baud Rate is between 100 to 100k
// Note: The result is calculated sligthly higher for StopBits > 1.
function GetComCharTransmissionTime(ComSettings: TFComSettings): word;
begin
  Result:=(1000000*(ComSettings.ByteSize+ComSettings.StopBits+1)) DIV ComSettings.BaudRate;
end;

// SerialComUtils functions and procedures

function GetMaxSerialPortNumber: dword;
var fKeyNamesStringSet: TStringList;
    fAString: string;
    I: integer;
    CN: dword;
begin
  Result:=0;
  fKeyNamesStringSet:=TStringList.Create;
  GetSubKeyNamesFromRegistry(HKEY_LOCAL_MACHINE,CSerialCommInfo,fKeyNamesStringSet);

  for I:=0 to fKeyNamesStringSet.Count-1 do
      begin
      fAString:=LoadStringFromRegistry(HKEY_LOCAL_MACHINE,CSerialCommInfo,fKeyNamesStringSet.Strings[I],'');
      if (fAString<>'') and (System.Pos('COM',UpperCase(fAString))=1) then
         begin
         System.Delete(fAstring,1,3);
         CN:=StrToIntDef(fAString,0);
         if (CN>Result) then
            Result:=CN;
         end;
      end;
  fKeyNamesStringSet.Destroy;
end;

function SerialPortExists(ComNumber: word): boolean;
var fKeyNamesStringSet: TStringList;
    fAString: string;
    I: integer;
begin
  Result:=false;
  if (ComNumber in [1..99]) then
     begin
     fAString:='COM'+IntToStr(ComNumber);
     fKeyNamesStringSet:=TStringList.Create;
     GetSubKeyNamesFromRegistry(HKEY_LOCAL_MACHINE,CSerialCommInfo,fKeyNamesStringSet);
     for I:=0 to fKeyNamesStringSet.Count-1 do
         begin
         if fAString=LoadStringFromRegistry(HKEY_LOCAL_MACHINE,CSerialCommInfo,fKeyNamesStringSet.Strings[I],IntToStr(I)) then
            begin
            Result:=true;
            break;
            end;
         end;
     fKeyNamesStringSet.Destroy;
     end;
end;

function SerialPortAvailable(ComNumber: word): boolean;
var fPString: PChar;
    fComHandle: THandle;
begin
  if (ComNumber in [1..99]) then
     begin
     fPString:=StrAlloc(6);
     StrPCopy(fPString,'COM'+IntToStr(ComNumber));
     fComHandle:=CreateFile(PChar(fPString),
                            GENERIC_READ or GENERIC_WRITE, 0, nil,
                            OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
     StrDispose(fPString);
     Result:=(fComHandle<>INVALID_HANDLE_VALUE) and (GetFileType(fComHandle)=FILE_TYPE_CHAR);
     CloseHandle(fComHandle);
     end
  else Result:=false;
end;

function GetSerialPortPresenceInfo(MaxComNumber: word): TComPresenceInfoArray;
var I: integer;
begin
  // be careful with this procedure, it may take very long !
  for I:=1 to 99 do
         if not(I>MaxComNumber) then
            begin
            Result[I].Exists:=SerialPortExists(I);
            Result[I].Available:=SerialPortAvailable(I);
            if ((I mod 10)=0) then           // call ProcessMessages every 10th loop
               Application.ProcessMessages;  // to avoid freezed effect operation
            end
         else
            begin
            Result[I].Exists:=false;
            Result[I].Available:=false;
            end;
end;

function ModemOnSerialPort(ComNumber: word): boolean;
var fAString,fTString,fRString: string;
    fKStrings: TStringList;
    I: integer;
begin
  if (ComNumber in [1..99]) then
     begin
     fKStrings:=TStringList.Create;
     GetSubKeySetFromRegistry(HKEY_LOCAL_MACHINE,CModemDeviceInfo,fKStrings);
     fAString:='COM'+IntToStr(ComNumber);
     for I:=0 to fKStrings.Count-1 do
         begin
         fTString:=fKStrings.Strings[I];
         fRString:=LoadStringFromRegistry(HKEY_LOCAL_MACHINE,CModemDeviceInfo+'\'+fTString,'AttachedTo','');
         if (fRString=fAString) then break;
         end;
     fKStrings.Free;
     Result:=(fRString=fAString);
     end
  else Result:=false;
end;

function OpenSerialPort(ComSettings: TFComSettings): THandle;
var fComTimeouts: TCommTimeouts;
    fComHandle: THandle;
    fCOMOpen: boolean;
    fPString: PChar;
    fComDCB: TDCB;
begin
  if SerialPortExists(ComSettings.Number) and
     not(ModemOnSerialPort(ComSettings.Number)) and
     SerialPortAvailable(ComSettings.Number) then
     begin
     fCOMOpen:=false;
     fPString:=StrAlloc(6);
     StrPCopy(fPString,'COM'+IntToStr(ComSettings.Number));
     fComHandle:=CreateFile(PChar(fPString),
                            GENERIC_READ or GENERIC_WRITE, 0, nil,
                            OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);

     if (fComHandle <> INVALID_HANDLE_VALUE) then
        begin
        if (GetFileType(fComHandle)=FILE_TYPE_CHAR) then
           begin
           FillChar(fComTimeOuts, Sizeof(fComTimeOuts), 0);
           fComTimeOuts.ReadIntervalTimeout:= MAXDWORD;
           if (SetCommTimeOuts(fComHandle, fComTimeOuts)) then
              begin
              if GetCommState(fComHandle, fComDCB) then
                 begin
                 // insert ComSettings configuration to the DCB structure
                 fComDCB.BaudRate:=ComSettings.BaudRate;
                 fComDCB.ByteSize:=ComSettings.ByteSize;
                 fComDCB.StopBits:=ComSettings.StopBits;
                 fComDCB.Parity:=ComSettings.Parity;
                 fComDCB.Flags:=dcb_Binary; // Win32 Platform will accept only binary mode

                 fComDCB.Flags:=fComDCB.Flags or ComSettings.Flags;

                 if ((fComDCB.Flags and (dcb_InX or dcb_OutX))>0) then
                    begin
                    fComDCB.XonChar:=ComSettings.XonChar;
                    fComDCB.XoffChar:=ComSettings.XoffChar;
                    end;
                 fComDCB.EofChar:=ComSettings.EOSChar;
                 // and configure Com to the Settings
                 if SetCommState(fComHandle, fComDCB) then
                    // condition "and SetupComm(fComHandle, 255, 255)" was removed
                    // because it will not work on computers with Windows NT
                    begin
                    fCOMOpen:=true;
                    PurgeComm(fComHandle, CPurgeReadWrite);
                    end;
                 end;
              end;
           end;
        end;

     StrDispose(fPString);

     if not(fCOMOpen) and (fComHandle <> INVALID_HANDLE_VALUE) then
        begin
        CloseHandle(fComHandle);
        fComHandle:=INVALID_HANDLE_VALUE;
        end;
     Result:=fComHandle;
     end
  else Result:=INVALID_HANDLE_VALUE;
end;

function SerialPortOpened(fComHandle: THandle): boolean;
begin
  Result:=(fComHandle <> INVALID_HANDLE_VALUE);
end;

procedure CloseSerialPort(var fComHandle: THandle);
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     CloseHandle(fComHandle);
     fComHandle:=INVALID_HANDLE_VALUE;
     end;
end;

function SetTXDLine(fComHandle: THandle; TXD: boolean): boolean;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     if TXD then EscapeCommFunction(fComHandle,SETBREAK)
     else EscapeCommFunction(fComHandle,CLRBREAK);
     Result:=(GetLastError=0);
     end
  else Result:=false;
end;

function SetDTRLine(fComHandle: THandle; DTR: boolean): boolean;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     if DTR then EscapeCommFunction(fComHandle,SETDTR)
     else EscapeCommFunction(fComHandle,CLRDTR);
     Result:=(GetLastError=0);
     end
  else Result:=false;
end;

function SetRTSLine(fComHandle: THandle; RTS: boolean): boolean;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     if RTS then EscapeCommFunction(fComHandle,SETRTS)
     else EscapeCommFunction(fComHandle,CLRRTS);
     Result:=(GetLastError=0);
     end
  else Result:=false;
end;

function SetDTRandRTSlines(fComHandle: THandle; DTR,RTS: boolean): boolean;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     if DTR then EscapeCommFunction(fComHandle,SETDTR)
     else EscapeCommFunction(fComHandle,CLRDTR);
     if RTS then EscapeCommFunction(fComHandle,SETRTS)
     else EscapeCommFunction(fComHandle,CLRRTS);
     Result:=(GetLastError=0);
     end
  else Result:=false;
end;

function GetCTSLine(fComHandle: THandle): boolean;
var fModemStat: dword;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) and
     (GetCommModemStatus(fComHandle,fModemStat)) then
     begin
     Result:=((fModemStat and MS_CTS_ON)>0) and (GetLastError=0);
     end
  else Result:=false;
end;

function GetDSRLine(fComHandle: THandle): boolean;
var fModemStat: dword;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) and
     (GetCommModemStatus(fComHandle,fModemStat)) then
     begin
     Result:=((fModemStat and MS_DSR_ON)>0) and (GetLastError=0);
     end
  else Result:=false;
end;

function GetCTSandDSRLines(fComHandle: THandle): TModemStatus;
var fModemStat: dword;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) and
     (GetCommModemStatus(fComHandle,fModemStat)) and
     (GetLastError=0) then
     begin
     Result.CTSLine:=((fModemStat and MS_CTS_ON)>0);
     Result.DSRLine:=((fModemStat and MS_DSR_ON)>0);
     end
  else
     begin
     Result.CTSLine:=false;
     Result.DSRLine:=false;
     end;
end;

procedure PurgeCOMReadBuffer(fComHandle: THandle);
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     PurgeComm(fComHandle, CPurgeRead);
     end;
end;

procedure PurgeCOMWriteBuffer(fComHandle: THandle);
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     PurgeComm(fComHandle, CPurgeWrite);
     end;
end;

procedure PurgeCOMReadWriteBuffer(fComHandle: THandle);
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     PurgeComm(fComHandle, CPurgeReadWrite);
     end;
end;

function GetCOMCharsInQue(fComHandle: THandle): dword;
var fStat: TComStat;
    fErrors: dword;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) and
     ClearCommError(fComHandle, fErrors, @fStat) then
     begin
     Result:=fStat.cbInQue;
     end
  else Result:=0;
end;

function GetCOMCharsOutQue(fComHandle: THandle): dword;
var fStat: TComStat;
    fErrors: dword;
begin
  // Attention! Please note, that for non-overlapped write access
  // this function will always return zero (and not the que size)

  // Furthermore, please note that for overlapped write access
  // the result means the software interface buffer
  // and not the hardware communication transmitter !!!
    if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     ClearCommError(fComHandle, fErrors, @fStat);
     Result:=fStat.cbOutQue;
     end
  else Result:=$FFFF;
end;

function SetCOMMonitoredEvents(fComHandle: THandle; EvtMask: dword): boolean;
begin
  if (fComHandle <> INVALID_HANDLE_VALUE) and
     SetCommMask(fComHandle,EvtMask) then
     Result:=true
  else Result:=false;
end;

function GetCOMMonitoredEventsStatus(fComHandle: THandle; EvtMask: dword): boolean;
var fEvtMask: dword;
begin
  // Attention! Please note, that the to be monitored event must be enabled before
  // (see the above function). Please note also, that for multiple event mask
  // (more than one event) this function will return ambiguous result.
  // Furthermore, please note that e.g. the EV_TXEMPTY event status means software
  // interface buffer and not the hardware communication transmitter !!!
  if (fComHandle <> INVALID_HANDLE_VALUE) and
     GetCommMask(fComHandle,fEvtMask) then
     begin
     Result:=(fEvtMask and EvtMask)>0;
     end
  else Result:=false;
end;

function PutStringToCOM(fComHandle: THandle; Str: string): boolean;
var fOverlapped: TOverlapped;
    fPString: TCharBuffer;
    fErrorCode: dword;
    fLength: integer;
    fStat: TComStat;
begin
  // This function may return with success before the IO write operation is completed
  // The success result means that the operation was correctly passed to the API interface
  // and when the system indication is ERROR_IO_PENDING, then success is pressumed.
  Result:=false;
  if (fComHandle <> INVALID_HANDLE_VALUE) and             // serial port is open
     ClearCommError(fComHandle, fErrorCode, @fStat) and   // function call succeeds
     ((fErrorCode and CE_TXFULL) = 0) then                // output buffer is not full
     begin
     fLength:=System.Length(Str);
     StrCopyBuff(fPString,Str);

     FillChar(fOverlapped,SizeOf(fOverlapped),0);

     if not(WriteFile(fComHandle, fPString, fLength, fErrorCode, @fOverlapped)) then
        begin                                   // function call fail (but don't worry)
        fErrorCode:=GetLastError;               // possible only because of
        Result:=(fErrorCode=ERROR_IO_PENDING);  // the IO write is pending
        end
     else Result:=(fErrorCode=dword(fLength));
     end;
end;

function PutStringToCOM(fComHandle: THandle; Str: string; TOC: TTimeOutController; TimeOut: dword): boolean; overload;
var fOverlapped: TOverlapped;
    fPString: TCharBuffer;
    fErrorCode: dword;
    fLength: integer;
    fStat: TComStat;
begin
  // This function will return with success when the IO write operation is correctly completed,
  // however it means that the last byte was successful passed to the hardware interface.
  // It requires in the interface parameters pointer to specialised Event object and TimeOut value
  // When TimeOut is elapsed before the IO operation is completed, then the return value is false
  Result:=false;
  if (fComHandle <> INVALID_HANDLE_VALUE) and             // serial port is open
     ClearCommError(fComHandle, fErrorCode, @fStat) and   // function call succeeds
     ((fErrorCode and CE_TXFULL) = 0) then                // output buffer is not full
     begin
     fLength:=System.Length(Str);
     StrCopyBuff(fPString,Str);

     FillChar(fOverlapped,SizeOf(fOverlapped),0);
     TOC.Initialize(TimeOut);

     if not(WriteFile(fComHandle, fPString, fLength, fErrorCode, @fOverlapped)) then
        begin
        if (GetLastError=ERROR_IO_PENDING) then
           begin
           while not(TOC.Elapsed) and not(GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false)) do;
           Result:=(fErrorCode=dword(fLength));
           end;
        end
     else Result:=(fErrorCode=dword(fLength));
     end;
end;

function WaitUntilCOMCharsAvailableOrTimeOutElapsed(fComHandle: THandle; CharCount: byte; TOC: TTimeOutController; TimeOut: dword): boolean;
var fErrorCode: dword;
    fStat: TComStat;
begin
  if (ClearCommError(fComHandle, fErrorCode, @fStat)) then
     begin
     TerminateWaitSerialComLoop:=false;
     if (TimeOut>0) then // wait until received char pending or timeout elapsed
        begin
        TOC.Initialize(TimeOut);
        while not(TerminateWaitSerialComLoop) and not(TOC.Elapsed) and
              ClearCommError(fComHandle, fErrorCode, @fStat) and (fStat.cbInQue<CharCount) do;
        end;
     Result:=not(TerminateWaitSerialComLoop) and not(fStat.cbInQue<CharCount);
     end
  else Result:=false;
end;

function GetCharFromCOM(fComHandle: THandle; var Ch: char): boolean;
var fOverlapped: TOverlapped;
    fPString: TTwoChars;
    fErrorCode: dword;
begin
  // If at least one character is available, this function will perform the IO read operation
  // and will return with true (and set Ch), otherwise it will return immediately with false.
  Result:=false;
  // Note: The function "GetCOMCharsInQue" will also check if the fComHandle value is valid
  if (GetCOMCharsInQue(fComHandle)>0) then
     begin
     FillChar(fOverlapped,SizeOf(fOverlapped),0);
     if ReadFile(fComHandle, fPString, 1, fErrorCode, @fOverlapped) and
        GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false) and
        (fErrorCode=1) then
        begin
        Ch:=fPString[0];
        Result:=true;
        end;
     end;
end;

function GetCharFromCOM(fComHandle: THandle; var Ch: char; TOC: TTimeOutController; TimeOut: dword): boolean;
var fOverlapped: TOverlapped;
    fPString: TTwoChars;
    fErrorCode: dword;
begin
  // This function will wait until at least one character is available (and perform IO read operation
  // and set Ch value) or until the TimeOut is elapsed (and return in this case with false).
  // Used with parameter TimeOut=0 it is identical as the previous overloaded version.
  Result:=false;
  if (fComHandle <> INVALID_HANDLE_VALUE) then
     begin
     if WaitUntilCOMCharsAvailableOrTimeOutElapsed(fComHandle,1,TOC,TimeOut) then
        begin
        FillChar(fOverlapped,SizeOf(fOverlapped),0);
        if ReadFile(fComHandle, fPString, 1, fErrorCode, @fOverlapped) and
           GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false) and
           (fErrorCode=1) then
           begin
           Ch:=fPString[0];
           Result:=true;
           end;
        end;
     end;
end;

function GetStringFromCOM(fComHandle: THandle; CharCount: byte; BreakReadOnEOS: boolean; EOSChar: char): string;
var fOverlapped: TOverlapped;
    fPString: TCharBuffer;
    fTempReceive: string;
    fLength,I: integer;
    fErrorCode: dword;
    fStat: TComStat;
begin
  fTempReceive:='';
  if (CharCount>0) and
     (fComHandle <> INVALID_HANDLE_VALUE) and
     (ClearCommError(fComHandle, fErrorCode, @fStat)) then
     begin
     if not(fStat.cbInQue<CharCount) then fLength:=CharCount else fLength:=fStat.cbInQue;
     if (fLength>CMaxBufferLength) then fLength:=CMaxBufferLength;
     if (fLength>0) then
        begin
        FillChar(fOverlapped,SizeOf(fOverlapped),0);
        if not(BreakReadOnEOS) then
           begin
           if ReadFile(fComHandle, fPString, fLength, fErrorCode, @fOverlapped) and
              GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false) and
              (fErrorCode>0) then
              for I:=0 to fErrorCode-1 do
                  fTempReceive:=fTempReceive+fPString[I];
           end
        else  // BreakReadOnEOS
           begin
           for I:=1 to fLength do
               begin
               if ReadFile(fComHandle, fPString, 1, fErrorCode, @fOverlapped) and
                  GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false) and
                  (fErrorCode=1) then
                  begin
                  fTempReceive:=fTempReceive+fPString[0];
                  if (fPString[0]=EOSChar) then break;
                  end
               else break;
               end;
           end;
        end;
     end;
  Result:=fTempReceive;
end;

function GetStringFromCOM(fComHandle: THandle; CharCount: byte; BreakReadOnEOS: boolean; EOSChar: char; TOC: TTimeOutController; TimeOut: dword): string;
var fOverlapped: TOverlapped;
    fPString: TCharBuffer;
    fTempReceive: string;
    fLength,I: integer;
    fErrorCode: dword;
    fStat: TComStat;
begin
  fTempReceive:='';
  if (CharCount>0) and
     (fComHandle <> INVALID_HANDLE_VALUE) and
     (ClearCommError(fComHandle, fErrorCode, @fStat)) then
     begin
     if not(BreakReadOnEOS) then
        begin
        if WaitUntilCOMCharsAvailableOrTimeOutElapsed(fComHandle,CharCount,TOC,TimeOut) then
           fLength:=CharCount
        else if ClearCommError(fComHandle, fErrorCode, @fStat) then
                fLength:=fStat.cbInQue
             else fLength:=0;
        if (fLength>0) then
           begin
           FillChar(fOverlapped,SizeOf(fOverlapped),0);
           if ReadFile(fComHandle, fPString, fLength, fErrorCode,@fOverlapped) and
              GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false) and
              (fErrorCode>0) then
              for I:=0 to fErrorCode-1 do
                  fTempReceive:=fTempReceive+fPString[I];
           end;
        end
     else  // BreakReadOnEOS
        begin
        if (TimeOut>0) then
           begin
           FillChar(fOverlapped,SizeOf(fOverlapped),0);
           TOC.Initialize(TimeOut);
           TerminateWaitSerialComLoop:=false;
           while not(TerminateWaitSerialComLoop) and not(TOC.Elapsed) do
                 begin
                 if ClearCommError(fComHandle, fErrorCode, @fStat) and
                    (fStat.cbInQue>0) then
                    begin
                    fErrorCode:=GetLastError;
                    if ReadFile(fComHandle, fPString, 1, fErrorCode, @fOverlapped) and
                       GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false) and
                       (fErrorCode=1) then
                       begin
                       fTempReceive:=fTempReceive+fPString[0];
                       if (fPString[0]=EOSChar) or
                          not(System.Length(fTempReceive)<CharCount) then break;
                       end
                    else break;
                    end;
                 end;
           end
        else // not(TimeOut>0)
           begin
           if not(fStat.cbInQue<CharCount) then fLength:=CharCount else fLength:=fStat.cbInQue;
           if (fLength>CMaxBufferLength) then fLength:=CMaxBufferLength;
           if (fLength>0) then
              begin
              FillChar(fOverlapped,SizeOf(fOverlapped),0);
              for I:=1 to fLength do
                  begin
                  if ReadFile(fComHandle, fPString, 1, fErrorCode, @fOverlapped) and
                     GetOverlappedResult(fComHandle, fOverlapped, fErrorCode, false) and
                     (fErrorCode=1) then
                     begin
                     fTempReceive:=fTempReceive+fPString[0];
                     if (fPString[0]=EOSChar) then break;
                     end
                  else break;
                  end;
              end;
           end;
        end;
     end;

  Result:=fTempReceive;
end;

// TComReceiver

{$IFDEF WithThread}
constructor TComReceiver.Create;
begin
  inherited Create(true);

  pComHandle:=nil;
  pBreakReadOnEOS:=nil;
  pRemoveEOSOnRead:=nil;
  pSimulationMode:=nil;

  pEOSchar:=nil;
  pReadTimeOut:=nil;
  pMaxStringLength:=nil;
  pReceptionString:=nil;
  pSimulationString:=nil;

  fOnThreadReceiveChar:=nil;
  fOnThreadReceptionFinished:=nil;

  fOnThreadReadChar:=nil;
  fOnThreadReadString:=nil;
  fOnThreadReceptionTimeOut:=nil;
  // fOnThreadReceptionReview:=nil;
  fOnThreadReadTermination:=nil;
end;

destructor TComReceiver.Destroy;
begin

  inherited;
end;

procedure TComReceiver.Resume;
begin
  if not(fEnabled) then
     begin
     fReceptionFinished:=false;
     fReceptionTimeOut:=false;
     fThreadReceptionString:='';

     inherited Resume;
     fEnabled:=true;
     end;
end;

procedure TComReceiver.Suspend;
begin
  if fEnabled then
     begin
     fEnabled:=false;
     while not(fDisabled) do;
     inherited Suspend;
     end;
end;

function TComReceiver.GetSimulatedChar(var Ch: char): boolean;
begin
  if (pSimulationString<>nil) and
     (pSimulationString^<>'') then
     begin
     Ch:=pSimulationString^[1];
     System.Delete(pSimulationString^,1,1);
     Result:=true;
     end
  else Result:=false;
end;

procedure TComReceiver.Execute;
var StringLength: word;
    NewChar: char;
begin
  StringLength:=0;

  while True do
        begin
        if fEnabled then
           begin
           if fDisabled then fDisabled:=false;

           if (pComHandle<>nil) and
              (pSimulationMode<>nil) then
              begin

              if (not(pSimulationMode^) and
                 (GetCOMCharsInQue(pComHandle^)>0) and
                 GetCharFromCOM(pComHandle^,NewChar)) or
                 (pSimulationMode^ and (GetSimulatedChar(NewChar))) then
                 begin

                 if Assigned(fOnThreadReceiveChar) then fOnThreadReceiveChar;

                 if Assigned(fOnThreadReadChar) then fOnThreadReadChar(NewChar)
                 else // not(Assigned(fOnThreadReadChar))
                    begin
                    if (fThreadReceptionString='') then
                       begin
                       fReceptionBegin:=GetTickCount;
                       if (pReadTimeOut<>nil) then
                          fReceptionTimeOutLimit:=fReceptionBegin+pReadTimeOut^
                       else fReceptionTimeOutLimit:=fReceptionBegin+CDefaultReadTimeOut;
                       end;

                    fThreadReceptionString:=fThreadReceptionString+NewChar;
                    StringLength:=System.Length(fThreadReceptionString);

                    if (StringLength=10) then// wait more patiently for long strings in reception
                       begin
                       if (pReadTimeOut<>nil) then
                          fReceptionTimeOutLimit:=fReceptionTimeOutLimit+pReadTimeOut^
                       else fReceptionTimeOutLimit:=fReceptionTimeOutLimit+CDefaultReadTimeOut;
                       end;

                    fReceptionFinished:=((pBreakReadOnEOS<>nil) and (pBreakReadOnEOS^) and (pEOSchar<>nil) and (NewChar=pEOSChar^)) or
                                         ((pMaxStringLength<>nil) and not(StringLength<pMaxStringLength^));
                    end;
                 if not(fReceptionFinished) and
                    Assigned(fOnThreadReadTermination) then
                    fOnThreadReadTermination(fThreadReceptionString,GetTickCount-fReceptionBegin,fReceptionFinished);
                 end
              else // no any char available, so control the time out
                 begin
                 if (fThreadReceptionString<>'') and not(fReceptionTimeOut) then
                    fReceptionTimeOut:=not(fReceptionTimeOutLimit>GetTickCount);
                 end;

              if fReceptionFinished or fReceptionTimeOut then
                 begin
                 if ((pRemoveEOSOnRead<>nil) and (pRemoveEOSOnRead^) and (pEOSchar<>nil) and (NewChar=pEOSChar^)) then
                    System.Delete(fThreadReceptionString,StringLength,1);

                 if Assigned(fOnThreadReceptionFinished) then fOnThreadReceptionFinished;

                 if not(fReceptionTimeOut) then
                    begin
                    if Assigned(fOnThreadReadString) then
                       fOnThreadReadString(fThreadReceptionString)
                    else if (pReceptionString<>nil) then
                            begin
                            while (pReceptionString^<>'') do; // wait ages until serviced
                            pReceptionString^:=fThreadReceptionString;
                            end;
                    end
                 else // ReceptionTimeOut
                    begin
                    if Assigned(fOnThreadReceptionTimeOut) then
                       fOnThreadReceptionTimeOut(fThreadReceptionString)
                    else if (pReceptionString<>nil) then
                            begin
                            while (pReceptionString^<>'') do; // wait ages until serviced
                            pReceptionString^:=fThreadReceptionString;
                            end;
                    end;

                 fReceptionFinished:=false;
                 fReceptionTimeOut:=false;
                 fThreadReceptionString:='';
                 end;
              end;
           end
        else // not(fEnabled)
           if not(fDisabled) then fDisabled:=true;
        end;
end;
{$ENDIF}

// TCustomComDevice

function TCustomComDevice.GetComOpenStatus: boolean;
begin
  Result:=(fComHandle <> INVALID_HANDLE_VALUE);
end;

procedure TCustomComDevice.SetComNumber(Value: integer);
begin
  if not(GetComOpenStatus) and (fComNumber<>Value) and (Value in [0..99]) then fComNumber:=Value;
end;

procedure TCustomComDevice.SetBaudRate(Value: dword);
begin
  if not(GetComOpenStatus) and (fBaudRate<>Value) and
     (not(fValidateBaudRate) or AcceptableComBaudRate(Value)) then
     begin
     fBaudRate:=Value;
     fCharTransmitDuration:=(1000000*(fByteSize+fStopBits+1)) DIV fBaudRate;
     end;
end;

procedure TCustomComDevice.SetByteSize(Value: byte);
begin
  if not(GetComOpenStatus) and (fByteSize<>Value) and (Value in [4..8]) then
     begin
     fByteSize:=Value;
     fCharTransmitDuration:=(1000000*(fByteSize+fStopBits+1)) DIV fBaudRate;
     end;
end;

procedure TCustomComDevice.SetStopBits(Value: byte);
begin
  if not(GetComOpenStatus) and (fStopBits<>Value) and (Value in [ONESTOPBIT..TWOSTOPBITS]) then
     begin
     fStopBits:=Value;
     fCharTransmitDuration:=(1000000*(fByteSize+fStopBits+1)) DIV fBaudRate;
     end;
end;

procedure TCustomComDevice.SetParityMode(Value: byte);
begin
  if not(GetComOpenStatus) and (fParityMode<>Value) and (Value in [NOPARITY..SPACEPARITY]) then fParityMode:=Value;
end;

procedure TCustomComDevice.SetEOSChar(Value: char);
begin
  if not(GetComOpenStatus) and (fEOSChar<>Value) then fEOSChar:=Value;
end;

procedure TCustomComDevice.SetXonChar(Value: char);
begin
  if not(GetComOpenStatus) and (fXonChar<>Value) then fXonChar:=Value;
end;

procedure TCustomComDevice.SetXoffChar(Value: char);
begin
  if not(GetComOpenStatus) and (fXoffChar<>Value) then fXoffChar:=Value;
end;

procedure TCustomComDevice.SetDTRControl(Value: byte);
begin
  if not(GetComOpenStatus) and (fDTRControl<>Value) and (Value in [DTR_CONTROL_DISABLE..DTR_CONTROL_HANDSHAKE]) then fDTRControl:=Value;
end;

procedure TCustomComDevice.SetRTSControl(Value: byte);
begin
  if not(GetComOpenStatus) and (fRTSControl<>Value) and (Value in [RTS_CONTROL_DISABLE..RTS_CONTROL_TOGGLE]) then fRTSControl:=Value;
end;

procedure TCustomComDevice.SetOutputControlMode(Value: TOutputControlMode);
begin
  if not(GetComOpenStatus) and (fOutputControlMode<>Value) and (Value in [ocNone..ocCtsDsr]) then fOutputControlMode:=Value;
end;

procedure TCustomComDevice.SetXonXoffControlMode(Value: TXonXoffControlMode);
begin
  if not(GetComOpenStatus) and (fXonXoffControlMode<>Value) and (Value in [xcNone..xcInputOutput]) then fXonXoffControlMode:=Value;
end;

procedure TCustomComDevice.SetInputDSRSensivity(Value: boolean);
begin
  if not(GetComOpenStatus) and (fInputDSRSensivity<>Value) then fInputDSRSensivity:=Value;
end;

procedure TCustomComDevice.SetMaxStringLength(Value: word);
begin
  if (Value<>fMaxStringLength) and not(Value>1000) then
     fMaxStringLength:=Value;
end;

procedure TCustomComDevice.SetValidateBaudRate(Value: boolean);
begin
  if not(GetComOpenStatus) and (fValidateBaudRate<>Value) then
     begin
     fValidateBaudRate:=Value;
     if fValidateBaudRate and not(AcceptableComBaudRate(fBaudRate)) then
        begin
        fBaudRate:= CDefaultComBaudRate;
        fCharTransmitDuration:=(1000000*(fByteSize+fStopBits+1)) DIV fBaudRate;
        end;
     end;
end;

procedure TCustomComDevice.SetCharTransmitInterval(Value: word);
begin
  // Don't accept Char Transmit Interval > 1000
  if (fCharTransmitInterval<>Value) and not(fCharTransmitInterval>1000) then
     fCharTransmitInterval:=Value;
end;

procedure TCustomComDevice.SetWriteTimeOut(Value: word);
begin
  // Don't accept Time Out > 10000 (which is 10s)
  if (fWriteTimeOut<>Value) and not(fWriteTimeOut>10000) then
     fWriteTimeOut:=Value;
end;

procedure TCustomComDevice.SetReadTimeOut(Value: word);
begin
  // Don't accept Time Out > 60000 (which is 60s = 1min)
  if (fReadTimeOut<>Value) and not(fReadTimeOut>60000) then
     fReadTimeOut:=Value;
end;

function TCustomComDevice.GetCTSLineStatus: boolean;
begin
  Result:=GetCTSLine(fComHandle);
end;

function TCustomComDevice.GetDSRLineStatus: boolean;
begin
  Result:=GetDSRLine(fComHandle);
end;

function TCustomComDevice.PutString(Source: string): boolean;
var fMessageTransmissionTime: word;
begin
  if (fWriteTimeOut=0) then
     Result:=PutStringToCOM(fComHandle,Source)
  else
     begin
     fMessageTransmissionTime:=System.Round(System.Length(Source)*fCharTransmitDuration/1000);
     Result:=PutStringToCOM(fComHandle,Source,fSmallTimeOut,fMessageTransmissionTime+fWriteTimeOut);
     end;
end;

function TCustomComDevice.TransmitString(Source: string): boolean;
var fLength,I: integer;
begin
  if (fCharTransmitInterval=0) then
     begin
     if not(Assigned(fOnTransmitChar)) then
        Result:=PutString(Source)
     else // Assigned(fOnTransmitChar)
        begin
        Result:=true;
        fLength:=System.Length(Source);
        for I:=1 to fLength do
            begin
            if PutString(Source[I]) then fOnTransmitChar
            else
               begin
               Result:=false;
               break;
               end;
            end;
        end;
     end
  else  // fCharTransmitInterval > 0
     begin
     Result:=true;
     fLength:=System.Length(Source);
     if (fLength>CMaxBufferLength) then fLength:=CMaxBufferLength;
     for I:=1 to fLength do
         begin
         if PutString(Source[I]) then
            begin
            WaitMicroseconds(fCharTransmitDuration+fCharTransmitInterval*1000);
            if Assigned(fOnTransmitChar) then fOnTransmitChar;
            end
         else
            begin
            Result:=false;
            break;
            end;
         end;
     end;
end;

function TCustomComDevice.SimulateTransmission (Source: string): boolean;
var fLength,I: integer;
begin
  fLength:=System.Length(Source);

  if Assigned(fOnTransmitChar) then
     begin
     for I:=1 to fLength do
         begin
         WaitMicroseconds(fCharTransmitDuration+fCharTransmitInterval*1000);
         fOnTransmitChar;
         end;
     end
  else WaitMilliseconds((fCharTransmitDuration*fLength) DIV 1000);
  Result:=true;
end;

function TCustomComDevice.CharsInQue: word;
begin
  Result:=GetCOMCharsInQue(fComHandle);
end;

function TCustomComDevice.CharsOutQue: word;
begin
  Result:=GetCOMCharsOutQue(fComHandle);
end;

function TCustomComDevice.GetChar(var Target: char): boolean;
var C: char;
begin
  if (fReadTimeOut=0) then
     Result:=GetCharFromCOM(fComHandle,C)
  else Result:=GetCharFromCOM(fComHandle,C,fSmallTimeOut,fReadTimeOut);
  if Result then Target:=C;
end;

function TCustomComDevice.GetString(CharCount: byte): string;
begin
  if (fReadTimeOut=0) then
     Result:=GetStringFromCOM(fComHandle,CharCount,fBreakReadOnEOS,fEOSChar)
  else Result:=GetStringFromCOM(fComHandle,CharCount,fBreakReadOnEOS,fEOSChar,fSmallTimeOut,fReadTimeOut);
end;

function TCustomComDevice.EvaluateComSettings: TFComSettings;
var lComSettings: TFComSettings;
begin
  with lComSettings do
       begin
       Number:=fComNumber;
       BaudRate:=fBaudRate;
       ByteSize:=fByteSize;
       StopBits:=fStopBits;
       Parity:=fParityMode;
       Flags:=0;
       // create DCB Flag items
       case fDTRControl of
            DTR_CONTROL_ENABLE:    Flags:=Flags or dcb_DtrControlEnable;
            DTR_CONTROL_HANDSHAKE: Flags:=Flags or dcb_DtrControlHandshake;
            end;
       case fRTSControl of
            RTS_CONTROL_ENABLE:    Flags:=Flags or dcb_RtsControlEnable;
            RTS_CONTROL_HANDSHAKE: Flags:=Flags or dcb_RtsControlHandshake;
            RTS_CONTROL_TOGGLE:    Flags:=Flags or dcb_RtsControlToggle;
            end;
       case fOutputControlMode of
            ocCtsOnly: Flags:=Flags or dcb_OutxCtsFlow;
            ocDsrOnly: Flags:=Flags or dcb_OutxDsrFlow;
            ocCtsDsr:  Flags:=Flags or dcb_OutxCtsFlow or dcb_OutxDsrFlow;
            end;
       if fInputDSRSensivity then Flags:=Flags or dcb_DsrSensitivity;
       case fXonXoffControlMode of
            xcInputOnly:   Flags:=Flags or dcb_InX;
            xcOutputOnly:  Flags:=Flags or dcb_OutX;
            xcInputOutput: Flags:=Flags or dcb_InX or dcb_OutX;
            end;
       if (fXonXoffControlMode<>xcNone) then
          begin
          XonChar:=fXonChar;
          XoffChar:=fXoffChar;
          end;
       EOSChar:=fEOSChar;
       end;
  Result:=lComSettings;
end;

constructor TCustomComDevice.Create (AOwner: TComponent);
begin
  inherited Create (AOwner);

  fComNumber:=-1;

  fComHandle:=INVALID_HANDLE_VALUE;
  fSmallTimeOut := TTimeOutController.Create;
  fLargeTimeOut := TTimeOutController.Create;

  fBaudRate:=CDefaultComBaudRate;
  fByteSize:=DATABITS_8;  // 8
  fStopBits:=ONESTOPBIT;  // 1
  fCharTransmitDuration:=(1000000*(fByteSize+fStopBits+1)) DIV fBaudRate;

  fParityMode:=NOPARITY;

  fEOSChar:=CLF;
  fXonChar:=CXon;
  fXoffChar:=CXoff;

  fDTRControl:=DTR_CONTROL_DISABLE;
  fRTSControl:=RTS_CONTROL_DISABLE;
  fOutputControlMode:=ocNone;
  fXonXoffControlMode:=xcNone;

  fInputDSRSensivity:=false;

  fAppendEOSonWrite:=false;
  fBreakReadOnEOS:=false;
  fRemoveEOSonRead:=false;

  fValidateBaudRate:=false;

  fSimulationMode:=false;
  fThreadActive:=false;

  fComSimulationString:='';

  fCharTransmitInterval:=0;
  fCharTransmitDuration:=0;

  fWriteTimeOut:=0;
  fReadTimeOut:=0;
  fMaxStringLength:=CDefaultMaxStringLength;

{$IFDEF WithThread}
  fComReceiver:=TComReceiver.Create;
  with fComReceiver do
       begin
       FreeOnTerminate:=true;
       Priority:=tpLower;

       pComHandle:=@fComHandle;
       pBreakReadOnEOS:=@fBreakReadOnEOS;
       pRemoveEOSOnRead:=@pRemoveEOSOnRead;
       pSimulationMode:=@fSimulationMode;

       pEOSchar:=@fEOSchar;

       pReadTimeOut:=@fReadTimeOut;
       pMaxStringLength:=@fMaxStringLength;
       pReceptionString:=@fComReceptionString;
       pSimulationString:=@fComSimulationString;

       OnThreadReceiveChar:=nil;
       OnThreadReceptionFinished:=nil;

       OnThreadReadChar:=nil;
       OnThreadReadString:=nil;
       OnThreadReceptionTimeOut:=nil;
       // OnThreadReceptionReview:=nil;
       OnThreadReadTermination:=nil;
       end;

  fThreadActive:=true;

  fComReceptionString:='';
{$ENDIF}
end;

destructor TCustomComDevice.Destroy;
begin
  if (GetComOpenStatus) then CloseHandle(fComHandle);
  fSmallTimeOut.Free;
  fLargeTimeOut.Free;
  inherited Destroy;
end;

function TCustomComDevice.Open: boolean;
begin
  if (fComNumber>0) then
     begin
     fComHandle:=OpenSerialPort(EvaluateComSettings);
     fSimulationMode:=false;
     Result:=GetComOpenStatus;
     end
  else
     begin
     fSimulationMode:=true;
     Result:=true;
     end;

{$IFDEF WithThread}
  if Result then
     begin
     with fComReceiver do
          begin
          if Assigned(fOnReceiveChar) then
             OnThreadReceiveChar:=OnReceiveChar;
          if Assigned(fOnReceptionFinished) then
             OnThreadReceptionFinished:=OnReceptionFinished;

          if Assigned(fOnReadChar) then
             OnThreadReadChar:=OnReadChar;
          if Assigned(fOnReadString) then
             OnThreadReadString:=OnReadString;

          if Assigned(fOnReceptionTimeOut) then
             OnThreadReceptionTimeOut:=OnReceptionTimeOut;
          {
          if Assigned(fOnReceptionReview) then
             OnThreadReceptionReview:=OnReceptionReview;
          }
          
          if Assigned(fOnReadTermination) then
             OnThreadReadTermination:=OnReadTermination;

          Resume;
          end;
     end;
{$ENDIF}
end;

procedure TCustomComDevice.Close;
begin
{$IFDEF WithThread}
  if not(fComReceiver.Suspended) then
     fComReceiver.Suspend;
{$ENDIF}
  fSmallTimeOut.Terminate;
  fLargeTimeOut.Terminate;
  CloseSerialPort(fComHandle);
  fSimulationMode:=false;
end;

procedure TCustomComDevice.PurgeReadBuffer;
begin
  if not(csDesigning in ComponentState) then PurgeCOMReadBuffer(fComHandle);
end;

procedure TCustomComDevice.PurgeWriteBuffer;
begin
  if not(csDesigning in ComponentState) then PurgeCOMWriteBuffer(fComHandle);
end;

procedure TCustomComDevice.PurgeReadWriteBuffer;
begin
  if not(csDesigning in ComponentState) then PurgeCOMReadWriteBuffer(fComHandle);
end;

function TCustomComDevice.GetCTSandDSRLineStatus: TModemStatus;
begin
  Result:=GetCTSandDSRLines(fComHandle);
end;

function TCustomComDevice.WriteString(Source: string): boolean;
begin
  if fAppendEOSonWrite then Source:=Source+fEOSChar;

  if (not(fSimulationMode) and TransmitString(Source)) or
     (fSimulationMode and SimulateTransmission(Source)) then
     begin
     if (Assigned(fOnTransmissionFinished)) then fOnTransmissionFinished;
     Result:=true;
     end
  else Result:=false;
end;

function TCustomComDevice.ReadString(var Target: string; CharCount: byte): boolean;
var fTarget: string;
{$IFNDEF WithThread}
    fLength,I: integer;
    fC: char;
{$ENDIF}
begin
{$IFDEF WithThread}
  if (CharCount>0) and
     not(Assigned(fOnReadChar)) and
     not(Assigned(fOnReadString)) and
     not(Assigned(fOnReceptionTimeOut)) then
     begin

     fLargeTimeOut.Initialize(fReadTimeOut);
     while (fComReceptionString='') and not(fLargeTimeOut.Elapsed) do;

     fTarget:=fComReceptionString;
     fComReceptionString:='';
     Result:=true;
{$ELSE}
  if (CharCount>0) then
     begin
     if not(fSimulationMode) then
        begin
        if not(Assigned(fOnReceiveChar)) then
           begin
           fTarget:=GetString(CharCount);
           fLength:=System.Length(fTarget);
           end
        else // Assigned(fOnReceiveChar)
           begin
           fTarget:=''; fLength:=0;
           fLargeTimeOut.Initialize(fReadTimeOut);
           while not(fLargeTimeOut.Elapsed) do
                 begin
                 if GetChar(fC) then
                    begin
                    if Assigned(fOnReceiveChar) then fOnReceiveChar;
                    fTarget:=fTarget+fC;
                    fLength:=System.Length(fTarget);
                    if fBreakReadOnEOS and (fC=fEOSChar) then break;
                    end;
                 end;
            end;
        end
     else  // fSimulationMode
        begin
        if Assigned(fOnReadStringSimulation) then fOnReadStringSimulation(fTarget)
        else
           begin
           fTarget:=fComSimulationString;
           fComSimulationString:='';
           end;
        fLength:=System.Length(fTarget);
        if (Assigned(fOnReceiveChar)) then
           begin
           for I:=1 to fLength do
               begin
               WaitMilliseconds((fCharTransmitDuration DIV 1000)+1);
               if Assigned(fOnReceiveChar) then fOnReceiveChar;
               end;
           end;
        end;

     if fRemoveEOSonRead and (fLength>0) and (fTarget[fLength]=fEOSChar) then System.Delete(fTarget,fLength,1);
     if (fLength>0) and ((fLength=CharCount) or (fBreakReadOnEOS and (fTarget[fLength]=fEOSChar))) then
        begin
        if Assigned(fOnReceptionFinished) then fOnReceptionFinished;
        Result:=true;
        end
     else Result:=false;
{$ENDIF}
     end
  else
     begin
     fTarget:='';
     Result:=false;
     end;
  Target:=fTarget;
end;

function TCustomComDevice.QueryString(Question: string; BreakTime: integer; var Answer: string; CharCount: byte): boolean;
begin
  if WriteString(Question) then
     begin
     WaitMilliseconds(BreakTime);
     Result:=ReadString(Answer,CharCount);
     end
  else Result:=false;
end;

function TCustomComDevice.ForceReceiverSimulation(Value: string): boolean;
begin
  // Attention! don't use the function "ForceReceiverSimulation"
  // and assign Event "OnReadStringSimulation" in the same compilation
  // just as a simple rule: use "ForceReceiverSimulation" for slave
  // and "OnReadStringSimulation" for master operation.
  if fSimulationMode and
     not(Assigned(fOnReadStringSimulation)) then
     begin
     fComSimulationString:=Value;
     Result:=true;
     end
  else Result:=false;
end;

end.
