//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//*
//* ORDER DESCRIPTION - SERIAL COMMUNICATION
//* USEFULOBJECTS.PAS
//*
//* Handle and Event objects, mostly for serial communication but also for other
//* thread purposes
//*
//* Author : Georg  Hahn
//* (the first three objects derived from "CommObjs", VARIAN SOFTWARE SERVICES NL )
//* Datum  :  21.07.09 / 29.10.09
//*
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
unit UsefulObjects;
{$M+}
{$UNSAFECODE ON}

{$DEFINE WithThread}

interface

uses
  Windows, Classes;

type

  TEventWaitResult = (wrSignaled, wrTimeout, wrAbandoned, wrError);

  TEventObject = class(TObject)
  private
    fHandle: THandle;
    fError: integer;
  public
    constructor Create(EventAttributes: PSecurityAttributes; ManualReset,
      InitialState: boolean; const Name: string);
    destructor Destroy; override;
    function WaitFor(Timeout: dWord): TEventWaitResult;
    procedure SetEvent;
    procedure ResetEvent;
    property Handle: THandle read fHandle;
    property Error: Integer read fError;
  end;

  TSimpleEventObject = class(TEventObject)
  public
    constructor Create;
  end;

  TCriticalSection = class(TObject)
  private
    fSection: TRTLCriticalSection;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Enter;
    procedure Leave;
  end;

  TTimerPerformance = (tpNone, tpStandard, tpHighPerformance);

  TTimeOutController = class(TObject)
  private
    { Private declarations }
    fTimerPerformance,
    fActualPerformance: TTimerPerformance;
    fNextProcessMessagesTime,
    fFinishTime: int64;
    fProcessMessagesInterval: integer;
    fSystemResolutionInMilliseconds,
    fSystemCountForOneMicrosecond: real;
    fTimeOutPending: boolean;
    procedure SetTimerPerformance(Value: TTimerPerformance);
    procedure SetProcessMessagesInterval(Value: integer);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;

    function Initialize (TimeOut: word): boolean; overload;
    function Initialize (TimeOut: word; ActualPerformance: TTimerPerformance): boolean; overload;
    procedure Extend (TimeOut: word);
    procedure Terminate;
    function Elapsed: boolean;
  published
    { Published declarations }
    property TimerPerformance: TTimerPerformance read fTimerPerformance write SetTimerPerformance;
    property ProcessMessagesInterval: integer read fProcessMessagesInterval write SetProcessMessagesInterval;
  end;

  TSimpleEvent  = procedure of object;

  TRxCharRead  = procedure(RxChar: char) of object;
  TRxStringRead = procedure(RxString: string) of object;
  TRxStringRequest = procedure(var RxString: string) of object;

{$IFDEF WithThread}
  PReceiver = ^TReceiver;
  TReceiver = class(TThread)
  private
    { Private declarations }
    pReceiverHandle: PHandle;
    pBreakReadOnEOS,
    pRemoveEOSOnRead,
    pSimulationMode: pboolean;
    pEOSchar: pchar;
    pReadTimeOut,
    pMaxStringLength: pdword;
    pReceptionString,
    pSimulationString: pstring;

    fEnabled,
    fDisabled,
    fReceptionFinished,
    fReceptionTimeOut: boolean;
    fReceptionTimeOutLimit: dword;
    fThreadReceptionString: string;

    fOnThreadReceiveChar,
    fOnThreadReceptionFinished: TSimpleEvent;

    fOnThreadReadChar: TRxCharRead;
    fOnThreadReadString,
    fOnThreadReceptionTimeOut: TRxStringRead;
  protected
    { Protected-Deklarationen }
    function GetSimulatedChar(var Ch: char): boolean;
    function GetReceivedChar(var Ch: char): boolean; virtual; abstract;
    procedure Execute; override;
  public
    { Public-Deklarationen}
    constructor Create;
    destructor Destroy; override;
    procedure Resume;
    procedure Suspend;
  published
    { Published-Deklarationen }
    property OnThreadReceiveChar: TSimpleEvent read fOnThreadReceiveChar write fOnThreadReceiveChar;
    property OnThreadReceptionFinished: TSimpleEvent read fOnThreadReceptionFinished write fOnThreadReceptionFinished;

    property OnThreadReadChar: TRxCharRead read fOnThreadReadChar write fOnThreadReadChar;
    property OnThreadReadString: TRxStringRead read fOnThreadReadString write fOnThreadReadString;
    property OnThreadReceptionTimeOut: TRxStringRead read fOnThreadReceptionTimeOut write fOnThreadReceptionTimeOut;
  end;
{$ENDIF}

  TCurveType = (ctLine,ctRamp,ctExpotential,ctSinus);

  TSignalSimulator = class(TObject)
  private
    { Private declarations }
    fCurveType: TCurveType;
    fTimeLimit,
    fValueLimit,
    fNoisePercentage: real;
    procedure SetCurveType (Value: TCurveType);
    procedure SetTimeLimit (Value: real);
    procedure SetValueLimit (Value: real);
    // procedure SetNoisePercentage (Value: real);
  public
    { Public-Deklarationen }
    constructor Create;
    destructor Free;

    function GetValue(Time: real): real;
  published
    { Published-Deklarationen }
    property CurveType: TCurveType read fCurveType write SetCurveType default ctExpotential;
    property TimeLimit: real read fTimeLimit write SetTimeLimit;
    property ValueLimit: real read fValueLimit write SetValueLimit;
    property NoisePercentage: real read fNoisePercentage write fNoisePercentage;
  end;

const
  CDefaultReadTimeOut     = 1000;
  CDefaultMaxStringLength =  100;

implementation

uses SysUtils, Forms;

//TEventObject

constructor TEventObject.Create(EventAttributes: PSecurityAttributes; ManualReset,
  InitialState: boolean; const Name: string);
begin
  fHandle := CreateEvent(EventAttributes, ManualReset, InitialState, PChar(Name));
end;

destructor TEventObject.Destroy;
begin
  CloseHandle(fHandle);
  inherited Destroy;
end;

function TEventObject.WaitFor(Timeout: dWord): TEventWaitResult;
begin
  case WaitForSingleObject(fHandle, Timeout) of
    WAIT_ABANDONED:
      Result := wrAbandoned;
    WAIT_OBJECT_0:
      Result := wrSignaled;
    WAIT_TIMEOUT:
      Result := wrTimeout;
    WAIT_FAILED:
      begin
        Result := wrError;
        fError := GetLastError;
      end;
  else
    Result := wrError;
  end;
end;

procedure TEventObject.SetEvent;
begin
  Windows.SetEvent(fHandle);
end;

procedure TEventObject.ResetEvent;
begin
  Windows.ResetEvent(fHandle);
end;

//TSimpleEventObject

constructor TSimpleEventObject.Create;
begin
  fHandle := CreateEvent(nil, true, false, nil);
end;

// TCriticalSection

constructor TCriticalSection.Create;
begin
  inherited Create;
  InitializeCriticalSection(fSection);
end;

destructor TCriticalSection.Destroy;
begin
  DeleteCriticalSection(fSection);
  inherited Destroy;
end;

procedure TCriticalSection.Enter;
begin
  EnterCriticalSection(fSection);
end;

procedure TCriticalSection.Leave;
begin
  LeaveCriticalSection(fSection);
end;

// TTimeOutController

procedure TTimeOutController.SetTimerPerformance(Value: TTimerPerformance);
begin
  if not(fTimeOutPending) and (Value<>fTimerPerformance) then
     begin
     if ((Value=tpHighPerformance) and (fSystemCountForOneMicrosecond>0.0)) or
        ((Value=tpStandard) and (fSystemResolutionInMilliseconds>0.0)) or
        (Value=tpNone) then
        fTimerPerformance:=Value;
     end;
end;

procedure TTimeOutController.Terminate;
begin
  fTimeOutPending:=false;
end;

procedure TTimeOutController.SetProcessMessagesInterval(Value: integer);
begin
  if not(fTimeOutPending) and (Value<>fProcessMessagesInterval) and
     not(Value<50) and not(Value>1000) then
     fProcessMessagesInterval:=Value;
end;

constructor TTimeOutController.Create;
var   fta, fti: dword;
      ftad: longbool;
      mCounterFrequency, mProcessInitialisationTime: int64;
begin
  inherited;

  // Note: if high performance timer is available on the host system
  // then evaluate the count number per microsecond (as real value).
  // This number value will be then used to create required delay loops
  // by the functions "WaitMicrosecond" and "WaitMilliseconds".
  if QueryPerformanceFrequency(mCounterFrequency) and
     QueryPerformanceCounter(mProcessInitialisationTime) then
     begin
     fSystemCountForOneMicrosecond:=mCounterFrequency/1000000.0;
     end
  else fSystemCountForOneMicrosecond:=0.0;

  // Note: if the standard (low performance) timer is available
  // then evaluate the system resolution in milliseconds (as real value).
  // Please note that the fti value is given in 100ns steps
  if GetSystemTimeAdjustment(fta, fti, ftad) then
     fSystemResolutionInMilliseconds:=fti/10000.0
  else fSystemResolutionInMilliseconds:=0.0;

  if (fSystemCountForOneMicrosecond>0.001) then
     fTimerPerformance:=tpHighPerformance
  else if (fSystemResolutionInMilliseconds>0.001) then
          fTimerPerformance:=tpStandard
       else fTimerPerformance:=tpNone;

  fProcessMessagesInterval:=250;
  fTimeOutPending:=false;
end;

destructor TTimeOutController.Destroy;
begin

  inherited;
end;

function TTimeOutController.Initialize(TimeOut: word): boolean;
var fStartTime: int64;
begin
  Result:=false;
  if (TimeOut>0) and
     ((fTimerPerformance=tpHighPerformance) or (fTimerPerformance=tpStandard)) then
     begin
     if (fTimerPerformance=tpHighPerformance) then
        begin
        if QueryPerformanceCounter(fStartTime) then
           begin
           fFinishTime:=fStartTime+System.Round(TimeOut*1000*fSystemCountForOneMicrosecond);
           fNextProcessMessagesTime:=fStartTime+fProcessMessagesInterval*1000;
           fTimeOutPending:=(fFinishTime>fStartTime);
           end;
        end
     else if (fTimerPerformance=tpStandard) and
             (TimeOut>System.Round(fSystemResolutionInMilliseconds)) then
             begin
             fStartTime:=GetTickCount; fFinishTime:=fStartTime+TimeOut;
             fNextProcessMessagesTime:=fStartTime+fProcessMessagesInterval;
             fTimeOutPending:=(fFinishTime>fStartTime);
             end;
     fActualPerformance:=fTimerPerformance;
     Result:=fTimeOutPending;
     end;
end;

function TTimeOutController.Initialize(TimeOut: word; ActualPerformance: TTimerPerformance): boolean;
var fStartTime: int64;
begin
  Result:=false;
  if (TimeOut>0) and
     (((ActualPerformance=tpHighPerformance) and (fSystemCountForOneMicrosecond>0.001)) or
      ((ActualPerformance=tpStandard) and (fSystemResolutionInMilliseconds>0.001))) then
     begin
     if (ActualPerformance=tpHighPerformance) then
        begin
        if QueryPerformanceCounter(fStartTime) then
           begin
           fFinishTime:=fStartTime+System.Round(TimeOut*1000*fSystemCountForOneMicrosecond);
           fNextProcessMessagesTime:=fStartTime+fProcessMessagesInterval*1000;
           fTimeOutPending:=(fFinishTime>fStartTime);
           end;
        end
     else if (ActualPerformance=tpStandard) and
             (TimeOut>System.Round(fSystemResolutionInMilliseconds)) then
             begin
             fStartTime:=GetTickCount; fFinishTime:=fStartTime+TimeOut;
             fNextProcessMessagesTime:=fStartTime+fProcessMessagesInterval;
             fTimeOutPending:=(fFinishTime>fStartTime);
             end;
     fActualPerformance:=ActualPerformance;
     Result:=fTimeOutPending;
     end;
end;

procedure TTimeOutController.Extend(TimeOut: word);
begin
  if (TimeOut>0) and fTimeOutPending and
     ((fTimerPerformance=tpHighPerformance) or (fTimerPerformance=tpStandard)) then
     begin
     if (fActualPerformance=tpHighPerformance) then
        fFinishTime:=fFinishTime+System.Round(TimeOut*1000*fSystemCountForOneMicrosecond)
     else if (fActualPerformance=tpStandard) then
             fFinishTime:=fFinishTime+TimeOut;
    end;
end;

function TTimeOutController.Elapsed: boolean;
var fStartTime: int64;
begin
  Result:=true;
  if fTimeOutPending and ((fTimerPerformance=tpHighPerformance) or (fTimerPerformance=tpStandard)) then
     begin
     if (fActualPerformance=tpHighPerformance) then
        begin
        if QueryPerformanceCounter(fStartTime) and (fFinishTime>fStartTime) then
           begin
           if (fStartTime>fNextProcessMessagesTime) then
              begin
              fNextProcessMessagesTime:=fNextProcessMessagesTime+fProcessMessagesInterval*1000;
              Application.ProcessMessages;
              end;
           end
        else fTimeOutPending:=false;
        end
     else if (fActualPerformance=tpStandard) then
             begin
             fStartTime:=GetTickCount;
             if (fFinishTime>fStartTime) then
                begin
                if (fStartTime>fNextProcessMessagesTime) then
                   begin
                   fNextProcessMessagesTime:=fNextProcessMessagesTime+fProcessMessagesInterval;
                   Application.ProcessMessages;
                   end;
                end
             else fTimeOutPending:=false;
             end;
    Result:=not(fTimeOutPending);
    end;
end;

// TReceiver (Thread Object)

{$IFDEF WithThread}
constructor TReceiver.Create;
begin
  inherited Create(true);

  pReceiverHandle:=nil;
  pBreakReadOnEOS:=nil;
  pRemoveEOSOnRead:=nil;
  pSimulationMode:=nil;

  pEOSchar:=nil;
  pReadTimeOut:=nil;
  pMaxStringLength:=nil;
  pReceptionString:=nil;
  pSimulationString:=nil;

  fOnThreadReceiveChar:=nil;
  fOnThreadReceptionFinished:=nil;

  fOnThreadReadChar:=nil;
  fOnThreadReadString:=nil;
  fOnThreadReceptionTimeOut:=nil;
end;

destructor TReceiver.Destroy;
begin

  inherited;
end;

procedure TReceiver.Resume;
begin
  if not(fEnabled) then
     begin
     fReceptionFinished:=false;
     fReceptionTimeOut:=false;
     fThreadReceptionString:='';

     inherited Resume;
     fEnabled:=true;
     end;
end;

procedure TReceiver.Suspend;
begin
  if fEnabled then
     begin
     fEnabled:=false;
     while not(fDisabled) do;
     inherited Suspend;
     end;
end;

function TReceiver.GetSimulatedChar(var Ch: char): boolean;
begin
  if (pSimulationString<>nil) and
     (pSimulationString^<>'') then
     begin
     Ch:=pSimulationString^[1];
     System.Delete(pSimulationString^,1,1);
     Result:=true;
     end
  else Result:=false;
end;

procedure TReceiver.Execute;
var StringLength: word;
    NewChar: char;
begin
  StringLength:=0;

  while True do
        begin
        if fEnabled then
           begin
           if fDisabled then fDisabled:=false;

           if (pReceiverHandle<>nil) and
              (pSimulationMode<>nil) then
              begin

              if (not(pSimulationMode^) and (GetReceivedChar(NewChar))) or
                 (pSimulationMode^ and (GetSimulatedChar(NewChar))) then
                 begin

                 if Assigned(fOnThreadReceiveChar) then fOnThreadReceiveChar;

                 if Assigned(fOnThreadReadChar) then fOnThreadReadChar(NewChar)
                 else // not(Assigned(fOnThreadReadChar))
                    begin
                    if (fThreadReceptionString='') then
                       begin
                       fReceptionTimeOutLimit:=GetTickCount;
                       if (pReadTimeOut<>nil) then
                          fReceptionTimeOutLimit:=fReceptionTimeOutLimit+pReadTimeOut^
                       else fReceptionTimeOutLimit:=fReceptionTimeOutLimit+CDefaultReadTimeOut;
                       end;

                    fThreadReceptionString:=fThreadReceptionString+NewChar;
                    StringLength:=System.Length(fThreadReceptionString);

                    if (StringLength=10) then// wait more patiently for long strings in reception
                       begin
                       if (pReadTimeOut<>nil) then
                          fReceptionTimeOutLimit:=fReceptionTimeOutLimit+pReadTimeOut^
                       else fReceptionTimeOutLimit:=fReceptionTimeOutLimit+CDefaultReadTimeOut;
                       end;

                    fReceptionFinished:=((pBreakReadOnEOS<>nil) and (pBreakReadOnEOS^) and (pEOSchar<>nil) and (NewChar=pEOSChar^)) or
                                         ((pMaxStringLength<>nil) and not(StringLength<pMaxStringLength^));
                    end;
                 end
              else // no any char available, so control the time out
                 begin
                 if (fThreadReceptionString<>'') and not(fReceptionTimeOut) then
                    fReceptionTimeOut:=not(fReceptionTimeOutLimit>GetTickCount);
                 end;

              if fReceptionFinished or fReceptionTimeOut then
                 begin
                 if ((pRemoveEOSOnRead<>nil) and (pRemoveEOSOnRead^) and (pEOSchar<>nil) and (NewChar=pEOSChar^)) then
                    System.Delete(fThreadReceptionString,StringLength,1);

                 if Assigned(fOnThreadReceptionFinished) then fOnThreadReceptionFinished;

                 if not(fReceptionTimeOut) then
                    begin
                    if Assigned(fOnThreadReadString) then
                       fOnThreadReadString(fThreadReceptionString)
                    else if (pReceptionString<>nil) then
                            begin
                            while (pReceptionString^<>'') do; // wait ages until serviced
                            pReceptionString^:=fThreadReceptionString;
                            end;
                    end
                 else // ReceptionTimeOut
                    begin
                    if Assigned(fOnThreadReceptionTimeOut) then
                       fOnThreadReceptionTimeOut(fThreadReceptionString)
                    else if (pReceptionString<>nil) then
                            begin
                            while (pReceptionString^<>'') do; // wait ages until serviced
                            pReceptionString^:=fThreadReceptionString;
                            end;
                    end;

                 fReceptionFinished:=false;
                 fReceptionTimeOut:=false;
                 fThreadReceptionString:='';
                 end;
              end;
           end
        else // not(fEnabled)
           if not(fDisabled) then fDisabled:=true;
        end;
end;
{$ENDIF}

  // TSignalSimulator

constructor TSignalSimulator.Create;
begin
  inherited Create;
  fCurveType:=ctExpotential;

  fTimeLimit:=60.0;
  fValueLimit:=100.0;
  fNoisePercentage:=0.0;

  System.Randomize;
end;

destructor TSignalSimulator.Free;
begin
  inherited Free;
end;

procedure TSignalSimulator.SetCurveType(Value: TCurveType);
begin
  if (Value<>fCurveType) then
     fCurveType:=Value;
end;

{
procedure TSignalSimulator.SetNoisePercentage(Value: real);
begin
  if (Value<>fNoisePercentage) then
     fNoisePercentage:=Value;
end;
}

procedure TSignalSimulator.SetTimeLimit(Value: real);
begin
  if (Value<>fTimeLimit) then
     fTimeLimit:=Value;
end;

procedure TSignalSimulator.SetValueLimit(Value: real);
begin
  if (Value<>fValueLimit) then
     fValueLimit:=Value;
end;

function TSignalSimulator.GetValue(Time: real): real;
var lNoise: real;
begin
  Result:=0.0;
  if not(Time>fTimeLimit) then
     begin
     lNoise:=(2.0*System.Random-1.0)*fValueLimit*fNoisePercentage/100.0;
     case fCurveType of
          ctLine: Result:=fValueLimit+lNoise;
          ctRamp: Result:=fValueLimit*Time/fTimeLimit+lNoise;
          ctExpotential: Result:=fValueLimit*(1.0-exp(-5.0*Time/fTimeLimit))+lNoise;
          ctSinus: Result:=fValueLimit*Sin(2.0*pi*Time/fTimeLimit)+lNoise;
          end;
     end;
end;

end.
