//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//**
//**	ORDER DESCRIPTION - CONFIGURATION SOFTWARE WITH HART COMMUNICATION (PCS45)
//**	HARTDATATYPES.PAS
//**
//**	Description of data records for applications with HART communication (PCS45)
//**  (derived from the previous unit "DataTypes.pas")
//**
//**	Autor : Georg  Hahn
//**	Datum :  14.11.1996 / .... / 19.06.2009 / 10.11.2009 / 14.02.2013 / 12.04.2014 / 29.07.15
//**
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************

unit HartDataTypes;

interface

type

// The below data record definitions for universal (0 to 19) and common practice
// (33 to 112) HART commands are based on the Rosemount HART(TM) documentation
// HCF_SPEC-99, HCF_SPEC-127, HCF_SPEC-151 and HCF_SPEC-181.
// If possible, the names of the data records and the members are identical with
// that used in the referred document, otherwise they are at least are similar.

// The data record definitions for transmitter specific commands (128 to ...)
// are based on MTL documentation, which is referred in the individual section.
// However, the document HCF_SPEC-99 shall be always attended and considered.

// Note: commands using just one data type (byte, single etc.) are not listed here.
// It will be assumed, that this is not necessary and would rather complicate
// the understanding of the source text than effectively help.

//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to Universal Commands
//**  number 0 to 19 - HCF_SPEC-127
//*************************************************************************************************

  T24BitUnsigned = array [1..3] of byte;        // 24-bit unsigned integer

// Command 0 (and 11) - Read Identifier
  PHDeviceIdentifier = ^THDeviceIdentifier;
  THDeviceIdentifier = record
    DeviceTypeCode254       : byte;
    ManufacturerIdentCode   : byte;             // Note: Measurement Technology code is 64
    ManufacturersDeviceType : byte;
    NumberOfPreambles       : byte;
    RevisionLevelOfUCD      : byte;
    RevisionLevelOfTSD      : byte;
    SoftwareRevision        : byte;
    HardwareRevision        : byte;             // xxxxx.yyy (5 bit and 3 bit joined to 8 bit)
    Flags                   : byte;
    DeviceIdentNumber       : T24BitUnsigned;   // 24-bit unsigned integer (former "array [1..3] of byte")
    end;

// Command 1 - Read Primary Variable
  PHPrimaryVariable = ^THPrimaryVariable;
  THPrimaryVariable = record
    PrimaryUnit             : byte;
    PrimaryVariable         : single;
    end;

// Command 2 - Read Primary Variable
  THCurrentRangePercent = record
    LoopCurrent             : single;           // Primary Variable Current (in mA)
    PercentOfRange          : single;           // Primary Variable Percent of Range
    end;

// Command 3 - Read Dynamic Variables
  PHDynamicVariables = ^THDynamicVariables;
  THDynamicVariables = record
    LoopCurrent             : single;
    PrimaryUnit             : byte;
    PrimaryVariable         : single;           // Temperature, Frequency etc.
    SecondaryUnit           : byte;
    SecondaryVariable       : single;           // CJ Temperature (for temp. converter)
    // the structure can be extended by another unit (byte) and variable (single)
    end;

// The HART command 6 uses just one byte for request and response, therefore not listed

// The HART command 8 uses "string [24]" PACKED ASCII for response, therefore not listed

// Command 13 (and 18) - Read (and Write) Tag, Descriptior and Date
  PHTagDescriptorDate = ^THTagDescriptorDate;
  THTagDescriptorDate = record
    TagName                 : string[8];        // PACKED ASCII
    Descriptor              : string[16];       // PACKED ASCII
    Day                     : byte;
    Month                   : byte;
    Year                    : byte;             // AC - 1900 (e.g. 2007 - 1900 = 107)
    end;

// Command 14 - Read Primary Sensor Information
  PHPrimarySensorInformation = ^THPrimarySensorInformation; // HART command 14
  THPrimarySensorInformation = record
    SerialNumber            : T24BitUnsigned;   // 24-bit unsigned integer
    PrimaryUnit             : byte;
    UpperLimit              : single;
    LowerLimit              : single;
    MinumumSpan             : single;
    end;

// Command 15 - Read Primary Variable Output Information
  PHPrimaryVariableOutput = ^THPrimaryVariableOutput;
  THPrimaryVariableOutput = record
    PVAlarm                 : byte;
    TransferFunction        : byte;             // Linearisation, Reverse etc.
    PVUnits                 : byte;
    Point20mA               : single;
    Point4mA                : single;
    DampingTimeConstant     : single;
    end;

// Command 16 (and 19) - Read (and Write) Assembly Number
  PHFinalAssemblyNumber = ^THFinalAssemblyNumber;
  THFinalAssemblyNumber = record
    Number                  : T24BitUnsigned;   // 24-bit unsigned integer (former "array [1..3] of byte")
    end;

//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to Common Practice Commands
//**  number 33 to 112 - HCF_SPEC-151
//*************************************************************************************************

// The HART command 33 is not considered for use by MTL, therefore not listed

// The HART command 34 uses just one single value for request and response, therefore not listed

// Command 35 - Read Primary Variable Range
  THPrimaryVariableRange = record
    PrimaryUnit             : byte;
    UpperLimit              : single;
    LowerLimit              : single;
    end;

// The HART command 35 to 47 are not considered for use by MTL or they don't use any request data,
// or use even just one byte or single value and therefore are not listed

// Note: the data structure for command 48 was modified differing from the HCF_SPEC-151
// during the first use in 1993 by D.C. and this specification will be now used by MTL.
// Furthermore, in difference to all other commands, the communication and the device Status
// byte will be prepended to the transmitted string to allow detailed analysis.
// The StatusInfo string will be also truncated to the transmitter relevant length.

// Command 48 - Read Transmitter Status Information
  PHTransmitterStatus = ^THTransmitterStatus;
  THTransmitterStatus = record
    ComInfo,DevInfo         : byte;
    Status                  : array of char; // former definition was "array [1..10 (or 24)] of char;
    end;

// The HART command 49 to 112 are not considered for use by MTL or they don't use any request data,
// or use even just one byte or single value and therefore are not listed

//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to Transmitter Specific Commands
//**  number 128 to ... - HCF_SPEC-99
//*************************************************************************************************

//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to MTL.073 to MTL.07.XK (and successors)
//**  refer to the Test Procedure TP4073-2, appendix 3
//*************************************************************************************************

// Note: the command 128 for transmitter parameter interrogation (or 129 for configuration)
// uses selector byte request and the response is always just one data (byte or word).
// The below structure is the transmitter summary parameter interrogated (or configured)
// by multiple use of these commands (sequentiell with all selectors).
  PHx073ParameterSet= ^THx073ParameterSet;        // HART command 128 (and 129)
  THx073ParameterSet = record
    OpenCircuitAlarmLevel         : byte;       // selector 0
    TransmitterFailureAlarmLevel  : byte;       // selector 1
    FilterFrequencySelector       : byte;       // selector 2
    InputType                     : byte;       // selector 3
    SmoothingTimeConstant         : word;       // selector 4
    ColdJunctionCompensationOnOff : byte;       // selector 5
    ColdJunctionFailureAlarmLevel : byte;       // selector 70
    end;

// Command 200 - Read Primary Variables
  PHPrimaryVariables = ^THPrimaryVariables;
  THPrimaryVariables = record
    ScaledInput                   : single;
    SmoothedInput                 : single;
    ColdJunctionResistance        : single;
    ColdJunctionTemperature       : single;
    ColdJunctionCompensation      : single;
    ProcessVariable               : single;
    SensorFlags                   : byte;
    end;

// Command 201 - Read Secondary Variables
  PHSecondaryVariables = ^THSecondaryVariables;
  THSecondaryVariables = record
    ProcessVariableScaledInput    : single;
    DampedProcessVariable         : single;
    LoopCurrent                   : single;
    NoiseFloor                    : single;
    NotSignificant                : array [0..7] of byte;
    SensorFlags                   : byte;
    end;

// Command 202 - Read RAW Input Output Values
  PHRAWInputOutputValues = ^THRAWInputOutputValues;
  THRAWInputOutputValues = record
    DACValue                : SmallInt;         // this is signed 16-bit integer
    MainInput               : SmallInt;
    AuxInput                : SmallInt;
    MainExcitation          : SmallInt;
    AuxExcitation           : SmallInt;
    LeadOffTest1            : SmallInt;
    LeadOffTest2            : SmallInt;
    LeadOffTest5            : SmallInt;
    LeadOffTest6            : SmallInt;
    NotSignificant          : array [0..5] of byte;
    SensorFlags             : byte;
    end;

// Command 203 - Read Calibration Data
  PHCalibrationData = ^THCalibrationData;
  THCalibrationData = record
    RrefCal                 : single;
    Rin5Cal                 : single;
    Rin6Cal                 : single;
    VrefCal                 : single;
    CalFlag                 : byte;
    NotSignificant          : array [0..6] of byte;
    SensorFlags             : byte;
    end;

// The HART command 220 to 225 don't use any request data or use even just one byte or single value
// and therefore are not listed (covered by qualified one  byte or single value procedures).

//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to MTL4075 (and posibly to the successors)
//**  based on informations given from MCV by email (13/04/2005 and 21/06/2005)
//*************************************************************************************************

// Command 204 (and 205) - Read (and Write) Trip Parameters for the MTL4075 (and MTL4575)
  PHx075TripParameters = ^THx075TripParameters;
  THx075TripParameters = record
    TripValue               : single;
    Hysteresis              : single;
    TripOutputFunction      : byte;             //  0 = off, 1 = EBD, 2 = TRIP
    end;

//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to MTL4575 (and posibly to the successors)
//**  refer to the Test Procedure TP4575-3/5, appendix 2
//*************************************************************************************************

// The HART command 207 to 210 don't use any request or response data or use even just one byte
// and therefore are not listed


//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to the MTL4576
//**  refer to the communication protocoll description, appendix B
//*************************************************************************************************

// Command 151 (and 152) - Read (and Write) Input Parameters for the MTL4576
   PHInputParameters_x576 = ^THInputParameters_x576;
   THInputParameters_x576 = packed record
      Channel                   : byte;
      SensorType                : byte;
      PVUnits                   : byte;
      Linearisation             : byte;
      CJC_Option                : byte;
      Point4mA                  : single;
      Point20mA                 : single;
      CJC_UserValue             : single;
      end;

   THInputParameterArray_x576 = array [0..1] of THInputParameters_x576;

// Command 161 (and 162) - Read (and Write) Output Parameters for the MTL4576
   PHOutputParameters_x576 = ^THOutputParameters_x576;
   THOutputParameters_x576 = packed record
      Channel                   : byte;
      OpenCircuitAlarmLevel     : byte;
      ShortCircuitAlarmLevel    : byte;
      TransmitterFailureAlarmLevel  : byte;
      ColdJunctionFailureAlarmLevel : byte;
      NAMUR_MTL                 : byte;  // Namur = 0, MTL = 1
      DampingTimeConstant       : single;
      SmoothingTimeConstant     : single;
      EBD_Option                : byte;
      TripLimit                 : single;
      TripHysteresis            : single;
      end;

   THOutputParameterArray_x576 = array [0..1] of THOutputParameters_x576;

// Command 208 (and 207) - Read (and Write) Common Parameters for the MTL4576
   PHCommonParameters_x576 = ^THCommonParameters_x576;
   THCommonParameters_x576 = packed record
      response_time             : byte;  // Slow / Medium / Fast
      FilterFrequencySelector   : byte;  // 50Hz / 60Hz
      InputOption               : byte;
      end;

   THUserSensorHeader_x576 = record
      SensorType    : byte;
      TMIN          : single;
      TMAX          : single;
      TSEG          : single;
      end;

   THUserSensorTable = array [0..64] of word;

   // Command 209 (and 210) - Read (and Write) User Sensor Table for the MTL4576
   PHUserSensorData_x576 = ^THUserSensorData_x576;
   THUserSensorData_x576 = packed record
      SensorType    : byte;
      TMIN          : single;
      TMAX          : single;
      TSEG          : single;
      reserved      : byte;
      CoeffTable    : array [0..64] of word;
      end;

//*************************************************************************************************
//**  Data Records for Read and Write functions assigned to all MTL Frequency/Current Converters
//*************************************************************************************************

 // Command 138 (and 139) - Write (and Read) Pulse Divider Parameters for the E01003 and MTL4532
  PHDividerParameters = ^THDividerParameters;
  THDividerParameters = record
    FunctionSelect          : byte;
    Divider                 : single;
    Duration                : single;
    end;

// Command 140 (and 141) - Write (and Read) Trimm Range Parameters for the E01003 and MTL4532
  PHAdjustOutputParameters = ^THAdjustOutputParameters;
  THAdjustOutputParameters = record    // former named "TTrimmRangeParameters"
    AdjMode                 : byte;
    AdjUnits                : byte;
    OffsetValue             : single;
    GainValue               : single;
    end;

//*************************************************************************************************
//**  Data Records for Read and Write functions used for the Frequency Converter MTL4532
//*************************************************************************************************

// Command 131 (and 132) - Write (and Read) Input Configuration for the MTL4532
// The parameters associated with the commands 131 and 132 were configured
// using the "Read_One_Word" or "WriteRead_One_Word" procedures.
  PHInputConfiguration_x532 = ^THInputConfiguration_x532;
  THInputConfiguration_x532 = record
    InputSource             : byte;
    SignalInvert            : byte;
    end;

// Command 133 (and 134) - Write (and Read) Leak Flow Off Parameters for the MTL4532
  PHLeakFlowOffParameters_x532 = ^THLeakFlowOffParameters_x532;
  THLeakFlowOffParameters_x532 = record
    FunctionSelect          : byte;
    LFUnits                 : byte;
    LeakFlowOffValue        : single;
    LeakFlowOffCurrent      : single;
    end;

// Command 136 (and 137) - Write (and Read) Trip Parameters (A and B) for the MTL4532
  PHTripParameters_x532 = ^THTripParameters_x532;
  THTripParameters_x532 = record
    FunctionSelect          : byte;
    TPAUnits                : byte;
    TripAValue              : single;
    AHysteresis             : single;
    TPBUnits                : byte;
    TripBValue              : single;
    BHysteresis             : single;
    end;

//*************************************************************************************************
//**  Data Records for Read and Write functions used for the Eurocard Frequency Converter E01003
//*************************************************************************************************

// Command 131 (and 132) - Write (and Read) LFD and Input Configuration for the E01003
  PHInputAndLFDParameters_1003 = ^THInputAndLFDParameters_1003;
  THInputAndLFDParameters_1003 = record             // former named "TInputAndLFDParameters"
    FunctionSelect          : byte;
    end;

// Command 133 (and 134) - Write (and Read) Leak Flow Off Parameters for the E01003
  PHLeakFlowOffParameters_1003 = ^THLeakFlowOffParameters_1003;
  THLeakFlowOffParameters_1003 = record          // former named "TLowCutOffParameters"
    FunctionSelect          : byte;
    LeakFlowOffValue        : single;
    LeakFlowOffCurrent      : single;
    end;

// Command 136 (and 137) - Write (and Read) Trip Parameters (A or B) for the E01003
  PHTripParameters_1003 = ^THTripParameters_1003;
  THTripParameters_1003 = record                 // former named "TTripParameters"
    FunctionSelect          : byte;
    TPUnits                 : byte;
    TripValue               : single;
    Hysteresis              : single;
    end;

// Command 142 (and 143) - Write (and Read) Trimmer Parameters for the E01003
  PHTrimmerVariables_1003 = ^THTrimmerVariables_1003;
  THTrimmerVariables_1003 = record               // former named "TTrimmerVariables"
    OffsetValue             : word;
    GainValue               : word;
    end;

// Unlike the "LFD and Input Configuration" (associated with the commands 131 and 132),
// the just one byte data records for Hardware Configuration (command numbers 144 and 145)
// and for the Full Scale Option (commands 146 and 147) were not explicitely defined.
// These parameters were specified even as byte in the highest level record and configured
// using the "Read_One_Byte" or "WriteRead_One_Byte" procedures.

//*************************************************************************************************
//**  Complex Data Records assigned to different MTL converters
//*************************************************************************************************

// Basic data record, used for all MTL converter (and expected also by other manufacturer)
  PHBasicDeviceData = ^THBasicDeviceData;
  THBasicDeviceData = record
    UID     : THDeviceIdentifier;
    TDD     : THTagDescriptorDate;
    end;

// Master configuration data record, used for all single channel MTL converter
  PHMainDeviceData = ^THMainDeviceData;
  THMainDeviceData = record
    UID     : THDeviceIdentifier;
    TDD     : THTagDescriptorDate;
    PVO     : THPrimaryVariableOutput;
    end;

// Additional data record used for the temperature converter MTLx073 (and MTLx074, +XK)
  PHPVOandx073Variables = ^THPVOandx073Variables;
  THPVOandx073Variables = record
    PVO:     THPrimaryVariableOutput;
    x073Set: THx073ParameterSet;
    end;

// Additional data record used for the frequency converter MTLx532
  PHAddDeviceData_x532 = ^THAddDeviceData_x532;
  THAddDeviceData_x532 = record
    AOP     : THAdjustOutputParameters;
    LFP     : THLeakFlowOffParameters_x532;
    TransmitterFailureAlarmLevel  : byte;
    UnderOverRangeMode : byte;
    end;

// Additional data record used for the frequency converter E01003
  PHAddDeviceData_1003 = ^THAddDeviceData_1003;
  THAddDeviceData_1003 = record
    AOP     : THAdjustOutputParameters;
    LFP     : THLeakFlowOffParameters_1003;
    TransmitterFailureAlarmLevel : byte;
    FullScaleFlag : byte;
    end;

  PHLFDInputParameters_1003 = ^THLFDInputParameters_1003;
  THLFDInputParameters_1003 = record
    FunctionSelect : byte;
    end;

  PHFreqTripData_1003 = ^THFreqTripData_1003;
  THFreqTripData_1003 = record
    DVP     : THDividerParameters;
    TPA     : THTripParameters_1003;   //  Trip Point A
    TPB     : THTripParameters_1003;   //  Trip Point B
    end;

//*************************************************************************************************
//* Complete Configuration Set Data Records assigned to different MTL converters
//*************************************************************************************************

//* Temperature Converter MTLxy7z

  PHx073 = ^THx073;
  THx073 = record
    MDD     : THMainDeviceData;
    x073Set : THx073ParameterSet;
    Parity  : byte;
    end;

//* Temperature Converter MTL4075
  PHx075 = ^THx075;
  THx075 = record
    MDD     : THMainDeviceData;
    x073Set : THx073ParameterSet;
    x075TPA : THx075TripParameters;
    Parity  : byte;
    end;

//* Temperature Converter MTL4575
  PHx575 = ^THx575;
  THx575 = record
    MDD     : THMainDeviceData;
    x073Set : THx073ParameterSet;
    x075TPA : THx075TripParameters;
    Speed,
    NAMUR   : byte;
    Parity  : byte;
    end;

//* Temperature Converter MTL4573
  PHx573 = ^THx573;
  THx573 = record
    MDD     : THMainDeviceData;
    x073Set : THx073ParameterSet;
    x075TPA : THx075TripParameters;
    Speed,
    NAMUR,
    ShortCircuitAlarmLevel : byte; // selector 6
    Parity  : byte;
    end;

//* Frequency Converter MTL4532
  PHx532 = ^THx532;
  THx532 = record
    MDD     : THMainDeviceData;
    ADD     : THAddDeviceData_x532;
    TPP     : THTripParameters_x532;
    DVP     : THDividerParameters;
    InpConf : THInputConfiguration_x532;
    Parity  : byte;
    end;


//* Frequency Converter E01003 (newer version SwRev 4)
  PH1003 = ^TH1003;
  TH1003 = record
    MDD     : THMainDeviceData;
    ADD     : THAddDeviceData_1003;
    LFD_Inp : THLFDInputParameters_1003;
    FTD     : THFreqTripData_1003;
    Parity  : byte;
    end;

//* Temperature Converter MTL4576
  PHx576 = ^THx576;
  THx576 = record
    MDD     : THMainDeviceData;
    INPAR   : THInputParameterArray_x576;
    OUTPAR  : THOutputParameterArray_x576;
    COMMPAR : THCommonParameters_x576;
    // USEN: TUserSensorTable;
    Parity  : byte;
    end;

implementation

end.
