//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//**
//**	ORDER DESCRIPTION - CONFIGURATION SOFTWARE WITH HART COMMUNICATION (PCS45)
//**	HARTMODULES.PAS
//**
//**	Description of Module objects for applications with HART communication (PCS45)
//**  (derived from the Dave Posner units "ModuleObjects.pas" and "ModuleX576.pas")
//**
//**	Autor : Dave Posner and Georg  Hahn
//**	Datum :  xx.zz.2009 / .... / 17.04.2014
//**
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
unit HartModules;

interface

uses Types, Classes, SerialComDevices, HartDevices, DataTypes, HartDataTypes;

type
  TTransferDirection = (tdDownload,tdUpload);

  PBasicHartModule = ^TBasicHartModule;
  TBasicHartModule = class
  private
    fDeviceType: TDeviceType;
    lHartMaster: PHartMaster;
    fDeviceDataStream: TMemoryStream;
    fActualBasicDeviceData,
    fRequiredBasicDeviceData: THBasicDeviceData;
    fBasicDeviceDataAvailable,
    fSetTDD: boolean;

    procedure SeTHBasicDeviceData(Value: THBasicDeviceData);
    procedure ClearModuleSpecificDeviceDataStatusFlags; virtual;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; virtual;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; virtual;

    procedure DownloadActualModuleSpecificDeviceData; virtual;
    procedure UploadRequiredModuleSpecificDeviceData; virtual;
    function VerifyRequiredModuleSpecificDeviceData: boolean; virtual;

    procedure SynchronizeActualModuleSpecificDeviceData; virtual;
    procedure SynchronizeRequiredModuleSpecificDeviceData; virtual;

    procedure ReadModuleSpecificDaviceDataFromStream; virtual; abstract;
    procedure WriteModuleSpecificDeviceDataToStream; virtual; abstract;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; virtual;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; virtual;
    procedure SetModuleSpecificDeviceDataAvailableStatus; virtual;
  public
    constructor Create;
    destructor Destroy; override;

    procedure AssignHartMaster(var Value: THartMaster);
    procedure AssignReceivedCharEvent(Value: TSimpleEvent);
    procedure UnAssignReceivedCharEvent;

    function GetConnectedDeviceType: TDeviceType;
    function GetModuleTransferDataLength(Value: TTransferDirection): integer;

    procedure DownloadActualDeviceData;
    procedure UploadRequiredDeviceData;
    function VerifyDeviceData: boolean;

    procedure SynchronizeActualDeviceData;
    procedure SynchronizeRequiredDeviceData;

    function SetActualDeviceData(SourceStream: TMemoryStream): boolean;
    function GetActualDeviceDataStream: TMemoryStream;

    procedure ClearModuleDeviceDataAvailableStatus;
    procedure SetModuleDeviceDataAvailableStatus;

    procedure EvaluateRequiredDeviceDataStatusFlags;

    function ResetModule: boolean;
    function GeTHDynamicVariables(var Target: THDynamicVariables): boolean; virtual;
    function SetOutputCurrent(Value: single): boolean; virtual;
    function CalibratePoint20mA(Value: single): boolean; virtual;
    function CalibratePoint4mA(Value: single): boolean; virtual;
    function GeTHTransmitterStatus(var Target: THTransmitterStatus): boolean; virtual;

    property DeviceType: TDeviceType read fDeviceType;
    property BasicDeviceData: THBasicDeviceData read fActualBasicDeviceData write SeTHBasicDeviceData;
    property DeviceDataAvailableStatus: boolean read GetModuleSpecificDeviceDataAvailableStatus;
  end;

  TSingleChannelHartModule = class (TBasicHartModule)
  private
    fActualPrimaryVariableOutputData,
    fRequiredPrimaryVariableOutputData: THPrimaryVariableOutput;
    fPrimaryVariableOutputDataAvailable,
    fSetTransfer,fSetUnits,fSetRangeBlock,fSetDamping: boolean;

    procedure SeTHPrimaryVariableOutputData(Value: THPrimaryVariableOutput);
    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    property PrimaryVariableOutputData: THPrimaryVariableOutput read fActualPrimaryVariableOutputData write SeTHPrimaryVariableOutputData;
  end;

  THartModule_x073 = class (TSingleChannelHartModule)
  private
    fActualx073ParameterSet,
    fRequiredx073ParameterSet: THx073ParameterSet;
    fx073ParameterSetAvailable,
    fSetOCAL,fSetTFAL,fSetFFS,fSetIT,fSetSTC,fSetCJCOO,fSetCJFAL: boolean;

    procedure Set_x073_ParameterSet(Value: THx073ParameterSet);
    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    procedure ReadModuleSpecificDaviceDataFromStream; override;
    procedure WriteModuleSpecificDeviceDataToStream; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    // function Calibrate R5R6etc;

    property x073ParameterSet: THx073ParameterSet read fActualx073ParameterSet write Set_x073_ParameterSet;
  end;

  THartModule_x075 = class (THartModule_x073)
  private
    fActualx075TripParameters,
    fRequiredx075TripParameters: THx075TripParameters;
    fx075TripParametersAvailable,
    fSetx075Trip: boolean;

    procedure Set_x075_Parameters(Value: THx075TripParameters);
    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    procedure ReadModuleSpecificDaviceDataFromStream; override;
    procedure WriteModuleSpecificDeviceDataToStream; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    property x075TripParameters: THx075TripParameters read fActualx075TripParameters write Set_x075_Parameters;
  end;

  THartModule_x575 = class (THartModule_x075)
  private
    fActualResponseTime,
    fRequiredResponseTime,
    fActualNE43Mode,
    fRequiredNE43Mode: byte;
    fx575ParametersAvailable,
    fSetResponseTime, fSetNE43: boolean;

    procedure SetResponseTime(Value: byte);
    procedure SetNE43(Value: byte);
    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    procedure ReadModuleSpecificDaviceDataFromStream; override;
    procedure WriteModuleSpecificDeviceDataToStream; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    property ResponseTime: byte read fActualResponseTime write SetResponseTime;
    property NE43Mode: byte read fActualNE43Mode write SetNE43;
  end;

   THartModule_x573 = class (THartModule_x575)
  private
    fActualShortCircuitAlarmLevel,
    fRequiredShortCircuitAlarmLevel: byte;
    fx573ParametersAvailable,
    fSetShortCircuitAlarmLevel: boolean;

    procedure SetShortCircuitAlarmLevel(Value: byte);
    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    procedure ReadModuleSpecificDaviceDataFromStream; override;
    procedure WriteModuleSpecificDeviceDataToStream; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    property ShortCircuitAlarmLevel: byte read fActualShortCircuitAlarmLevel write SetShortCircuitAlarmLevel;
  end;

  THartModule_Freq = class (TSingleChannelHartModule)
  private
    fActualFullScale,
    fRequiredFullScale,
    fActualTransmitterFailureAlarmLevel,
    fRequiredTransmitterFailureAlarmLevel: byte;
    fActualAdjustOutputParameters,
    fRequiredAdjustOutputParameters: THAdjustOutputParameters;

    fActualDividerParameters,
    fRequiredDividerParameters: THDividerParameters;

    fxFreqParametersAvailable,
    fSetAdjust, fSetTFAL,
    fSetFullScale, fSetDivider,
    fClearTFAL, fClearFullScale,
    fClearDivider: boolean;

    procedure SetTransmitterFailureAlarmLevel(Value: byte);
    procedure SeTHAdjustOutputParameters(Value: THAdjustOutputParameters);

    procedure SetFullScale(Value: byte);
    procedure SeTHDividerParameters(Value: THDividerParameters);

    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    function SetDigitalDiagnostic(Selector: byte): boolean; overload; virtual; abstract;
    function SetDigitalDiagnostic(Selector: byte; Value: single): boolean; overload; virtual; abstract;

    function GetInternalValues(Selector: byte; var Target: TByteDynArray): boolean;

    function ResetConfigurationFlag: boolean;
    function ResetFactoryConfiguration: boolean;

    property TransmitterFailureAlarmLevel: byte read fActualTransmitterFailureAlarmLevel write SetTransmitterFailureAlarmLevel;
    property AdjustOutputParameters: THAdjustOutputParameters read fActualAdjustOutputParameters write SeTHAdjustOutputParameters;

    property FullScale: byte read fActualFullScale write SetFullScale;
    property DividerParameters: THDividerParameters read fActualDividerParameters write SeTHDividerParameters;
  end;


  THartModule_x532 = class (THartModule_Freq)
  private
    fActualNE43Mode,
    fRequiredNE43Mode: byte;
    fActualLeakFlowOffParameters,
    fRequiredLeakFlowOffParameters: THLeakFlowOffParameters_x532;
    fActualInputConfig,
    fRequiredInputConfig: THInputConfiguration_x532;
    fActualTripParameters,
    fRequiredTripParameters: THTripParameters_x532;

    fx532ParametersAvailable,
    fSetNE43,fSetLeakFlow,
    fSetInputConfig, fSetTrip,
    fClearNE43,fClearLeakFlow,
    fClearInputConfig, fClearTrip: boolean;

    procedure SetNE43(Value: byte);
    procedure SetLeakFlowOffParameters(Value: THLeakFlowOffParameters_x532);
    procedure SetInputConfig(Value: THInputConfiguration_x532);
    procedure SetTripParameters(Value: THTripParameters_x532);

    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    procedure ReadModuleSpecificDaviceDataFromStream; override;
    procedure WriteModuleSpecificDeviceDataToStream; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    function GetAssemblyNumber(var Target: dword): boolean;
    function SetAssemblyNumber(Source: dword): boolean;

    function SetDigitalDiagnostic(Selector: byte): boolean; overload; override;
    function SetDigitalDiagnostic(Selector: byte; Value: single): boolean; overload; override;

    property NE43Mode: byte read fActualNE43Mode write SetNE43;

    property LeakFlowOffParameters: THLeakFlowOffParameters_x532 read fActualLeakFlowOffParameters write SeTLeakFlowOffParameters;
    property InputConfiguration: THInputConfiguration_x532 read fActualInputConfig write SetInputConfig;
    property TripParameters: THTripParameters_x532 read fActualTripParameters write SetTripParameters;
  end;

  THartModule_1003 = class (THartModule_Freq)
  private
    fActualInputAndLFDMode,
    fRequiredInputAndLFDMode: byte;
    fActualLeakFlowOffParameters,
    fRequiredLeakFlowOffParameters: THLeakFlowOffParameters_1003;
    fActualTripAParameters,
    fActualTripBParameters,
    fRequiredTripAParameters,
    fRequiredTripBParameters: THTripParameters_1003;
    fActualHardwareCircuitry: byte;
    f1003ParametersAvailable,
    fSetLFDAndInputMode, fSetLeakFlow,
    fSetTripA, fSetTripB,
    fClearLFDAndInputMode, fClearDivider,
    fClearTripA, fClearTripB: boolean;

    procedure SetInputAndLFDMode(Value: byte);
    procedure SetLeakFlowOffParameters(Value: THLeakFlowOffParameters_1003);
    procedure SetTripAParameters(Value: THTripParameters_1003);
    procedure SetTripBParameters(Value: THTripParameters_1003);

    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    procedure ReadModuleSpecificDaviceDataFromStream; override;
    procedure WriteModuleSpecificDeviceDataToStream; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    function SetDigitalDiagnostic(Selector: byte): boolean; overload; override;
    function SetDigitalDiagnostic(Selector: byte; Value: single): boolean; overload; override;

    function GetHardwareCircuitryConfiguration(var HardConfig: byte): boolean;
    function SetHardwareCircuitryConfiguration(HardConfig: byte): boolean;

    function GetTrimmerInformation(Selector: byte; var Target: THTrimmerVariables_1003): boolean;
    function SetTrimmerInformation(Selector: byte; Source: THTrimmerVariables_1003): boolean;

    property HardwareCircuitry: byte read fActualHardwareCircuitry;
    property InputAndLFDMode: byte read fActualInputAndLFDMode write SetInputAndLFDMode;
    property LeakFlowOffParameters: THLeakFlowOffParameters_1003 read fActualLeakFlowOffParameters write SeTLeakFlowOffParameters;
    property TripAParameters: THTripParameters_1003 read fActualTripAParameters write SetTripAParameters;
    property TripBParameters: THTripParameters_1003 read fActualTripBParameters write SetTripBParameters;
  end;

  TMultiChannelHartModule = class (TBasicHartModule)
  private
    fActualChannel,
    fMaxChannels: byte;
    procedure SetChannelNumber(Value: byte);
    procedure SetMaxChannelCount(Value: byte);
  protected
  public
    constructor Create;
    destructor Destroy; override;

    property ChannelNumber: byte read fActualChannel write SetChannelNumber;
    property MaxChannelCount: byte read fMaxChannels write SetMaxChannelCount;
  end;

  THartModule_x576 = class (TMultiChannelHartModule)
  private
    fActualInputParameters,
    fRequiredInputParameters: THInputParameterArray_x576;
    fActualOutputParameters,
    fRequiredOutputParameters: THOutputParameterArray_x576;
    fActualCommonParameters,
    fRequiredCommonParameters: THCommonParameters_x576;
    fActualUserSensorHeader,
    fRequiredUserSensorHeader: THUserSensorHeader_x576;
    fx576ParametersAvailable,
    fSetInputs0, fSetInputs1, fSetOutputs0, fSetOutputs1, fSetCommons: boolean;

    procedure SetInputParameters(Value: THInputParameterArray_x576);
    procedure SetOutputParameters(Value: THOutputParameterArray_x576);
    procedure SetCommonParameters(Value: THCommonParameters_x576);
    procedure SetUserSensorHeader(Value: THUserSensorHeader_x576);

    procedure ClearModuleSpecificDeviceDataStatusFlags; override;
    procedure EvaluateModuleSpecificDeviceDataStatusFlags; override;
  protected
    function GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer; override;

    procedure DownloadActualModuleSpecificDeviceData; override;
    procedure UploadRequiredModuleSpecificDeviceData; override;
    function VerifyRequiredModuleSpecificDeviceData: boolean; override;

    procedure SynchronizeActualModuleSpecificDeviceData; override;
    procedure SynchronizeRequiredModuleSpecificDeviceData; override;

    procedure ReadModuleSpecificDaviceDataFromStream; override;
    procedure WriteModuleSpecificDeviceDataToStream; override;

    function GetModuleSpecificDeviceDataAvailableStatus: boolean; override;

    procedure ClearModuleSpecificDeviceDataAvailableStatus; override;
    procedure SetModuleSpecificDeviceDataAvailableStatus; override;
  public
    constructor Create;
    destructor Destroy; override;

    function GeTHDynamicVariables(var Target: THDynamicVariables): boolean; override;
    function SetOutputCurrent(Value: single): boolean; override;
    function CalibratePoint4mA(Value: single): boolean; override;
    function CalibratePoint20mA(Value: single): boolean; override;
    function GeTHTransmitterStatus(var Target: THTransmitterStatus): boolean; override;

    function DownloadUserSensorTableCoefficients(var Target: THUserSensorTable): boolean;
    function UploadUserSensorData(Source: THUserSensorData_x576): boolean;

    property InputParameters: THInputParameterArray_x576 read fActualInputParameters write SetInputParameters;
    property OutputParameters: THOutputParameterArray_x576 read fActualOutputParameters write SetOutputParameters;
    property CommonParameters: THCommonParameters_x576 read fActualCommonParameters write SetCommonParameters;
    property UserSensorHeader: THUserSensorHeader_x576 read fActualUserSensorHeader write SetUserSensorHeader;
  end;

  procedure GetSpecificHartModule(DevType: TDeviceType; var Target: TBasicHartModule);
  procedure ChangeTypeHartModule(DevType: TDeviceType; var Target: TBasicHartModule);

implementation

uses Dialogs, ProcessUtils, ConversionUtils, HartCommands, HartDataConstants, DataInstances,
     DataConstants, ConversionProcs, ComparisonProcs, LabelNumbers, HelpContextNumbers;

procedure GetSpecificHartModule(DevType: TDeviceType; var Target: TBasicHartModule);
var lBDD: THBasicDeviceData;
begin
  if (Target<>nil) and
     (IdentToType(Target.BasicDeviceData.UID)<>DevType) then  // why is this ?
     begin
     Target.Free;
     Target:=nil;
     end;

  if (Target=nil) then
     begin
     case DevType of
          dt507x,
          dtx073,
          dt5074XK,
          dt4073XK: THartModule_x073(Target):=THartModule_x073.Create;
          dt4075:   THartModule_x075(Target):=THartModule_x075.Create;
          dtx575:   THartModule_x575(Target):=THartModule_x575.Create;
          dtx573:   THartModule_x575(Target):=THartModule_x573.Create;
          dtx576:   THartModule_x576(Target):=THartModule_x576.Create;
          dtx532:   THartModule_x532(Target):=THartModule_x532.Create;
          dt1003:   THartModule_1003(Target):=THartModule_1003.Create;
          else      Target:=TBasicHartModule.Create;
          end;
     Target.fDeviceType:=DevType;
     case DevType of
          dt507x,
          dtx073,
          dt5074XK,
          dt4073XK: begin
                    if not(DevType in [dt5074XK,dt4073XK]) then
                       lBDD.UID:=CDef_UID_x073
                    else lBDD.UID:=CDef_UID_x073XK;
                    lBDD.TDD:=CDef_TDD;
                    lBDD.UID.ManufacturersDeviceType:=CDeviceIdentifiers[DevType];
                    Target.fActualBasicDeviceData:=lBDD;
                    TSingleChannelHartModule(Target).fActualPrimaryVariableOutputData:=CDef_PVO_x073;
                    THartModule_x073(Target).fActualx073ParameterSet:=CDef_x073_THC_Set;
                    end;
          dt4075:   begin
                    lBDD.UID:=CDef_UID_x075;
                    lBDD.TDD:=CDef_TDD;
                    Target.fActualBasicDeviceData:=lBDD;
                    TSingleChannelHartModule(Target).fActualPrimaryVariableOutputData:=CDef_PVO_x073;
                    THartModule_x073(Target).fActualx073ParameterSet:=CDef_x073_THC_Set;
                    THartModule_x075(Target).fActualx075TripParameters:=CDef_x075_Trip;
                    end;
          dtx575:   begin
                    lBDD.UID:=CDef_UID_x575;
                    lBDD.TDD:=CDef_TDD;
                    Target.fActualBasicDeviceData:=lBDD;
                    TSingleChannelHartModule(Target).fActualPrimaryVariableOutputData:=CDef_PVO_x073;
                    THartModule_x073(Target).fActualx073ParameterSet:=CDef_x073_THC_Set;
                    THartModule_x075(Target).fActualx075TripParameters:=CDef_x075_Trip;
                    THartModule_x575(Target).fActualResponseTime:=CHart_x575_Slow;
                    THartModule_x575(Target).fActualNE43Mode:=CHart_x575_NE43_Off;
                    end;
          dtx573:   begin
                    lBDD.UID:=CDef_UID_x573;
                    lBDD.TDD:=CDef_TDD;
                    lBDD.UID.ManufacturersDeviceType:=CDeviceIdentifiers[Target.DeviceType];
                    Target.fActualBasicDeviceData:=lBDD;
                    TSingleChannelHartModule(Target).fActualPrimaryVariableOutputData:=CDef_PVO_x073;
                    THartModule_x073(Target).fActualx073ParameterSet:=CDef_x073_THC_Set;
                    THartModule_x075(Target).fActualx075TripParameters:=CDef_x075_Trip;
                    THartModule_x575(Target).fActualResponseTime:=CHart_x575_Slow;
                    THartModule_x575(Target).fActualNE43Mode:=CHart_x575_NE43_Off;
                    THartModule_x573(Target).fActualShortCircuitAlarmLevel:=CHart_Option_Off;
                    end;
          dtx576:   begin
                    lBDD.UID:=CDef_UID_x576;
                    lBDD.TDD:=CDef_TDD;
                    Target.fActualBasicDeviceData:=lBDD;
                    THartModule_x576(Target).fActualInputParameters[0]:=CDef_x576_InPar;
                    THartModule_x576(Target).fActualInputParameters[1]:=CDef_x576_InPar;
                    THartModule_x576(Target).fActualInputParameters[1].Channel:=1;
                    THartModule_x576(Target).fActualOutputParameters[0]:=CDef_x576_OutPar;
                    THartModule_x576(Target).fActualOutputParameters[1]:=CDef_x576_OutPar;
                    THartModule_x576(Target).fActualOutputParameters[1].Channel:=1;
                    THartModule_x576(Target).fActualCommonParameters:=CDef_x576_Common;
                    //THartModule_x576(Target).fActualUserSensoData:=
                    end;
          dtx532:   begin
                    lBDD.UID:=CDef_UID_x532;
                    lBDD.TDD:=CDef_TDD;
                    Target.fActualBasicDeviceData:=lBDD;
                    TSingleChannelHartModule(Target).fActualPrimaryVariableOutputData:=CDef_PVO_x532;
                    THartModule_Freq(Target).fActualTransmitterFailureAlarmLevel:=CHart_Alarm_Off;
                    THartModule_Freq(Target).fActualFullScale:=CHart_FullScale_Off;
                    THartModule_Freq(Target).fActualAdjustOutputParameters:=CDef_AdjustOutput;
                    THartModule_Freq(Target).fActualDividerParameters:=CDef_Divider;
                    THartModule_x532(Target).fActualNE43Mode:=CHart_x575_NE43_Off;
                    THartModule_x532(Target).fActualLeakFlowOffParameters:=CDef_x532_LeakOff;
                    THartModule_x532(Target).fActualInputConfig.InputSource:=CP_InputProxy1;
                    THartModule_x532(Target).fActualInputConfig.SignalInvert:=CP_Blank_Pattern;
                    THartModule_x532(Target).fActualTripParameters:=CDef_x532_Trip;
                    end;
          dt1003:   begin
                    lBDD.UID:=CDef_UID_1003;
                    lBDD.TDD:=CDef_TDD;
                    Target.fActualBasicDeviceData:=lBDD;
                    TSingleChannelHartModule(Target).fActualPrimaryVariableOutputData:=CDef_PVO_x532;
                    THartModule_Freq(Target).fActualTransmitterFailureAlarmLevel:=CHart_Alarm_Off;
                    THartModule_Freq(Target).fActualFullScale:=CHart_FullScale_Off;
                    THartModule_Freq(Target).fActualAdjustOutputParameters:=CDef_AdjustOutput;
                    THartModule_Freq(Target).fActualDividerParameters:=CDef_Divider;
                    THartModule_1003(Target).fActualInputAndLFDMode:=CP_Blank_Pattern;
                    THartModule_1003(Target).fActualLeakFlowOffParameters:=CDef_1003_LeakOff;
                    THartModule_1003(Target).fActualTripAParameters:=CDef_1003_TripA;
                    THartModule_1003(Target).fActualTripBParameters:=CDef_1003_TripB;
                    end;
          end;
     end;
end;

procedure ChangeTypeHartModule(DevType: TDeviceType; var Target: TBasicHartModule);
var lDeviceType: TDeviceType;
    lBDD: THBasicDeviceData;
    lTargetData: THx075;
begin
  if (Target<>nil) then
     begin
     lDeviceType:=IdentToType(Target.BasicDeviceData.UID);
     if (lDeviceType in [dt507x,dtx073,dt5074XK,dt4073XK,dt4075]) then
        begin
        lTargetData.MDD.TDD:=Target.BasicDeviceData.TDD;
        lTargetData.MDD.PVO:=TSingleChannelHartModule(Target).PrimaryVariableOutputData;
        lTargetData.x073Set:=THartModule_x073(Target).x073ParameterSet;

        if ((lDeviceType=dt4073XK) or(lDeviceType=dt5074XK)) and
           (lTargetData.x073Set.InputType=6) then // detect THC type XK
           lTargetData.x073Set.InputType:=20;     // and chnage adequately

        if (lDeviceType=dt4075) then
           lTargetData.x075TPA:=THartModule_x075(Target).x075TripParameters
        else lTargetData.x075TPA:=CDef_x075_Trip;

        Target.Free;
        Target:=nil;

        case DevType of
             dtx575:   THartModule_x575(Target):=THartModule_x575.Create;
             dtx573:   THartModule_x575(Target):=THartModule_x573.Create;
             end;
        Target.fDeviceType:=DevType;

        if (Target<>nil) then
           begin
           if (DevType=dtx573) then
              lBDD.UID:=CDef_UID_x573
           else lBDD.UID:=CDef_UID_x575;
           lBDD.TDD:=lTargetData.MDD.TDD;
           Target.fActualBasicDeviceData:=lBDD;
           TSingleChannelHartModule(Target).fActualPrimaryVariableOutputData:=lTargetData.MDD.PVO;
           THartModule_x073(Target).fActualx073ParameterSet:=lTargetData.x073Set;
           THartModule_x075(Target).fActualx075TripParameters:=lTargetData.x075TPA;
           THartModule_x575(Target).fActualResponseTime:=CHart_x575_Slow;
           THartModule_x575(Target).fActualNE43Mode:=CHart_x575_NE43_Off;
           if (DevType=dtx573) then
              THartModule_x573(Target).fActualShortCircuitAlarmLevel:=CHart_Option_Off;
           end;
        end;
     end;
end;
// ****************************************************************************

// TBasicHartModule

// private

procedure TBasicHartModule.SeTHBasicDeviceData(Value: THBasicDeviceData);
begin
  fRequiredBasicDeviceData:=Value;
  while (System.Length(fRequiredBasicDeviceData.TDD.TagName)<8) do
        fRequiredBasicDeviceData.TDD.TagName:=fRequiredBasicDeviceData.TDD.TagName+' ';
  while (System.Length(fRequiredBasicDeviceData.TDD.Descriptor)<16) do
        fRequiredBasicDeviceData.TDD.Descriptor:=fRequiredBasicDeviceData.TDD.Descriptor+' ';
end;

procedure TBasicHartModule.ClearModuleSpecificDeviceDataStatusFlags;
begin
  fSetTDD:=false;
end;

procedure TBasicHartModule.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  fSetTDD:=not(CompareTDDs(fRequiredBasicDeviceData.TDD,fActualBasicDeviceData.TDD));
end;

// protected

function TBasicHartModule.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  if (Value=tdDownload) then
     Result :=   8 + 12 // Command 0  - TDeviceIdentifier
               + 8 + 21 // Command 13 - TTagDescriptorDate
  else
     begin
     fDataLength:=0;
     if fSetTDD then fDataLength:=fDataLength + 8 + 21;  // Command 18 - TTagDescriptorDate
     Result:=fDataLength;
     end;
end;

procedure TBasicHartModule.DownloadActualModuleSpecificDeviceData;
begin
  if (lHartMaster<>nil) then
     fBasicDeviceDataAvailable:=(Read_Identifier(lHartMaster^,fActualBasicDeviceData.UID,true)) and
                                (Read_TagDescriptorDate(lHartMaster^,fActualBasicDeviceData.TDD,500,true))
  else fBasicDeviceDataAvailable:=false;
end;

procedure TBasicHartModule.UploadRequiredModuleSpecificDeviceData;
var lBasicDeviceData: THBasicDeviceData;
begin
  if (lHartMaster<>nil) then
     begin
     lBasicDeviceData:=fRequiredBasicDeviceData;
     if fSetTDD then
        if (WriteRead_TagDescriptorDate_Block (lHartMaster^, lBasicDeviceData.TDD, true)) then
           fActualBasicDeviceData.TDD:=lBasicDeviceData.TDD;
     end;
end;

function TBasicHartModule.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  // Don't compare the TDeviceIdentifier data
  Result:=(CompareTDDs(fRequiredBasicDeviceData.TDD,fActualBasicDeviceData.TDD));
end;

procedure TBasicHartModule.SynchronizeActualModuleSpecificDeviceData;
begin
  // Don't synchronize the TDeviceIdentifier data
  fActualBasicDeviceData.TDD:=fRequiredBasicDeviceData.TDD;
end;

procedure TBasicHartModule.SynchronizeRequiredModuleSpecificDeviceData;
begin
  // Don't synchronize the TDeviceIdentifier data
  fRequiredBasicDeviceData.TDD:=fActualBasicDeviceData.TDD;
end;

function TBasicHartModule.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result:=fBasicDeviceDataAvailable;
end;

procedure TBasicHartModule.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  fBasicDeviceDataAvailable:=false;
end;

procedure TBasicHartModule.SetModuleSpecificDeviceDataAvailableStatus;
begin
  fBasicDeviceDataAvailable:=true;
end;

// public

constructor TBasicHartModule.Create;
begin
  fDeviceType:=dtNone;
  FillChar(fActualBasicDeviceData.UID, SizeOf(fActualBasicDeviceData.UID), 0);

  with fActualBasicDeviceData.UID do
       begin
       DeviceTypeCode254:=254;    // According to HCF_SPEC-127, UID details this value shall be 254
       ManufacturerIdentCode:=CHart_Manufacturer_MTL;
       end;

  lHartMaster:=nil;

  fDeviceDataStream:=TMemoryStream.Create;

  fBasicDeviceDataAvailable:=false;
end;

destructor TBasicHartModule.Destroy;
begin
  inherited Destroy;

  fDeviceDataStream.Free;
end;

procedure TBasicHartModule.AssignHartMaster(var Value: THartMaster);
begin
  if (Value<>nil) then
     begin
     lHartMaster:=@Value;
     fBasicDeviceDataAvailable:=false;
     end;
end;

procedure TBasicHartModule.AssignReceivedCharEvent(Value: TSimpleEvent);
begin
  if (lHartMaster<>nil) then
     begin
     lHartMaster^.OnReceiveChar:=Value;
     lHartMaster^.ResumeReception;
     end;
end;

procedure TBasicHartModule.UnAssignReceivedCharEvent;
begin
  if (lHartMaster<>nil) then
     begin
     lHartMaster^.OnReceiveChar:=nil;
     lHartMaster^.SuspendReception;
     end;
end;

function TBasicHartModule.GetConnectedDeviceType: TDeviceType;
var NewDeviceIdentifier: THDeviceIdentifier;
    lDevType: TDeviceType;
begin
  Result:=dtNone;
  if (lHartMaster<>nil) then
     begin
     if Read_Identifier(lHartMaster^, NewDeviceIdentifier, true) then
        begin
        if (NewDeviceIdentifier.ManufacturerIdentCode<>CHart_Manufacturer_MTL) then
           MessageDlg(LanguageStrings[msUnknownManufacturer],mtError,[mbCancel],hcNoHelpTopic)
        else
           begin
           lDevType:=IdentToType(NewDeviceIdentifier);
           if (lDevType<>dtNone) then
              Result:=lDevType
           else
              MessageDlg(LanguageStrings[msUnknownDeviceDetected],mtError,[mbCancel],hcNoHelpTopic)
           end;
        end
     else
        begin
        if (lHartMaster^).TimeOutErrorOccur then
           MessageDlg(LanguageStrings[msDeviceNotResponding],mtError,[mbOK,mbHelp],hcInterfaceError)
        else MessageDlg(LanguageStrings[msCommunicationFaults],mtError,[mbOK,mbHelp],hcInterfaceError);
        GetConnectedDeviceType:=dtNone;
        (lHartMaster^).PurgeReadWrite;
        end;
     end;
end;

function TBasicHartModule.GetModuleTransferDataLength(Value: TTransferDirection): integer;
begin
  Result:=GetModuleSpecificTransferDataLength(Value);
end;

procedure TBasicHartModule.DownloadActualDeviceData;
begin
  DownloadActualModuleSpecificDeviceData;
end;

procedure TBasicHartModule.UploadRequiredDeviceData;
begin
  UploadRequiredModuleSpecificDeviceData;
end;

function TBasicHartModule.VerifyDeviceData: boolean;
begin
  Result:=VerifyRequiredModuleSpecificDeviceData;
end;

procedure TBasicHartModule.SynchronizeActualDeviceData;
begin
  SynchronizeActualModuleSpecificDeviceData;
  ClearModuleSpecificDeviceDataStatusFlags;
end;

procedure TBasicHartModule.SynchronizeRequiredDeviceData;
begin
  SynchronizeRequiredModuleSpecificDeviceData;
  ClearModuleSpecificDeviceDataStatusFlags;
end;

function TBasicHartModule.SetActualDeviceData(SourceStream: TMemoryStream): boolean;
begin
  fDeviceDataStream.Clear;

  SourceStream.SaveToStream(fDeviceDataStream);

  ReadModuleSpecificDaviceDataFromStream;

  Result:=GetModuleSpecificDeviceDataAvailableStatus;
end;

function TBasicHartModule.GetActualDeviceDataStream: TMemoryStream;
begin
  fDeviceDataStream.Clear;
  fDeviceDataStream.Write(CNew_Config_Header,System.Length(CNew_Config_Header));

  WriteModuleSpecificDeviceDataToStream;

  Result:=fDeviceDataStream;
end;

procedure TBasicHartModule.ClearModuleDeviceDataAvailableStatus;
begin
  ClearModuleSpecificDeviceDataAvailableStatus;
end;

procedure TBasicHartModule.SetModuleDeviceDataAvailableStatus;
begin
  SetModuleSpecificDeviceDataAvailableStatus;
end;

procedure TBasicHartModule.EvaluateRequiredDeviceDataStatusFlags;
begin
  EvaluateModuleSpecificDeviceDataStatusFlags;
end;

function TBasicHartModule.ResetModule: boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Write_Master_Reset(lHartMaster^, true, true)
  else Result:=false;
end;

function TBasicHartModule.GeTHDynamicVariables(var Target: THDynamicVariables): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Read_DynamicVariables(lHartMaster^, Target, 250, true)
  else Result:=false;
end;

function TBasicHartModule.SetOutputCurrent(Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=WriteRead_Current_Value(lHartMaster^, 40, Value, true)
  else Result:=false;
end;

function TBasicHartModule.CalibratePoint20mA(Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=WriteRead_Current_Value(lHartMaster^, 46, Value, true)
  else Result:=false;
end;

function TBasicHartModule.CalibratePoint4mA(Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=WriteRead_Current_Value(lHartMaster^, 45, Value, true)
  else Result:=false;
end;

function TBasicHartModule.GeTHTransmitterStatus(var Target: THTransmitterStatus): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Read_TransmitterStatus(lHartMaster^, Target, true)
  else Result:=false;
end;

// ****************************************************************************

// TSingleChannelHartModule

// private

procedure TSingleChannelHartModule.SeTHPrimaryVariableOutputData(Value: THPrimaryVariableOutput);
begin
  fRequiredPrimaryVariableOutputData:=Value;
end;

procedure TSingleChannelHartModule.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetTransfer:=false;
  fSetUnits:=false;
  fSetRangeBlock:=false;
  fSetDamping:=false;
end;

procedure TSingleChannelHartModule.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  fSetTransfer:=(fRequiredPrimaryVariableOutputData.TransferFunction<>fActualPrimaryVariableOutputData.TransferFunction);
  fSetUnits:=(fRequiredPrimaryVariableOutputData.PVUnits<>fActualPrimaryVariableOutputData.PVUnits);
  fSetRangeBlock:=not(CompareRNGs(fRequiredPrimaryVariableOutputData,fActualPrimaryVariableOutputData));
  fSetDamping:=not(CompareDMPs(fRequiredPrimaryVariableOutputData,fActualPrimaryVariableOutputData));
end;

// protected

function TSingleChannelHartModule.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 19   // Command 15 - THPrimaryVariableOutput
  else  // Value=tdUpload
     begin
     if fSetTransfer then fDataLength:=fDataLength + 8 + 1;    // Command 47 - TransferFunction
     if fSetUnits then fDataLength:=fDataLength + 8 + 1;       // Command 44 - PVUnits
     if fSetRangeBlock then fDataLength:=fDataLength + 8 + 9;  // Command 35 - PVUnits + Point20mA + Point4mA
     if fSetDamping then fDataLength:=fDataLength + 8 + 4;     // Command 34 - DampingTimeConstant
     Result:=fDataLength;
     end;
end;

procedure TSingleChannelHartModule.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     fPrimaryVariableOutputDataAvailable:=(Read_PrimaryVariableOutput(lHartMaster^,fActualPrimaryVariableOutputData,true))
  else fPrimaryVariableOutputDataAvailable:=false;
end;

procedure TSingleChannelHartModule.UploadRequiredModuleSpecificDeviceData;
var lPrimaryVariableOutputData: THPrimaryVariableOutput;
begin
  if (lHartMaster<>nil) then
     begin
     lPrimaryVariableOutputData:=fRequiredPrimaryVariableOutputData;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetTransfer then
        if (WriteRead_TransferFunction_Information (lHartMaster^, lPrimaryVariableOutputData, true)) then
           fActualPrimaryVariableOutputData.TransferFunction:=lPrimaryVariableOutputData.TransferFunction;
      // note: however the command 35 includes the PVUnits, but it will not update this module property
      // therefore the command 44 must be used and the PVUnits must be updated before command 35
      if fSetUnits then
        if (WriteRead_Process_Units (lHartMaster^, lPrimaryVariableOutputData, true)) then
           fActualPrimaryVariableOutputData.PVUnits:=lPrimaryVariableOutputData.PVUnits;
     if fSetRangeBlock then
        if (WriteRead_Primary_Variables_Block (lHartMaster^, lPrimaryVariableOutputData, true)) then
           begin
           fActualPrimaryVariableOutputData.PVUnits:=lPrimaryVariableOutputData.PVUnits;
           fActualPrimaryVariableOutputData.Point20mA:=lPrimaryVariableOutputData.Point20mA;
           fActualPrimaryVariableOutputData.Point4mA:=lPrimaryVariableOutputData.Point4mA;
           end;
     if fSetDamping then
        if (WriteRead_Damping_Value (lHartMaster^, lPrimaryVariableOutputData, true)) then
           fActualPrimaryVariableOutputData.DampingTimeConstant:=lPrimaryVariableOutputData.DampingTimeConstant;
     end;
end;

function TSingleChannelHartModule.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          (fRequiredPrimaryVariableOutputData.TransferFunction=fActualPrimaryVariableOutputData.TransferFunction) and
          (fRequiredPrimaryVariableOutputData.PVUnits=fActualPrimaryVariableOutputData.PVUnits) and
          (CompareRNGs(fRequiredPrimaryVariableOutputData,fActualPrimaryVariableOutputData)) and
          (CompareDMPs(fRequiredPrimaryVariableOutputData,fActualPrimaryVariableOutputData));
end;

procedure TSingleChannelHartModule.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualPrimaryVariableOutputData:=fRequiredPrimaryVariableOutputData;
end;

procedure TSingleChannelHartModule.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredPrimaryVariableOutputData:=fActualPrimaryVariableOutputData;
end;

function TSingleChannelHartModule.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
            fPrimaryVariableOutputDataAvailable;
end;

procedure TSingleChannelHartModule.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fPrimaryVariableOutputDataAvailable:=false;
end;

procedure TSingleChannelHartModule.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fPrimaryVariableOutputDataAvailable:=true;
end;

// public

constructor TSingleChannelHartModule.Create;
begin
  inherited Create;

  fPrimaryVariableOutputDataAvailable:=false;
end;

destructor TSingleChannelHartModule.Destroy;
begin
  inherited Destroy;
end;

// ****************************************************************************

// THartModule_x073

// private

procedure THartModule_x073.Set_x073_ParameterSet(Value: THx073ParameterSet);
begin
  fRequiredx073ParameterSet:=Value;
end;

procedure THartModule_x073.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetOCAL:=false;
  fSetTFAL:=false;
  fSetFFS:=false;
  fSetIT:=false;
  fSetSTC:=false;
  fSetCJCOO:=false;
  fSetCJFAL:=false;
end;

procedure THartModule_x073.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  fSetOCAL:=(fRequiredx073ParameterSet.OpenCircuitAlarmLevel<>fActualx073ParameterSet.OpenCircuitAlarmLevel);
  fSetTFAL:=(fRequiredx073ParameterSet.TransmitterFailureAlarmLevel<>fActualx073ParameterSet.TransmitterFailureAlarmLevel);
  fSetFFS:=(fRequiredx073ParameterSet.FilterFrequencySelector<>fActualx073ParameterSet.FilterFrequencySelector);
  fSetIT:=(fRequiredx073ParameterSet.InputType<>fActualx073ParameterSet.InputType);
  fSetSTC:=(fRequiredx073ParameterSet.SmoothingTimeConstant<>fActualx073ParameterSet.SmoothingTimeConstant);
  fSetCJCOO:=(fRequiredx073ParameterSet.ColdJunctionCompensationOnOff<>fActualx073ParameterSet.ColdJunctionCompensationOnOff);
  fSetCJFAL:=(fRequiredx073ParameterSet.ColdJunctionFailureAlarmLevel<>fActualx073ParameterSet.ColdJunctionFailureAlarmLevel);
end;

// protected

function THartModule_x073.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 1 // Command 128 (0)  - OpenCircuitAlarmLevel
               + 8 + 1 // Command 128 (1)  - TransmitterFailureAlarmLevel
               + 8 + 1 // Command 128 (2)  - FilterFrequencySelector
               + 8 + 1 // Command 128 (3)  - InputType
               + 8 + 2 // Command 128 (4)  - SmoothingTimeConstant
               + 8 + 1 // Command 128 (5)  - ColdJunctionCompensationOnOff
               + 8 + 1 // Command 128 (70) - ColdJunctionFailureAlarmLevel
  else  // Value=tdUpload
     begin
     if fSetOCAL then fDataLength:=fDataLength + 8 + 8 + 1;  // Command 129 (0)  - OpenCircuitAlarmLevel
     if fSetTFAL then fDataLength:=fDataLength + 8 + 8 + 1;  // Command 129 (1)  - TransmitterFailureAlarmLevel
     if fSetFFS then fDataLength:=fDataLength + 8 + 8 + 1;   // Command 129 (2)  - FilterFrequencySelector
     if fSetIT then fDataLength:=fDataLength + 8 + 8 + 1;    // Command 129 (3)  - InputType
     if fSetSTC then fDataLength:=fDataLength + 8 + 8 + 2;   // Command 129 (4)  - SmoothingTimeConstant
     if fSetCJCOO then fDataLength:=fDataLength + 8 + 8 + 1; // Command 129 (5)  - ColdJunctionCompensationOnOff
     if fSetCJFAL then fDataLength:=fDataLength + 8 + 8 + 1; // Command 129 (70) - ColdJunctionFailureAlarmLevel
     Result:=fDataLength;
     end;
end;

procedure THartModule_x073.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     fx073ParameterSetAvailable:=(Read_Selected_Byte(lHartMaster^,128,0,fActualx073ParameterSet.OpenCircuitAlarmLevel,true)) and
                                 (Read_Selected_Byte(lHartMaster^,128,1,fActualx073ParameterSet.TransmitterFailureAlarmLevel,true)) and
                                 (Read_Selected_Byte(lHartMaster^,128,2,fActualx073ParameterSet.FilterFrequencySelector,true)) and
                                 (Read_Selected_Byte(lHartMaster^,128,3,fActualx073ParameterSet.InputType,true)) and
                                 (Read_Selected_Word(lHartMaster^,128,4,fActualx073ParameterSet.SmoothingTimeConstant,true)) and
                                 (Read_Selected_Byte(lHartMaster^,128,5,fActualx073ParameterSet.ColdJunctionCompensationOnOff,true)) and
                                 (Read_Selected_Byte(lHartMaster^,128,70,fActualx073ParameterSet.ColdJunctionFailureAlarmLevel,true))
     else fx073ParameterSetAvailable:=false;
end;

procedure THartModule_x073.UploadRequiredModuleSpecificDeviceData;
var lx073ParameterSet: THx073ParameterSet;
begin
  if (lHartMaster<>nil) then
     begin
     lx073ParameterSet:=fRequiredx073ParameterSet;

     if fSetIT then
        if (Write_Selected_Byte (lHartMaster^, 129,3,lx073ParameterSet.InputType, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,3,lx073ParameterSet.InputType,true)) then
              fActualx073ParameterSet.InputType:=lx073ParameterSet.InputType;
           end;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetOCAL then
        if (Write_Selected_Byte (lHartMaster^, 129,0,lx073ParameterSet.OpenCircuitAlarmLevel, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,0,lx073ParameterSet.OpenCircuitAlarmLevel,true)) then
              fActualx073ParameterSet.OpenCircuitAlarmLevel:=lx073ParameterSet.OpenCircuitAlarmLevel;
           end;

     if fSetTFAL then
        if (Write_Selected_Byte (lHartMaster^, 129,1,lx073ParameterSet.TransmitterFailureAlarmLevel, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,1,lx073ParameterSet.TransmitterFailureAlarmLevel,true)) then
              fActualx073ParameterSet.TransmitterFailureAlarmLevel:=lx073ParameterSet.TransmitterFailureAlarmLevel;
           end;

     if fSetFFS then
        if (Write_Selected_Byte (lHartMaster^, 129,2,lx073ParameterSet.FilterFrequencySelector, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,2,lx073ParameterSet.FilterFrequencySelector,true)) then
              fActualx073ParameterSet.FilterFrequencySelector:=lx073ParameterSet.FilterFrequencySelector;
           end;

     {if fSetIT then
        if (Write_Selected_Byte (lHartMaster^, 129,3,lx073ParameterSet.InputType, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,3,lx073ParameterSet.InputType,true)) then
              fActualx073ParameterSet.InputType:=lx073ParameterSet.InputType;
           end;}

     if fSetSTC then
        if (Write_Selected_Word (lHartMaster^, 129,4,lx073ParameterSet.SmoothingTimeConstant, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Word(lHartMaster^,128,4,lx073ParameterSet.SmoothingTimeConstant,true)) then
              fActualx073ParameterSet.SmoothingTimeConstant:=lx073ParameterSet.SmoothingTimeConstant;
           end;

     if fSetCJCOO then
        if (Write_Selected_Byte (lHartMaster^, 129,5,lx073ParameterSet.ColdJunctionCompensationOnOff, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,5,lx073ParameterSet.ColdJunctionCompensationOnOff,true)) then
              fActualx073ParameterSet.ColdJunctionCompensationOnOff:=lx073ParameterSet.ColdJunctionCompensationOnOff;
           end;

     if fSetCJFAL then
        if (Write_Selected_Byte (lHartMaster^, 129,70,lx073ParameterSet.ColdJunctionFailureAlarmLevel, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,70,lx073ParameterSet.ColdJunctionFailureAlarmLevel,true)) then
              fActualx073ParameterSet.ColdJunctionFailureAlarmLevel:=lx073ParameterSet.ColdJunctionFailureAlarmLevel;
           end;
      end;
end;

function THartModule_x073.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          (fRequiredx073ParameterSet.OpenCircuitAlarmLevel=fActualx073ParameterSet.OpenCircuitAlarmLevel) and
          (fRequiredx073ParameterSet.TransmitterFailureAlarmLevel=fActualx073ParameterSet.TransmitterFailureAlarmLevel) and
          (fRequiredx073ParameterSet.FilterFrequencySelector=fActualx073ParameterSet.FilterFrequencySelector) and
          (fRequiredx073ParameterSet.InputType=fActualx073ParameterSet.InputType) and
          (fRequiredx073ParameterSet.SmoothingTimeConstant=fActualx073ParameterSet.SmoothingTimeConstant) and
          (fRequiredx073ParameterSet.ColdJunctionCompensationOnOff=fActualx073ParameterSet.ColdJunctionCompensationOnOff) and
          (fRequiredx073ParameterSet.ColdJunctionFailureAlarmLevel=fActualx073ParameterSet.ColdJunctionFailureAlarmLevel);
end;

procedure THartModule_x073.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

 fActualx073ParameterSet:=fRequiredx073ParameterSet;
end;

procedure THartModule_x073.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredx073ParameterSet:=fActualx073ParameterSet;
end;

procedure THartModule_x073.ReadModuleSpecificDaviceDataFromStream;
var lDx073: THx073;
begin
  fDeviceDataStream.Seek(System.Length(CNew_Config_Header),soFromBeginning);
  if ((fDeviceDataStream.Size-fDeviceDataStream.Position)=SizeOf(THx073)) then
     begin
     FillChar(lDx073,SizeOf(THx073),$FF);
     if (fDeviceDataStream.Read(lDx073,SizeOf(THx073))=SizeOf(THx073)) and
        (lDx073.Parity=GetDataStreamParityByte(lDx073,SizeOf(THx073)-4)) then
        begin
        while System.Length(lDx073.MDD.TDD.TagName)<8 do lDx073.MDD.TDD.TagName:=lDx073.MDD.TDD.TagName+' ';
        while System.Length(lDx073.MDD.TDD.Descriptor)<16 do lDx073.MDD.TDD.Descriptor:=lDx073.MDD.TDD.Descriptor+' ';
        if (lDx073.MDD.UID.ManufacturerIdentCode=CHart_Manufacturer_MTL) and
           (IdentToType(lDx073.MDD.UID) in [dt507x,dtx073,dt5074XK,dt4073XK]) then
           begin
           fActualBasicDeviceData.UID:=lDx073.MDD.UID;
           fActualBasicDeviceData.TDD:=lDx073.MDD.TDD;
           fActualPrimaryVariableOutputData:=lDx073.MDD.PVO;
           fActualx073ParameterSet:=lDx073.x073Set;
           SetModuleSpecificDeviceDataAvailableStatus;
           end;
        end;
     end;
end;

procedure THartModule_x073.WriteModuleSpecificDeviceDataToStream;
var lDx073: THx073;
begin
  lDx073.MDD.UID:=fActualBasicDeviceData.UID;
  lDx073.MDD.TDD:=fActualBasicDeviceData.TDD;
  lDx073.MDD.PVO:=fActualPrimaryVariableOutputData;
  lDx073.x073Set:=fActualx073ParameterSet;
  lDx073.Parity:=GetDataStreamParityByte(lDx073,SizeOf(THx073)-4);

  fDeviceDataStream.Write(lDx073,SizeOf(THx073));
end;

function THartModule_x073.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
          fx073ParameterSetAvailable;
end;

procedure THartModule_x073.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fx073ParameterSetAvailable:=false;
end;

procedure THartModule_x073.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fx073ParameterSetAvailable:=true;
end;

// public

constructor THartModule_x073.Create;
begin
  inherited Create;

  fx073ParameterSetAvailable:=false;
end;

destructor THartModule_x073.Destroy;
begin
  inherited Destroy;
end;

// ****************************************************************************

// THartModule_x075

// private

procedure THartModule_x075.Set_x075_Parameters(Value: THx075TripParameters);
begin
  fRequiredx075TripParameters:=Value;
end;

procedure THartModule_x075.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetx075Trip:=false;
end;

procedure THartModule_x075.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  fSetx075Trip:=not(Compare075s(fRequiredx075TripParameters,fActualx075TripParameters));
end;

// protected

function THartModule_x075.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 9 // Command 205 - x075TripParameters
  else  // Value=tdUpload
     begin
     if fSetx075Trip then fDataLength:=fDataLength + 8 + 9;  // Command 204 - x075TripParameters
     Result:=fDataLength;
     end;
end;

procedure THartModule_x075.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     fx075TripParametersAvailable:=(Read_TripParameters_x075(lHartMaster^,fActualx075TripParameters,true))
     else fx075TripParametersAvailable:=false;
end;

procedure THartModule_x075.UploadRequiredModuleSpecificDeviceData;
var lx075TripParameters: THx075TripParameters;
begin
  if (lHartMaster<>nil) then
     begin
     lx075TripParameters:=fRequiredx075TripParameters;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetx075Trip then
        if (WriteRead_TripParameters_x075 (lHartMaster^, lx075TripParameters, true)) then
           fActualx075TripParameters:=lx075TripParameters;
     end;
end;

function THartModule_x075.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          Compare075s(fRequiredx075TripParameters,fActualx075TripParameters);
end;

procedure THartModule_x075.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualx075TripParameters:=fRequiredx075TripParameters;
end;

procedure THartModule_x075.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredx075TripParameters:=fActualx075TripParameters;
end;

procedure THartModule_x075.ReadModuleSpecificDaviceDataFromStream;
var lDx075: THx075;
begin
  fDeviceDataStream.Seek(System.Length(CNew_Config_Header),soFromBeginning);
  if ((fDeviceDataStream.Size-fDeviceDataStream.Position)=SizeOf(THx075)) then
     begin
     FillChar(lDx075,SizeOf(THx075),$FF);
     if (fDeviceDataStream.Read(lDx075,SizeOf(THx075))=SizeOf(THx075)) and
        (lDx075.Parity=GetDataStreamParityByte(lDx075,SizeOf(THx075)-4)) then
        begin
        while System.Length(lDx075.MDD.TDD.TagName)<8 do lDx075.MDD.TDD.TagName:=lDx075.MDD.TDD.TagName+' ';
        while System.Length(lDx075.MDD.TDD.Descriptor)<16 do lDx075.MDD.TDD.Descriptor:=lDx075.MDD.TDD.Descriptor+' ';
        if (lDx075.MDD.UID.ManufacturerIdentCode=CHart_Manufacturer_MTL) and
           (IdentToType(lDx075.MDD.UID) in [dt4075]) then
           begin
           fActualBasicDeviceData.UID:=lDx075.MDD.UID;
           fActualBasicDeviceData.TDD:=lDx075.MDD.TDD;
           fActualPrimaryVariableOutputData:=lDx075.MDD.PVO;
           fActualx073ParameterSet:=lDx075.x073Set;
           fActualx075TripParameters:=lDx075.x075TPA;
           SetModuleSpecificDeviceDataAvailableStatus;
           end;
        end;
     end;
end;

procedure THartModule_x075.WriteModuleSpecificDeviceDataToStream;
var lDx075: THx075;
begin
  lDx075.MDD.UID:=fActualBasicDeviceData.UID;
  lDx075.MDD.TDD:=fActualBasicDeviceData.TDD;
  lDx075.MDD.PVO:=fActualPrimaryVariableOutputData;
  lDx075.x073Set:=fActualx073ParameterSet;
  lDx075.x075TPA:=fActualx075TripParameters;
  lDx075.Parity:=GetDataStreamParityByte(lDx075,SizeOf(THx075)-4);

  fDeviceDataStream.Write(lDx075,SizeOf(THx075));
end;

function THartModule_x075.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
            fx075TripParametersAvailable;
end;

procedure THartModule_x075.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fx075TripParametersAvailable:=false;
end;

procedure THartModule_x075.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fx075TripParametersAvailable:=true;
end;

// public

constructor THartModule_x075.Create;
begin
  inherited Create;

  fx075TripParametersAvailable:=false;
end;

destructor THartModule_x075.Destroy;
begin
  inherited Destroy;
end;

// ****************************************************************************

// THartModule_x575

// private

procedure THartModule_x575.SetResponseTime(Value: byte);
begin
  fRequiredResponseTime:=Value;
end;

procedure THartModule_x575.SetNE43(Value: byte);
begin
  fRequiredNE43Mode:=Value;
end;

procedure THartModule_x575.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetResponseTime:=false;
  fSetNE43:=false;
end;

procedure THartModule_x575.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  fSetResponseTime:=(fRequiredResponseTime<>fActualResponseTime);
  fSetNE43:=(fRequiredNE43Mode<>fActualNE43Mode);
end;

// protected

function THartModule_x575.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 1 // Command 208 - ResponseTime
               + 8 + 1 // Command 210 - NE43 Mode
  else  // Value=tdUpload
     begin
     if fSetResponseTime then fDataLength:=fDataLength + 8 + 1;  // Command 207 - ResponseTime
     if fSetNE43 then fDataLength:=fDataLength + 8 + 1;          // Command 209 - NE43 Mode
     Result:=fDataLength;
     end;
end;

procedure THartModule_x575.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     fx575ParametersAvailable:=(Read_One_Byte(lHartMaster^,208,fActualResponseTime,true)) and
                               (Read_One_Byte(lHartMaster^,210,fActualNE43Mode,true))
  else fx575ParametersAvailable:=false;
end;

procedure THartModule_x575.UploadRequiredModuleSpecificDeviceData;
var lResponseTime, lNE43Mode: byte;
begin
  if (lHartMaster<>nil) then
     begin
     lResponseTime:=fRequiredResponseTime;
     lNE43Mode:=fRequiredNE43Mode;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetResponseTime then
        if WriteRead_One_Byte(lHartMaster^, 207, lResponseTime, true) then
           fActualResponseTime:=lResponseTime;

     if fSetNE43 then
        if WriteRead_One_Byte(lHartMaster^, 209, lNE43Mode, true) then
           fActualNE43Mode:=lNE43Mode;
     end;
end;

function THartModule_x575.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          (fRequiredResponseTime=fActualResponseTime) and
          (fRequiredNE43Mode=fActualNE43Mode);
end;

procedure THartModule_x575.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualResponseTime:=fRequiredResponseTime;
  fActualNE43Mode:=fRequiredNE43Mode;
end;

procedure THartModule_x575.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredResponseTime:=fActualResponseTime;
  fRequiredNE43Mode:=fActualNE43Mode;
end;

procedure THartModule_x575.ReadModuleSpecificDaviceDataFromStream;
var lDx575: THx575;
begin
  fDeviceDataStream.Seek(System.Length(CNew_Config_Header),soFromBeginning);
  if ((fDeviceDataStream.Size-fDeviceDataStream.Position)=SizeOf(THx575)) then
     begin
     FillChar(lDx575,SizeOf(THx575),$FF);
     if (fDeviceDataStream.Read(lDx575,SizeOf(THx575))=SizeOf(THx575)) and
        (lDx575.Parity=GetDataStreamParityByte(lDx575,SizeOf(THx575)-4)) then
        begin
        while System.Length(lDx575.MDD.TDD.TagName)<8 do lDx575.MDD.TDD.TagName:=lDx575.MDD.TDD.TagName+' ';
        while System.Length(lDx575.MDD.TDD.Descriptor)<16 do lDx575.MDD.TDD.Descriptor:=lDx575.MDD.TDD.Descriptor+' ';
        if (lDx575.MDD.UID.ManufacturerIdentCode=CHart_Manufacturer_MTL) and
           (IdentToType(lDx575.MDD.UID) in [dtx575]) then
           begin
           fActualBasicDeviceData.UID:=lDx575.MDD.UID;
           fActualBasicDeviceData.TDD:=lDx575.MDD.TDD;
           fActualPrimaryVariableOutputData:=lDx575.MDD.PVO;
           fActualx073ParameterSet:=lDx575.x073Set;
           fActualx075TripParameters:=lDx575.x075TPA;
           fActualResponseTime:=lDx575.Speed;
           fActualNE43Mode:=lDx575.NAMUR;
           SetModuleSpecificDeviceDataAvailableStatus;
           end;
        end;
     end;
end;

procedure THartModule_x575.WriteModuleSpecificDeviceDataToStream;
var lDx575: THx575;
begin
  lDx575.MDD.UID:=fActualBasicDeviceData.UID;
  lDx575.MDD.TDD:=fActualBasicDeviceData.TDD;
  lDx575.MDD.PVO:=fActualPrimaryVariableOutputData;
  lDx575.x073Set:=fActualx073ParameterSet;
  lDx575.x075TPA:=fActualx075TripParameters;
  lDx575.Speed:=fActualResponseTime;
  lDx575.NAMUR:=fActualNE43Mode;
  lDx575.Parity:=GetDataStreamParityByte(lDx575,SizeOf(THx575)-4);

  fDeviceDataStream.Write(lDx575,SizeOf(THx575));
end;

function THartModule_x575.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
            fx575ParametersAvailable;
end;

procedure THartModule_x575.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fx575ParametersAvailable:=false;
end;

procedure THartModule_x575.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fx575ParametersAvailable:=true;
end;

// public

constructor THartModule_x575.Create;
begin
  inherited Create;

  fx575ParametersAvailable:=false;
end;

destructor THartModule_x575.Destroy;
begin
  inherited Destroy;
end;

// ****************************************************************************

// THartModule_x573

// private

procedure THartModule_x573.SetShortCircuitAlarmLevel(Value: byte);
begin
  fRequiredShortCircuitAlarmLevel:=Value;
end;

procedure THartModule_x573.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetShortCircuitAlarmLevel:=false;
end;

procedure THartModule_x573.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  fSetShortCircuitAlarmLevel:=(fActualShortCircuitAlarmLevel<>fRequiredShortCircuitAlarmLevel);
end;

// protected

function THartModule_x573.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 1 // Command 128 (6)  - ShortCircuitAlarmLevel
  else  // Value=tdUpload
     begin
     if fSetShortCircuitAlarmLevel then fDataLength:=fDataLength + 8 + 8 + 1;  // Command 129 (6)  - ShortCircuitAlarmLevel
     Result:=fDataLength;
     end;
end;

procedure THartModule_x573.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     fx573ParametersAvailable:=(Read_Selected_Byte(lHartMaster^,128,6,fActualShortCircuitAlarmLevel,true))
  else fx573ParametersAvailable:=false;
end;

procedure THartModule_x573.UploadRequiredModuleSpecificDeviceData;
var lShortCircuitAlarmLevel: byte;
begin
  if (lHartMaster<>nil) then
     begin
     lShortCircuitAlarmLevel:=fRequiredShortCircuitAlarmLevel;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetShortCircuitAlarmLevel then
        if (Write_Selected_Byte (lHartMaster^, 129,6,lShortCircuitAlarmLevel, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,6,lShortCircuitAlarmLevel,true)) then
              fActualShortCircuitAlarmLevel:=lShortCircuitAlarmLevel;
           end;
     end;
end;

function THartModule_x573.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          (fRequiredShortCircuitAlarmLevel=fActualShortCircuitAlarmLevel);
end;

procedure THartModule_x573.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualShortCircuitAlarmLevel:=fRequiredShortCircuitAlarmLevel;
end;

procedure THartModule_x573.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredShortCircuitAlarmLevel:=fActualShortCircuitAlarmLevel;
end;

procedure THartModule_x573.ReadModuleSpecificDaviceDataFromStream;
var lDx573: THx573;
begin
  fDeviceDataStream.Seek(System.Length(CNew_Config_Header),soFromBeginning);
  if ((fDeviceDataStream.Size-fDeviceDataStream.Position)=SizeOf(THx573)) then
     begin
     FillChar(lDx573,SizeOf(THx573),$FF);
     if (fDeviceDataStream.Read(lDx573,SizeOf(THx573))=SizeOf(THx573)) and
        (lDx573.Parity=GetDataStreamParityByte(lDx573,SizeOf(THx573)-4)) then
        begin
        while System.Length(lDx573.MDD.TDD.TagName)<8 do lDx573.MDD.TDD.TagName:=lDx573.MDD.TDD.TagName+' ';
        while System.Length(lDx573.MDD.TDD.Descriptor)<16 do lDx573.MDD.TDD.Descriptor:=lDx573.MDD.TDD.Descriptor+' ';
        if (lDx573.MDD.UID.ManufacturerIdentCode=CHart_Manufacturer_MTL) and
           (IdentToType(lDx573.MDD.UID) in [dtx573]) then
           begin
           fActualBasicDeviceData.UID:=lDx573.MDD.UID;
           fActualBasicDeviceData.TDD:=lDx573.MDD.TDD;
           fActualPrimaryVariableOutputData:=lDx573.MDD.PVO;
           fActualx073ParameterSet:=lDx573.x073Set;
           fActualx075TripParameters:=lDx573.x075TPA;
           fActualResponseTime:=lDx573.Speed;
           fActualNE43Mode:=lDx573.NAMUR;
           fActualShortCircuitAlarmLevel:=lDx573.ShortCircuitAlarmLevel;
           SetModuleSpecificDeviceDataAvailableStatus;
           end;
        end;
     end;
end;

procedure THartModule_x573.WriteModuleSpecificDeviceDataToStream;
var lDx573: THx573;
begin
  lDx573.MDD.UID:=fActualBasicDeviceData.UID;
  lDx573.MDD.TDD:=fActualBasicDeviceData.TDD;
  lDx573.MDD.PVO:=fActualPrimaryVariableOutputData;
  lDx573.x073Set:=fActualx073ParameterSet;
  lDx573.x075TPA:=fActualx075TripParameters;
  lDx573.Speed:=fActualResponseTime;
  lDx573.NAMUR:=fActualNE43Mode;
  lDx573.ShortCircuitAlarmLevel:=fActualShortCircuitAlarmLevel;
  lDx573.Parity:=GetDataStreamParityByte(lDx573,SizeOf(THx573)-4);

  fDeviceDataStream.Write(lDx573,SizeOf(THx573));
end;

function THartModule_x573.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
            fx573ParametersAvailable;
end;

procedure THartModule_x573.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fx573ParametersAvailable:=false;
end;

procedure THartModule_x573.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fx573ParametersAvailable:=true;
end;

// public

constructor THartModule_x573.Create;
begin
  inherited Create;

  fx575ParametersAvailable:=false;
end;

destructor THartModule_x573.Destroy;
begin
  inherited Destroy;
end;

// ****************************************************************************

// THartModule_Freq

// private

procedure THartModule_Freq.SetTransmitterFailureAlarmLevel(Value: byte);
begin
  fRequiredTransmitterFailureAlarmLevel:=Value;
end;

procedure THartModule_Freq.SeTHAdjustOutputParameters(Value: THAdjustOutputParameters);
begin
  fRequiredAdjustOutputParameters:=Value;
end;

procedure THartModule_Freq.SetFullScale(Value: byte);
begin
  fRequiredFullScale:=Value;
end;

procedure THartModule_Freq.SeTHDividerParameters(Value: THDividerParameters);
begin
  fRequiredDividerParameters:=Value;
end;

procedure THartModule_Freq.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetTFAL:=false;
  fSetAdjust:=false;

  fSetFullScale:=false;
  fSetDivider:=false;

  // Note: On Frequency modules the below parameter must be cleared before other can be set
  // this will ensure smoothly module reconfiguration
  fClearTFAL:=false;
  fClearFullScale:=false;
end;

procedure THartModule_Freq.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  fSetAdjust:=not(CompareAOPs(fRequiredAdjustOutputParameters,fActualAdjustOutputParameters));
  fSetDivider:=not(CompareDVFs(fRequiredDividerParameters,fActualDividerParameters));

  // Note: On Frequency modules the below parameter must be cleared before other can be set
  // this will ensure smoothly module reconfiguration
  fSetTFAL:=(fRequiredTransmitterFailureAlarmLevel<>fActualTransmitterFailureAlarmLevel);
  if fSetTFAL then
     begin
     fClearTFAL:=(AlarmLevelFromHART(fRequiredTransmitterFailureAlarmLevel)=taOff);
     if fClearTFAL then
        fSetTFAL:=false;
     end
  else fClearTFAL:=false;

  fSetFullScale:=(fRequiredFullScale<>fActualFullScale);
  if fSetFullScale then
     begin
     fClearFullScale:=(fRequiredFullScale=0);
     if fClearFullScale then
        fSetFullScale:=false;
     end
  else fClearFullScale:=false;
end;

// protected

function THartModule_Freq.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 1  // Command 128 (1) - TransmitterFailureAlarmLevel
               + 8 + 10 // Command 141     - Trimmer Mode (+/- x %)
               + 8 + 1  // Command 147     - Full Scale (0 to 20mA)
               + 8 + 9  // Command 139     - Pulse Output / Divider Parameter
  else  // Value=tdUpload
     begin
     if fSetTFAL then fDataLength:=fDataLength + 8 + 4;        // Command 129 (1) - TransmitterFailureAlarmLevel
     if fSetAdjust then fDataLength:=fDataLength + 8 + 1;      // Command 140     - Trimmer Mode (+/- x %)
     if fSetFullScale then fDataLength:=fDataLength + 8 + 1;   // Command 146     - Full Scale (0 to 20mA)
     if fSetDivider then fDataLength:=fDataLength + 8 + 4;     // Command 138     - Pulse Output / Divider Parameter
     // Note: On Frequency modules the below parameter must be cleared before other can be set
     // this will ensure smoothly module reconfiguration
     if fClearTFAL then fDataLength:=fDataLength + 8 + 4;      // Command 129 (1) - Clear TransmitterFailureAlarmLevel
     if fClearFullScale then fDataLength:=fDataLength + 8 + 1; // Command 146     - Clear Full Scale (0 to 20mA)
     Result:=fDataLength;
     end;
end;

procedure THartModule_Freq.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     fxFreqParametersAvailable:=(Read_Selected_Byte(lHartMaster^,128,1,fActualTransmitterFailureAlarmLevel,true)) and
                                (Read_AdjustOutputParameters(lHartMaster^,fActualAdjustOutputParameters,true)) and
                                (Read_One_Byte(lHartMaster^,147,fActualFullScale,true)) and
                                (Read_DividerParameters(lHartMaster^,fActualDividerParameters,true))
  else fxFreqParametersAvailable:=false;
end;

procedure THartModule_Freq.UploadRequiredModuleSpecificDeviceData;
var lAdjustOutputParameters: THAdjustOutputParameters;
    lDividerParameters: THDividerParameters;
    lTransmitterFailureAlarmLevel: byte;
    lFullScale: byte;
begin

  if (lHartMaster<>nil) then
     begin
     lTransmitterFailureAlarmLevel:=fRequiredTransmitterFailureAlarmLevel;
     lAdjustOutputParameters:=fRequiredAdjustOutputParameters;
     lDividerParameters:=fRequiredDividerParameters;
     lFullScale:=fRequiredFullScale;

     inherited UploadRequiredModuleSpecificDeviceData;

  // Note: On Frequency modules some parameter must be cleared before other can be set
  // this will ensure smoothly module reconfiguration
     if fClearTFAL then
        if (Write_Selected_Byte (lHartMaster^, 129,1,lTransmitterFailureAlarmLevel, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,1,lTransmitterFailureAlarmLevel,true)) then
              fActualTransmitterFailureAlarmLevel:=lTransmitterFailureAlarmLevel;
           end;

     if fClearFullScale then
        if (Write_One_Byte(lHartMaster^,146,lFullScale, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_One_Byte(lHartMaster^,147,lFullScale,true)) then
              fActualFullScale:=lFullScale;
           end;

     if fSetTFAL then
        if (Write_Selected_Byte (lHartMaster^, 129,1,lTransmitterFailureAlarmLevel, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,1,lTransmitterFailureAlarmLevel,true)) then
              fActualTransmitterFailureAlarmLevel:=lTransmitterFailureAlarmLevel;
           end;

     if fSetAdjust then
        if Write_AdjustOutputParameters(lHartMaster^, lAdjustOutputParameters, true) then
           begin
           WaitMilliseconds(250);
           if (Read_AdjustOutputParameters(lHartMaster^,lAdjustOutputParameters,true)) then
              fActualAdjustOutputParameters:=lAdjustOutputParameters;
           end;

     if fSetFullScale then
        if (Write_One_Byte(lHartMaster^,146,lFullScale, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_One_Byte(lHartMaster^,147,lFullScale,true)) then
              fActualFullScale:=lFullScale;
           end;

     if fSetDivider then
        if Write_DividerParameters(lHartMaster^, lDividerParameters, true) then
           begin
           WaitMilliseconds(250);
           if (Read_DividerParameters(lHartMaster^,lDividerParameters,true)) then
              fActualDividerParameters:=lDividerParameters;
           end;
      end;
end;

function THartModule_Freq.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          (fRequiredTransmitterFailureAlarmLevel=fActualTransmitterFailureAlarmLevel) and
          (CompareAOPs(fRequiredAdjustOutputParameters,fActualAdjustOutputParameters)) and
          (fRequiredFullScale=fActualFullScale) and
          (CompareDVFs(fRequiredDividerParameters,fActualDividerParameters));
end;

procedure THartModule_Freq.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualTransmitterFailureAlarmLevel:=fRequiredTransmitterFailureAlarmLevel;
  fActualAdjustOutputParameters:=fRequiredAdjustOutputParameters;
  fActualFullScale:=fRequiredFullScale;
  fActualDividerParameters:=fRequiredDividerParameters;
end;

procedure THartModule_Freq.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredTransmitterFailureAlarmLevel:=fActualTransmitterFailureAlarmLevel;
  fRequiredAdjustOutputParameters:=fActualAdjustOutputParameters;
  fRequiredFullScale:=fActualFullScale;
  fRequiredDividerParameters:=fActualDividerParameters;
end;


function THartModule_Freq.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
            fxFreqParametersAvailable;
end;

procedure THartModule_Freq.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fxFreqParametersAvailable:=false;
end;

procedure THartModule_Freq.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fxFreqParametersAvailable:=true;
end;

// public

constructor THartModule_Freq.Create;
begin
  inherited Create;

  fxFreqParametersAvailable:=false;
end;

destructor THartModule_Freq.Destroy;
begin

  inherited Destroy;
end;

function THartModule_Freq.GetInternalValues(Selector: byte; var Target: TByteDynArray): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=(Read_Internal_Value(lHartMaster^, Target, Selector, true))
  else Result:=false;
end;

function THartModule_Freq.ResetConfigurationFlag: boolean;
begin
  Result:=WriteRead_Reset_Configuration_Changed(lHartMaster^,true);
end;

function THartModule_Freq.ResetFactoryConfiguration: boolean;
begin
  Result:=Write_SetDeviceToDefault(lHartMaster^,true);
end;

// ****************************************************************************

// THartModule_x532

// private

procedure THartModule_x532.SetNE43(Value: byte);
begin
  fRequiredNE43Mode:=Value;
end;

procedure THartModule_x532.SetLeakFlowOffParameters(Value: THLeakFlowOffParameters_x532);
begin
  fRequiredLeakFlowOffParameters:=Value;
end;

procedure THartModule_x532.SetInputConfig(Value: THInputConfiguration_x532);
begin
  fRequiredInputConfig:=Value;
end;

procedure THartModule_x532.SetTripParameters(Value: THTripParameters_x532);
begin
  fRequiredTripParameters:=Value;
end;

procedure THartModule_x532.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetNE43:=false;
  fSetLeakFlow:=false;
  fSetInputConfig:=false;
  fSetTrip:=false;

  // Note: On MTL4532 the below parameter must be cleared before other can be set
  // this will ensure smoothly module reconfiguration
  fClearNE43:=false;
end;

procedure THartModule_x532.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  // Note: On MTL4532 the below parameter must be ignored;
  fSetTransfer:=false;

  fSetLeakFlow:=not(CompareLFPs(fRequiredLeakFlowOffParameters,fActualLeakFlowOffParameters));
  fSetTrip:=not(CompareTPXs(fRequiredTripParameters,fActualTripParameters));

  // Note: On MTL4532 the below parameter must be cleared before other can be set
  // this will ensure smoothly module reconfiguration
  fSetNE43:=(fRequiredNE43Mode<>fActualNE43Mode);
  if fSetNE43 then
     begin
     fClearNE43:=((fRequiredNE43Mode and CHart_x575_NE43_On)=0);
     if fClearNE43 then
        fSetNE43:=false;
     end
  else fClearNE43:=false;

  fSetInputConfig:=(fRequiredInputConfig.InputSource<>fActualInputConfig.InputSource) or
                   (fRequiredInputConfig.SignalInvert<>fActualInputConfig.SignalInvert);
  if fSetInputConfig then
     begin
     fClearInputConfig:=((fRequiredInputConfig.InputSource and CP_OutputSelector)=0);
     if fClearInputConfig then
        fSetInputConfig:=false;
     end;
end;

// protected

function THartModule_x532.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 2   // Command 132     - Input configuration
               + 8 + 10  // Command 134     - Leak Flow Parameters
               + 8 + 1   // Command 128 (6) - Output Characteristic (NE43)
               + 8 + 19  // Command 137     - Trip Parameter (assume most complex window trip)
  else  // Value=tdUpload
     begin
     if fSetInputConfig then fDataLength:=fDataLength + 8 + 4; // Command 131     - Input configuration
     if fSetLeakFlow then fDataLength:=fDataLength + 8 + 10;   // Command 133     - Leak Flow Parameters
     if fSetNE43 then fDataLength:=fDataLength + 8 + 1;        // Command 129 (6) - Set Output Characteristic (NE43)
     if fSetTrip then fDataLength:=fDataLength + 8 + 4 + 4;    // Command 136     - Trip Parameter (assume most complex window trip)
     // Note: On MTL4532 the command 35 reply is occasionally wrong
     // Therefore the command 15 must be used to verify correct configuration
     // and consequently module specific data length must be accordingly increased
     if fSetRangeBlock then fDataLength:=fDataLength + 8 + 19; // Command 15 - THPrimaryVariableOutput

     // Note: On MTL4532 the below parameter must be cleared before other can be set
     // this will ensure smoothly module reconfiguration
     if fClearNE43 then fDataLength:=fDataLength + 8 + 1;        // Command 129 (6) - Clear Output Characteristic (NE43)
     if fClearInputConfig then fDataLength:=fDataLength + 8 + 4; // Command 131     - Clear Input configuration
     Result:=fDataLength;
     end;
end;

procedure THartModule_x532.DownloadActualModuleSpecificDeviceData;
var lInputConfig: word;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  // Note: On MTL4532 the below parameter must be set to 0;
  if fPrimaryVariableOutputDataAvailable then
     fActualPrimaryVariableOutputData.TransferFunction:=0;

  if (lHartMaster<>nil) then
     fx532ParametersAvailable:=(Read_One_Word(lHartMaster^,132,lInputConfig,true)) and
                               (Read_LeakFlowOffParameters_x532(lHartMaster^,fActualLeakFlowOffParameters,true)) and
                               (Read_Selected_Byte(lHartMaster^,128,6,fActualNE43Mode,true)) and
                               (Read_TripParameters_x532(lHartMaster^,fActualTripParameters,true))
  else fx532ParametersAvailable:=false;
  if fx532ParametersAvailable then
     begin
     fActualInputConfig.InputSource:=(lInputConfig DIV 256);
     fActualInputConfig.SignalInvert:=(lInputConfig MOD 256);
     end;
end;

procedure THartModule_x532.UploadRequiredModuleSpecificDeviceData;
var lLeakFlowOffParameters: THLeakFlowOffParameters_x532;
    lPrimaryVariableOutputDataWrite,
    lPrimaryVariableOutputDataRead: THPrimaryVariableOutput;
    lInputConfiguration: THInputConfiguration_x532;
    lTripParameters: THTripParameters_x532;
    lInputConfig: word;
    lNE43Mode: byte;
    lRB: boolean;
begin
  if (lHartMaster<>nil) then
     begin
     lNE43Mode:=fRequiredNE43Mode;
     lLeakFlowOffParameters:=fRequiredLeakFlowOffParameters;
     lInputConfiguration:=fRequiredInputConfig;
     lTripParameters:=fRequiredTripParameters;

     // Note: On MTL4532 the below parameter must be cleared before other can be set
     // this will ensure smoothly module reconfiguration
     if fClearNE43 then
        if (Write_Selected_Byte (lHartMaster^, 129,6,lNE43Mode, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,6,lNE43Mode,true)) then
              fActualNE43Mode:=lNE43Mode;
           end;

     if fClearInputConfig then
        begin
        lInputConfig:=lInputConfiguration.InputSource*256+lInputConfiguration.SignalInvert;
        if (Write_One_Word (lHartMaster^, 131, word(lInputConfig), true)) then
           begin
           WaitMilliseconds(250);
           if (Read_One_Word(lHartMaster^,132,word(lInputConfig),true)) then
              begin
              lInputConfiguration.InputSource:=(lInputConfig DIV 256);
              lInputConfiguration.SignalInvert:=(lInputConfig MOD 256);
              fActualInputConfig:=lInputConfiguration;
              end;
           end;
         end;

     // Note: On MTL4532 there is apparently a firmware bug in the command 34 Range Block configuration
     // causing wrong Primary Variables reply (however the the parameters will be in fact correct changed)
     // Therefore if Range Block Upload is required, then MTL4532 module specific solution must be used
     // before the inherited method designed for standard (bug free) modules will cause comm failure.
     if fSetRangeBlock then
        begin
        lPrimaryVariableOutputDataWrite:=fRequiredPrimaryVariableOutputData;
        lRB:=WriteRead_Primary_Variables_Block (lHartMaster^, lPrimaryVariableOutputDataWrite, true);
        WaitMilliseconds(500);
        if (Read_PrimaryVariableOutput (lHartMaster^, lPrimaryVariableOutputDataRead, true)) and
           (CompareRNGs(fRequiredPrimaryVariableOutputData,lPrimaryVariableOutputDataRead)) then
           begin
           fActualPrimaryVariableOutputData.PVUnits:=lPrimaryVariableOutputDataRead.PVUnits;
           fActualPrimaryVariableOutputData.Point20mA:=lPrimaryVariableOutputDataRead.Point20mA;
           fActualPrimaryVariableOutputData.Point4mA:=lPrimaryVariableOutputDataRead.Point4mA;
           fSetRangeBlock:=false;
           end;
        end;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetNE43 then
        if (Write_Selected_Byte (lHartMaster^, 129,6,lNE43Mode, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_Selected_Byte(lHartMaster^,128,6,lNE43Mode,true)) then
              fActualNE43Mode:=lNE43Mode;
           end;

     if fSetLeakFlow then
        if Write_LeakFlowOffParameters_x532(lHartMaster^, lLeakFlowOffParameters, true) then
           begin
           WaitMilliseconds(250);
           if (Read_LeakFlowOffParameters_x532(lHartMaster^,lLeakFlowOffParameters,true)) then
              fActualLeakFlowOffParameters:=lLeakFlowOffParameters;
           end;

     if fSetInputConfig then
        begin
        lInputConfig:=lInputConfiguration.InputSource*256+lInputConfiguration.SignalInvert;
        if (Write_One_Word (lHartMaster^, 131, word(lInputConfig), true)) then
           begin
           WaitMilliseconds(250);
           if (Read_One_Word(lHartMaster^,132,word(lInputConfig),true)) then
              begin
              lInputConfiguration.InputSource:=(lInputConfig DIV 256);
              lInputConfiguration.SignalInvert:=(lInputConfig MOD 256);
              fActualInputConfig:=lInputConfiguration;
              end;
           end;
        end;

     if fSetTrip then
        if Write_TripParameters_x532(lHartMaster^, lTripParameters, true) then
           begin
           WaitMilliseconds(250);
           if (Read_TripParameters_x532(lHartMaster^,lTripParameters,true)) then
              fActualTripParameters:=lTripParameters;
           end;
     end;
end;

function THartModule_x532.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          (fRequiredNE43Mode=fActualNE43Mode) and
          (CompareLFPs(fRequiredLeakFlowOffParameters,fActualLeakFlowOffParameters)) and
          (fRequiredInputConfig.InputSource=fActualInputConfig.InputSource) and
          (fRequiredInputConfig.SignalInvert=fActualInputConfig.SignalInvert) and
          (CompareTPXs(fRequiredTripParameters,fActualTripParameters));
end;

procedure THartModule_x532.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualNE43Mode:=fRequiredNE43Mode;
  fActualLeakFlowOffParameters:=fRequiredLeakFlowOffParameters;
  fActualInputConfig:=fRequiredInputConfig;
  fActualTripParameters:=fRequiredTripParameters;
end;

procedure THartModule_x532.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredNE43Mode:=fActualNE43Mode;
  fRequiredLeakFlowOffParameters:=fActualLeakFlowOffParameters;
  fRequiredInputConfig:=fActualInputConfig;
  fRequiredTripParameters:=fActualTripParameters;
end;

procedure THartModule_x532.ReadModuleSpecificDaviceDataFromStream;
var lDx532: THx532;
    lSize: integer;
begin
   // Note: the Read procedure for MTLx532 differ from the previous modules
   // because of different file structure used in the software version 3.11
   lSize:=Sizeof(THMainDeviceData)+Sizeof(THAddDeviceData_x532)+SizeOf(THInputConfiguration_x532)+
         Sizeof(THDividerParameters)+Sizeof(THTripParameters_x532)+SizeOf(byte);
  fDeviceDataStream.Seek(System.Length(CNew_Config_Header),soFromBeginning);
  if ((fDeviceDataStream.Size-fDeviceDataStream.Position)=lSize) then
     begin
     FillChar(lDx532,SizeOf(THx532),$FF);
     if (fDeviceDataStream.Read(lDx532.MDD,SizeOf(THMainDeviceData))=SizeOf(THMainDeviceData)) and
        (fDeviceDataStream.Read(lDx532.ADD,SizeOf(THAddDeviceData_x532))=SizeOf(THAddDeviceData_x532)) and
        (fDeviceDataStream.Read(lDx532.InpConf,SizeOf(THInputConfiguration_x532))=SizeOf(THInputConfiguration_x532)) and
        (fDeviceDataStream.Read(lDx532.DVP,SizeOf(THDividerParameters))=SizeOf(THDividerParameters)) and
        (fDeviceDataStream.Read(lDx532.TPP,SizeOf(THTripParameters_x532))=SizeOf(THTripParameters_x532)) and
        (fDeviceDataStream.Read(lDx532.Parity,SizeOf(byte))=SizeOf(byte)) and
        (lDx532.Parity=GetDataStreamParityByte(lDx532.MDD,SizeOf(THMainDeviceData)) xor
                       GetDataStreamParityByte(lDx532.ADD,SizeOf(THAddDeviceData_x532)) xor
                       GetDataStreamParityByte(lDx532.InpConf,SizeOf(THInputConfiguration_x532)) xor
                       GetDataStreamParityByte(lDx532.DVP,SizeOf(THDividerParameters)) xor
                       GetDataStreamParityByte(lDx532.TPP,SizeOf(THTripParameters_x532))) then
        begin
        while System.Length(lDx532.MDD.TDD.TagName)<8 do lDx532.MDD.TDD.TagName:=lDx532.MDD.TDD.TagName+' ';
        while System.Length(lDx532.MDD.TDD.Descriptor)<16 do lDx532.MDD.TDD.Descriptor:=lDx532.MDD.TDD.Descriptor+' ';
        if (lDx532.MDD.UID.ManufacturerIdentCode=CHart_Manufacturer_MTL) and
           (IdentToType(lDx532.MDD.UID) in [dtx532]) then
           begin
           fActualBasicDeviceData.UID:=lDx532.MDD.UID;
           fActualBasicDeviceData.TDD:=lDx532.MDD.TDD;
           fActualPrimaryVariableOutputData:=lDx532.MDD.PVO;

           fActualFullScale:=lDx532.ADD.UnderOverRangeMode;  // fActualNE43Mode:=lDx532.TPP.
           fActualTransmitterFailureAlarmLevel:=lDx532.ADD.TransmitterFailureAlarmLevel;
           fActualAdjustOutputParameters:=lDx532.ADD.AOP;
           fActualDividerParameters:=lDx532.DVP;

           fActualLeakFlowOffParameters:=lDx532.ADD.LFP;
           fActualInputConfig:=lDx532.InpConf;
           fActualTripParameters:=lDx532.TPP;

           SetModuleSpecificDeviceDataAvailableStatus;
           end;
        end;
     end;
end;

procedure THartModule_x532.WriteModuleSpecificDeviceDataToStream;
var lDx532: THx532;
    lParity: byte;
begin
  // Note: the Write procedure for MTLx532 differ from the previous modules
  // because of different file structure used in the software version 3.11
  lDx532.MDD.UID:=fActualBasicDeviceData.UID;
  lDx532.MDD.TDD:=fActualBasicDeviceData.TDD;
  lDx532.MDD.PVO:=fActualPrimaryVariableOutputData;
  fDeviceDataStream.Write(lDx532.MDD,SizeOf(THMainDeviceData));

  lDx532.ADD.AOP:=fActualAdjustOutputParameters;
  lDx532.ADD.LFP:=fActualLeakFlowOffParameters;
  lDx532.ADD.TransmitterFailureAlarmLevel:=fActualTransmitterFailureAlarmLevel;
  lDx532.ADD.UnderOverRangeMode:=fActualFullScale;
  fDeviceDataStream.Write(lDx532.ADD,SizeOf(THAddDeviceData_x532));

  lDx532.InpConf:=fActualInputConfig;
  fDeviceDataStream.Write(lDx532.InpConf,SizeOf(THInputConfiguration_x532));

  lDx532.DVP:=fActualDividerParameters;
  fDeviceDataStream.Write(lDx532.DVP,SizeOf(THDividerParameters));

  lDx532.TPP:=fActualTripParameters;
  fDeviceDataStream.Write(lDx532.TPP,SizeOf(THTripParameters_x532));

  lParity:=GetDataStreamParityByte(lDx532.MDD,SizeOf(THMainDeviceData)) xor
           GetDataStreamParityByte(lDx532.ADD,SizeOf(THAddDeviceData_x532)) xor
           GetDataStreamParityByte(lDx532.InpConf,SizeOf(THInputConfiguration_x532)) xor
           GetDataStreamParityByte(lDx532.DVP,SizeOf(THDividerParameters)) xor
           GetDataStreamParityByte(lDx532.TPP,SizeOf(THTripParameters_x532));

  fDeviceDataStream.Write(lParity,SizeOf(byte));
end;

function THartModule_x532.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
            fx532ParametersAvailable;
end;

procedure THartModule_x532.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fx532ParametersAvailable:=false;
end;

procedure THartModule_x532.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fx532ParametersAvailable:=true;
end;

// public

constructor THartModule_x532.Create;
begin
  inherited Create;

  fx532ParametersAvailable:=false;
end;

destructor THartModule_x532.Destroy;
begin

  inherited Destroy;
end;

function THartModule_x532.GetAssemblyNumber(var Target: dword): boolean;
var lAssemblyNumber: THFinalAssemblyNumber;
begin
  if (lHartMaster<>nil) and
     Read_AssemblyNumber(lHartMaster^, lAssemblyNumber, true) then
     begin
     Target:=lAssemblyNumber.Number[1]*256*256+ // MSB
             lAssemblyNumber.Number[2]*256+
             lAssemblyNumber.Number[3];         // LSB
     Result:=true;
     end
  else Result:=false;
end;

function THartModule_x532.SetAssemblyNumber(Source: dword): boolean;
var lAssemblyNumber: THFinalAssemblyNumber;
begin
  if (lHartMaster<>nil) then
     begin
     lAssemblyNumber.Number[1]:=Source DIV (256*256);
     lAssemblyNumber.Number[2]:=(Source DIV 256) MOD 256;
     lAssemblyNumber.Number[3]:=Source MOD 256;
     Result:=WriteRead_AssemblyNumber(lHartMaster^, lAssemblyNumber, true);
     end
  else Result:=false;
end;

function THartModule_x532.SetDigitalDiagnostic(Selector: byte): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Write_One_Byte(lHartMaster^, 135, Selector, true)
  else Result:=false;
end;

function THartModule_x532.SetDigitalDiagnostic(Selector: byte; Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Write_Byte_And_Single(lHartMaster^, 135, Selector, Value, true)
  else Result:=false;
end;

// ****************************************************************************

// THartModule_1003

// private

procedure THartModule_1003.SetInputAndLFDMode(Value: byte);
begin
  fRequiredInputAndLFDMode:=Value;
end;

procedure THartModule_1003.SetLeakFlowOffParameters(Value: THLeakFlowOffParameters_1003);
begin
  fRequiredLeakFlowOffParameters:=Value;
end;

procedure THartModule_1003.SetTripAParameters(Value: THTripParameters_1003);
begin
  fRequiredTripAParameters:=Value;
end;

procedure THartModule_1003.SetTripBParameters(Value: THTripParameters_1003);
begin
  fRequiredTripBParameters:=Value;
end;

procedure THartModule_1003.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetLFDAndInputMode:=false;
  fSetLeakFlow:=false;
  fSetTripA:=false;
  fSetTripB:=false;

  // Note: On E01003 some parameters must be cleared before other can be set
  // to prevent multiple function assignment on the pulse and alarm output
  // and ensure smoothly module reconfiguration
  fClearLFDAndInputMode:=false;
  fClearDivider:=false;
  fClearTripA:=false;
  fClearTripB:=false;
end;

procedure THartModule_1003.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  // Note: On E01003 some parameters must be cleared before other can be set
  // to prevent multiple function assignment on the pulse and alarm output
  // and ensure smoothly module reconfiguration
  if fSetDivider then
     begin
     fClearDivider:=DetectClearDVFs(fRequiredDividerParameters.FunctionSelect);
     if fClearDivider then
        fSetDivider:=false;
     end
  else fClearDivider:=false;

  fSetLFDAndInputMode:=not(CompareInpLFDs(fRequiredInputAndLFDMode,fActualInputAndLFDMode));
  if fSetLFDAndInputMode then
     begin
     fClearLFDAndInputMode:=DetectClearInpLFDs(fRequiredInputAndLFDMode);
     if fClearLFDAndInputMode then
        fSetLFDAndInputMode:=false;
     end
  else fClearLFDAndInputMode:=false;

  fSetLeakFlow:=not(CompareLFPs(fRequiredLeakFlowOffParameters,fActualLeakFlowOffParameters));

  fSetTripA:=not(CompareTPXs(fRequiredTripAParameters,fActualTripAParameters));
  if fSetTripA then
     begin
     fClearTripA:=DetectClearTPXs(fRequiredTripAParameters.FunctionSelect);
     if fClearTripA then
        fSetTripA:=false;
     end
  else fClearTripA:=false;

  fSetTripB:=not(CompareTPXs(fRequiredTripBParameters,fActualTripBParameters));
  if fSetTripB then
     begin
     fClearTripB:=DetectClearTPXs(fRequiredTripBParameters.FunctionSelect);
     if fClearTripB then
        fSetTripB:=false;
     end
  else fClearTripB:=false;
end;

// protected

function THartModule_1003.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 1          // Command 132     - Input configuration
               + 8 + 9          // Command 134     - Leak Flow Parameters
               + 8 + 10         // Command 137 (0) - Trip A Parameter
               + 8 + 10         // Command 137 (1) - Trip B Parameter
               + 8 + 1          // Command 145     - Hardware Configuration (LFD circuit and Relay)
  else  // Value=tdUpload
     begin
     if fSetLFDAndInputMode then fDataLength:=fDataLength + 8 + 4;   // Command 131     - Input configuration
     if fSetLeakFlow then fDataLength:=fDataLength + 8 + 9;          // Command 133     - Leak Flow Parameters
     if fSetTripA then fDataLength:=fDataLength + 8 + 4;             // Command 136 (0) - Trip A Parameter
     if fSetTripB then fDataLength:=fDataLength + 8 + 4;             // Command 136 (1) - Trip B Parameter
     // Note: On E01003 some parameters must be cleared before other can be set
     // to prevent multiple function assignment on the pulse and alarm output
     // and ensure smoothly module reconfiguration
     if fClearLFDAndInputMode then fDataLength:=fDataLength + 8 + 4; // Command 131     - Clear Input configuration
     if fClearDivider then fDataLength:=fDataLength + 8 + 4;         // Command 138     - Clear Pulse Output / Divider Parameter
     if fClearTripA then fDataLength:=fDataLength + 8 + 4;           // Command 136 (0) - Clear Trip A Parameter
     if fClearTripB then fDataLength:=fDataLength + 8 + 4;           // Command 136 (1) - Clear Trip B Parameter
     Result:=fDataLength;
     end;
end;

procedure THartModule_1003.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     f1003ParametersAvailable:=(Read_One_Byte(lHartMaster^,132,fActualInputAndLFDMode,true)) and
                               (Read_LeakFlowOffParameters_1003(lHartMaster^,fActualLeakFlowOffParameters,true)) and
                               (Read_TripParameters_1003(lHartMaster^,fActualTripAParameters,0,true)) and
                               (Read_TripParameters_1003(lHartMaster^,fActualTripBParameters,1,true)) and
                               (Read_One_Byte(lHartMaster^,145,fActualHardwareCircuitry, true))
  else f1003ParametersAvailable:=false;
end;

procedure THartModule_1003.UploadRequiredModuleSpecificDeviceData;
var lLeakFlowOffParameters: THLeakFlowOffParameters_1003;
    lDividerParameters: THDividerParameters;
    lTripAWritten, lTripBWritten: boolean;
    lTripA,lTripB: THTripParameters_1003;
    lInputAndLFD: byte;
begin
  if (lHartMaster<>nil) then
     begin
     lDividerParameters:=fRequiredDividerParameters;
     lInputAndLFD:=fRequiredInputAndLFDMode;
     lTripA:=fRequiredTripAParameters;
     lTripB:=fRequiredTripBParameters;
     lLeakFlowOffParameters:=fRequiredLeakFlowOffParameters;

     lTripAWritten:=false;
     lTripBWritten:=false;

     if (fSetTripB) or (fClearTripB and not(fSetTripA))
        then lTripB.FunctionSelect:=lTripB.FunctionSelect or CP3_TripReconfig
     else if (fSetTripA) or (fClearTripA) then lTripA.FunctionSelect:=lTripA.FunctionSelect or CP3_TripReconfig;

     // Note: On E01003 some parameters must be cleared before other can be set
     // to prevent multiple function assignment on the pulse and alarm output
     // and ensure smoothly module reconfiguration
     if fClearDivider then
        if Write_DividerParameters(lHartMaster^, lDividerParameters, true) then
           begin
           WaitMilliseconds(250);
           if (Read_DividerParameters(lHartMaster^,lDividerParameters,true)) then
              fActualDividerParameters:=lDividerParameters;
           end;
     if fClearLFDAndInputMode then
        if (Write_One_Byte(lHartMaster^,131,lInputAndLFD, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_One_Byte(lHartMaster^,132,lInputAndLFD,true)) then
              fActualInputAndLFDMode:=lInputAndLFD;
           end;
     if fClearTripA then
        begin
        lTripA.FunctionSelect:=(lTripA.FunctionSelect and not(CP3_TripOutputInfo)) or (fActualTripAParameters.FunctionSelect and CP3_TripOutputInfo);
        lTripAWritten:=Write_TripParameters_1003(lHartMaster^, lTripA, true);
        WaitMilliseconds(250);
        end;
     if fClearTripB then
        begin
        lTripB.FunctionSelect:=(lTripB.FunctionSelect and not(CP3_TripOutputInfo)) or (fActualTripBParameters.FunctionSelect and CP3_TripOutputInfo);
        lTripBWritten:=Write_TripParameters_1003(lHartMaster^, lTripB, true);
        WaitMilliseconds(250);
        end;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetLFDAndInputMode then
        if (Write_One_Byte(lHartMaster^,131,lInputAndLFD, true)) then
           begin
           WaitMilliseconds(250);
           if (Read_One_Byte(lHartMaster^,132,lInputAndLFD,true)) then
              fActualInputAndLFDMode:=lInputAndLFD;
           end;

     if fSetLeakFlow then
        if Write_LeakFlowOffParameters_1003(lHartMaster^, lLeakFlowOffParameters, true) then
           begin
           WaitMilliseconds(250);
           if (Read_LeakFlowOffParameters_1003(lHartMaster^, lLeakFlowOffParameters, true)) then
              fActualLeakFlowOffParameters:=lLeakFlowOffParameters;
           end;

     if fSetTripA then
        begin
        lTripAWritten:=Write_TripParameters_1003(lHartMaster^, lTripA, true);
        WaitMilliseconds(250);
        end;

     if fSetTripB then
        begin
        lTripBWritten:=Write_TripParameters_1003(lHartMaster^, lTripB, true);
        WaitMilliseconds(250);
        end;

     if lTripAWritten then
        begin
        WaitMilliseconds(250);
        if (Read_TripParameters_1003(lHartMaster^, lTripA, 0, true)) then
           fActualTripAParameters:=lTripA;
        end;

     if lTripBWritten then
        begin
        WaitMilliseconds(250);
        if (Read_TripParameters_1003(lHartMaster^, lTripB, 1, true)) then
           fActualTripBParameters:=lTripB;
        end;
     end;
end;

function THartModule_1003.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          (CompareLFPs(fRequiredLeakFlowOffParameters,fActualLeakFlowOffParameters)) and
          (CompareInpLFDs(fRequiredInputAndLFDMode,fActualInputAndLFDMode)) and
          (CompareTPXs(fRequiredTripAParameters,fActualTripAParameters)) and
          (CompareTPXs(fRequiredTripBParameters,fActualTripBParameters));
end;

procedure THartModule_1003.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualInputAndLFDMode:=fRequiredInputAndLFDMode;
  fActualLeakFlowOffParameters:=fRequiredLeakFlowOffParameters;
  fActualTripAParameters:=fRequiredTripAParameters;
  fActualTripBParameters:=fRequiredTripBParameters;
end;

procedure THartModule_1003.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredInputAndLFDMode:=fActualInputAndLFDMode;
  fRequiredLeakFlowOffParameters:=fActualLeakFlowOffParameters;
  fRequiredTripAParameters:=fActualTripAParameters;
  fRequiredTripBParameters:=fActualTripBParameters;
end;

procedure THartModule_1003.ReadModuleSpecificDaviceDataFromStream;
var lD1003: TH1003;
begin
  fDeviceDataStream.Seek(System.Length(CNew_Config_Header),soFromBeginning);
  if ((fDeviceDataStream.Size-fDeviceDataStream.Position)=SizeOf(TH1003)) then
     begin
     FillChar(lD1003,SizeOf(TH1003),$FF);
     if (fDeviceDataStream.Read(lD1003,SizeOf(TH1003))=SizeOf(TH1003)) and
        (lD1003.Parity=GetDataStreamParityByte(lD1003,SizeOf(TH1003)-4)) then
        begin
        while System.Length(lD1003.MDD.TDD.TagName)<8 do lD1003.MDD.TDD.TagName:=lD1003.MDD.TDD.TagName+' ';
        while System.Length(lD1003.MDD.TDD.Descriptor)<16 do lD1003.MDD.TDD.Descriptor:=lD1003.MDD.TDD.Descriptor+' ';
        if (lD1003.MDD.UID.ManufacturerIdentCode=CHart_Manufacturer_MTL) and
           (IdentToType(lD1003.MDD.UID) in [dt1003]) then
           begin
           fActualBasicDeviceData.UID:=lD1003.MDD.UID;
           fActualBasicDeviceData.TDD:=lD1003.MDD.TDD;
           fActualPrimaryVariableOutputData:=lD1003.MDD.PVO;

           fActualFullScale:=lD1003.ADD.FullScaleFlag;
           fActualTransmitterFailureAlarmLevel:=lD1003.ADD.TransmitterFailureAlarmLevel;
           fActualAdjustOutputParameters:=lD1003.ADD.AOP;
           fActualDividerParameters:=lD1003.FTD.DVP;

           fActualInputAndLFDMode:=lD1003.LFD_Inp.FunctionSelect;
           fActualLeakFlowOffParameters:=lD1003.ADD.LFP;
           fActualTripAParameters:=lD1003.FTD.TPA;
           fActualTripBParameters:=lD1003.FTD.TPB;

           SetModuleSpecificDeviceDataAvailableStatus;
           end;
        end
     end;
end;

procedure THartModule_1003.WriteModuleSpecificDeviceDataToStream;
var lD1003: TH1003;
begin
  lD1003.MDD.UID:=fActualBasicDeviceData.UID;
  lD1003.MDD.TDD:=fActualBasicDeviceData.TDD;
  lD1003.MDD.PVO:=fActualPrimaryVariableOutputData;
  lD1003.ADD.AOP:=fActualAdjustOutputParameters;
  lD1003.ADD.LFP:=fActualLeakFlowOffParameters;
  lD1003.ADD.TransmitterFailureAlarmLevel:=fActualTransmitterFailureAlarmLevel;
  lD1003.ADD.FullScaleFlag:=fActualFullScale;
  lD1003.LFD_Inp.FunctionSelect:=fActualInputAndLFDMode;
  lD1003.FTD.DVP:=fActualDividerParameters;
  lD1003.FTD.TPA:=fActualTripAParameters;
  lD1003.FTD.TPB:=fActualTripBParameters;
  lD1003.Parity:=GetDataStreamParityByte(lD1003,SizeOf(TH1003)-4);

  fDeviceDataStream.Write(lD1003,SizeOf(TH1003));
end;

function THartModule_1003.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result:=inherited GetModuleSpecificDeviceDataAvailableStatus;
end;

procedure THartModule_1003.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  f1003ParametersAvailable:=false;
end;

procedure THartModule_1003.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  f1003ParametersAvailable:=false;
end;

// public

constructor THartModule_1003.Create;
begin
  inherited Create;
end;

destructor THartModule_1003.Destroy;
begin

  inherited Destroy;
end;

function THartModule_1003.SetDigitalDiagnostic(Selector: byte): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Write_One_Byte(lHartMaster^, 135, Selector, true)
  else Result:=false;
end;

function THartModule_1003.SetDigitalDiagnostic(Selector: byte; Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Write_Byte_And_Single(lHartMaster^, 135, Selector, Value, true)
  else Result:=false;
end;

function THartModule_1003.GetHardwareCircuitryConfiguration(var HardConfig: byte): boolean;
var lHardConfig: byte;
begin
  Result:=false;
  if (lHartMaster<>nil) then
     if (Read_One_Byte(lHartMaster^, 145, lHardConfig, true)) then
        begin
        HardConfig:=lHardConfig;
        fActualHardwareCircuitry:=lHardConfig;
        Result:=true;
        end;
end;

function THartModule_1003.SetHardwareCircuitryConfiguration(HardConfig: byte): boolean;
begin
  Result:=false;
  if (lHartMaster<>nil) then
     Result:=(Write_One_Byte(lHartMaster^, 144, HardConfig, true));
end;

function THartModule_1003.GetTrimmerInformation(Selector: byte; var Target: THTrimmerVariables_1003): boolean;
var lTrimmerVariables: THTrimmerVariables_1003;
begin
  Result:=false;
  if (lHartMaster<>nil) then
     if (Read_TrimmerVariables(lHartMaster^, lTrimmerVariables, Selector, true)) then
        begin
        Target:=lTrimmerVariables;
        Result:=true;
        end;
end;

function THartModule_1003.SetTrimmerInformation(Selector: byte; Source: THTrimmerVariables_1003): boolean;
begin
  Result:=false;
  if (lHartMaster<>nil) then
     Result:=(Write_TrimmerVariables(lHartMaster^, Source, Selector, true));
end;

// ****************************************************************************

// TMultiChannelHartModule

// private

procedure TMultiChannelHartModule.SetChannelNumber(Value: byte);
begin
  if (Value<fMaxChannels) then
     fActualChannel:=Value;
end;

procedure TMultiChannelHartModule.SetMaxChannelCount(Value: byte);
begin
  fMaxChannels:=Value;

  if not(fActualChannel<fMaxChannels) then
     fActualChannel:=0;
end;

// protected


// public

constructor TMultiChannelHartModule.Create;
begin
  inherited Create;

  fActualChannel:=0;
  fMaxChannels:=1;
end;

destructor TMultiChannelHartModule.Destroy;
begin

  inherited Destroy;
end;

// ****************************************************************************

// THartModule_x576

// private

procedure THartModule_x576.SetInputParameters(Value: THInputParameterArray_x576);
begin
  fRequiredInputParameters:=Value;
end;

procedure THartModule_x576.SetOutputParameters(Value: THOutputParameterArray_x576);
begin
  fRequiredOutputParameters:=Value;
end;

procedure THartModule_x576.SetCommonParameters(Value: THCommonParameters_x576);
begin
  fRequiredCommonParameters:=Value;
end;

procedure THartModule_x576.SetUserSensorHeader(Value: THUserSensorHeader_x576);
begin
  fRequiredUserSensorHeader:=Value;
end;

procedure THartModule_x576.ClearModuleSpecificDeviceDataStatusFlags;
begin
  inherited ClearModuleSpecificDeviceDataStatusFlags;

  fSetInputs0:=false;
  fSetInputs1:=false;
  fSetOutputs0:=false;
  fSetOutputs1:=false;
  fSetCommons:=false;
end;

procedure THartModule_x576.EvaluateModuleSpecificDeviceDataStatusFlags;
begin
  inherited EvaluateModuleSpecificDeviceDataStatusFlags;

  fSetInputs0:=not(Comparex576Inp(fRequiredInputParameters[0],fActualInputParameters[0]));
  fSetInputs1:=not(Comparex576Inp(fRequiredInputParameters[1],fActualInputParameters[1]));
  fSetOutputs0:=not(Comparex576Out(fRequiredOutputParameters[0],fActualOutputParameters[0]));
  fSetOutputs1:=not(Comparex576Out(fRequiredOutputParameters[1],fActualOutputParameters[1]));
  fSetCommons:=not(Comparex576Com(fRequiredCommonParameters,fActualCommonParameters));
end;

// protected

function THartModule_x576.GetModuleSpecificTransferDataLength(Value: TTransferDirection): integer;
var fDataLength: integer;
begin
  fDataLength:=inherited GetModuleSpecificTransferDataLength(Value);

  if (Value=tdDownload) then
     Result := fDataLength
               + 8 + 17  // Command 151 (0) - TInParameterSet_x576
               + 8 + 17  // Command 151 (1) - TInParameterSet_x576
               + 8 + 23  // Command 161 (0) - TOutParameterSet_x576
               + 8 + 23  // Command 161 (1) - TOutParameterSet_x576
               + 8 + 3   // Command 208     - THCommonParameters_x576
               + 8 + 14  // Command 209 (0) - THUserSensorData_x576 (only header data)
  else  // Value=tdUpload
     begin
     if fSetInputs0 then fDataLength:=fDataLength + 8 + 17 +8+17;  // Command 152 (0) - TInParameterSet_x576
     if fSetInputs1 then fDataLength:=fDataLength + 8 + 17 +8+17;  // Command 152 (1) - TInParameterSet_x576
     if fSetOutputs0 then fDataLength:=fDataLength + 8 + 23; // Command 162 (0) - TOutParameterSet_x576
     if fSetOutputs1 then fDataLength:=fDataLength + 8 + 23; // Command 162 (1) - TOutParameterSet_x576
     if fSetCommons then fDataLength:=fDataLength + 8 + 3;   // Command 207     - THCommonParameters_x576
     Result:=fDataLength;
     end;
end;

procedure THartModule_x576.DownloadActualModuleSpecificDeviceData;
begin
  inherited DownloadActualModuleSpecificDeviceData;

  if (lHartMaster<>nil) then
     fx576ParametersAvailable:=(Read_InputParameters_x576(lHartMaster^,0,fActualInputParameters[0],true)) and
                               (Read_InputParameters_x576(lHartMaster^,1,fActualInputParameters[1],true)) and
                               (Read_OutputParameters_x576(lHartMaster^,0,fActualOutputParameters[0],true)) and
                               (Read_OutputParameters_x576(lHartMaster^,1,fActualOutputParameters[1],true)) and
                               (Read_CommonParameters_x576(lHartMaster^,fActualCommonParameters,true)) and
                               (Read_UserSensorHeader_x576(lHartMaster^,fActualUserSensorHeader,true))
  else fx576ParametersAvailable:=false;
end;

procedure THartModule_x576.UploadRequiredModuleSpecificDeviceData;
var lInputParameters: THInputParameterArray_x576;
    lInputParametersWrite,
    lInputParametersRead: THInputParameters_x576;
    lOutputParameters: THOutputParameterArray_x576;
    lCommonParameters: THCommonParameters_x576;
begin
  if (lHartMaster<>nil) then
     begin
     lInputParameters:=fRequiredInputParameters;
     lOutputParameters:=fRequiredOutputParameters;
     lCommonParameters:=fRequiredCommonParameters;

     inherited UploadRequiredModuleSpecificDeviceData;

     if fSetInputs0 then
        begin
        lInputParametersWrite:=fRequiredInputParameters[0];
        if (WriteRead_InputParameters_x576 (lHartMaster^, 0, lInputParametersWrite, true)) and
           (Comparex576Inp(fRequiredInputParameters[0], lInputParametersWrite)) then
           begin
           fActualInputParameters[0]:=lInputParametersWrite;
           WaitMilliseconds(2000);
           end
        else
           begin
           WaitMilliseconds(2000);
           if (Read_InputParameters_x576 (lHartMaster^, 0, lInputParametersRead, true)) then
              fActualInputParameters[0]:=lInputParametersRead
           else WaitMilliseconds(100);
           end;
        end;
     if fSetInputs1 then
        begin
        lInputParametersWrite:=fRequiredInputParameters[1];
        if (WriteRead_InputParameters_x576 (lHartMaster^, 1, lInputParametersWrite, true)) and
           (Comparex576Inp(fRequiredInputParameters[1], lInputParametersWrite)) then
           begin
           fActualInputParameters[1]:=lInputParametersWrite;
           WaitMilliseconds(2000);
           end
        else
           begin
           WaitMilliseconds(2000);
           if (Read_InputParameters_x576 (lHartMaster^, 1, lInputParametersRead, true)) then
              fActualInputParameters[1]:=lInputParametersRead
           else WaitMilliseconds(100);
           end;
        end;
     if fSetOutputs0 then
        begin
        if (WriteRead_OutputParameters_x576 (lHartMaster^, 0, lOutputParameters[0], true)) then
           fActualOutputParameters[0]:=lOutputParameters[0];
        WaitMilliseconds(500);
        end;
     if fSetOutputs1 then
        begin
        if (WriteRead_OutputParameters_x576 (lHartMaster^, 1, lOutputParameters[1], true)) then
           fActualOutputParameters[1]:=lOutputParameters[1];
        WaitMilliseconds(500);
        end;
      if fSetCommons then
        if (WriteRead_CommonParameters_x576(lHartMaster^, lCommonParameters, true)) then
           fActualCommonParameters:=lCommonParameters;
     end;
end;

function THartModule_x576.VerifyRequiredModuleSpecificDeviceData: boolean;
begin
  Result:=inherited VerifyRequiredModuleSpecificDeviceData and
          Comparex576Inp(fRequiredInputParameters[0],fActualInputParameters[0]) and
          Comparex576Inp(fRequiredInputParameters[1],fActualInputParameters[1]) and
          Comparex576Out(fRequiredOutputParameters[0],fActualOutputParameters[0]) and
          Comparex576Out(fRequiredOutputParameters[1],fActualOutputParameters[1]) and
          Comparex576Com(fRequiredCommonParameters,fActualCommonParameters);
end;

procedure THartModule_x576.SynchronizeActualModuleSpecificDeviceData;
begin
  inherited SynchronizeActualModuleSpecificDeviceData;

  fActualInputParameters:=fRequiredInputParameters;
  fActualOutputParameters:=fRequiredOutputParameters;
  fActualCommonParameters:=fRequiredCommonParameters;
end;

procedure THartModule_x576.SynchronizeRequiredModuleSpecificDeviceData;
begin
  inherited SynchronizeRequiredModuleSpecificDeviceData;

  fRequiredInputParameters:=fActualInputParameters;
  fRequiredOutputParameters:=fActualOutputParameters;
  fRequiredCommonParameters:=fActualCommonParameters;
end;

procedure THartModule_x576.ReadModuleSpecificDaviceDataFromStream;
var lDx576: THx576;
    lSize: integer;
    lSLength: word;
    lParity: byte;
begin
  // Note: the Read procedure for MTLx576 differ from the previous modules
  // because of different file structure used in the software version 3.01
  lSize:=Sizeof(THMainDeviceData)+SizeOf(byte)+SizeOf(word)+
         Sizeof(THInputParameterArray_x576)+SizeOf(THOutputParameterArray_x576)+
         Sizeof(THCommonParameters_x576)+SizeOf(byte);
  fDeviceDataStream.Seek(System.Length(CNew_Config_Header),soFromBeginning);
  if ((fDeviceDataStream.Size-fDeviceDataStream.Position)=lSize) then
     begin
     FillChar(lDx576,SizeOf(THx576),$FF);
     if (fDeviceDataStream.Read(lDx576.MDD,SizeOf(THMainDeviceData))=SizeOf(THMainDeviceData)) and
        (fDeviceDataStream.Read(lParity,SizeOf(byte))=SizeOf(byte)) and
        (lParity=GetDataStreamParityByte(lDx576.MDD,SizeOf(THMainDeviceData))) and
        (fDeviceDataStream.Read(lSLength,SizeOf(word))=SizeOf(word)) and
        (lSLength=(SizeOf(THInputParameterArray_x576)+SizeOf(THOutputParameterArray_x576)+SizeOf(THCommonParameters_x576)+SizeOf(byte))) and
        (fDeviceDataStream.Read(lDx576.INPAR,SizeOf(THInputParameterArray_x576))=SizeOf(THInputParameterArray_x576)) and
        (fDeviceDataStream.Read(lDx576.OUTPAR,SizeOf(THOutputParameterArray_x576))=SizeOf(THOutputParameterArray_x576)) and
        (fDeviceDataStream.Read(lDx576.COMMPAR,SizeOf(THCommonParameters_x576))=SizeOf(THCommonParameters_x576)) and
        (fDeviceDataStream.Read(lDx576.Parity,SizeOf(byte))=SizeOf(byte)) and
        (lDx576.Parity=GetDataStreamParityByte(lDx576.INPAR,SizeOf(THInputParameterArray_x576)) xor
                       GetDataStreamParityByte(lDx576.OUTPAR,SizeOf(THOutputParameterArray_x576)) xor
                       GetDataStreamParityByte(lDx576.COMMPAR,SizeOf(THCommonParameters_x576))) then
        begin
        while System.Length(lDx576.MDD.TDD.TagName)<8 do lDx576.MDD.TDD.TagName:=lDx576.MDD.TDD.TagName+' ';
        while System.Length(lDx576.MDD.TDD.Descriptor)<16 do lDx576.MDD.TDD.Descriptor:=lDx576.MDD.TDD.Descriptor+' ';
        if (lDx576.MDD.UID.ManufacturerIdentCode=CHart_Manufacturer_MTL) and
           (IdentToType(lDx576.MDD.UID) in [dtx576]) then
           begin
           fActualBasicDeviceData.UID:=lDx576.MDD.UID;
           fActualBasicDeviceData.TDD:=lDx576.MDD.TDD;
           fActualInputParameters:=lDx576.INPAR;
           fActualOutputParameters:=lDx576.OUTPAR;
           fActualCommonParameters:=lDx576.COMMPAR;

           SetModuleSpecificDeviceDataAvailableStatus;
           end;
        end;
     end;
end;

procedure THartModule_x576.WriteModuleSpecificDeviceDataToStream;
var lDx576: THx576;
    lSLength: word;
    lParity: byte;
begin
   // Note: the Write procedure for MTLx576 differ from the previous modules
   // because of different file structure used in the software version 3.01

  lDx576.MDD.UID:=fActualBasicDeviceData.UID;
  lDx576.MDD.TDD:=fActualBasicDeviceData.TDD;
  lDx576.MDD.PVO:=CDef_PVO_x073; // not required for MTLx576 but for compatibility with 3.01
  fDeviceDataStream.Write(lDx576.MDD,SizeOf(THMainDeviceData));
  lParity:=GetDataStreamParityByte(lDx576.MDD,SizeOf(THMainDeviceData));
  fDeviceDataStream.Write(lParity,SizeOf(byte));

  lSLength:=SizeOf(THInputParameterArray_x576)+SizeOf(THOutputParameterArray_x576)+SizeOf(THCommonParameters_x576)+SizeOf(byte);
  fDeviceDataStream.Write(lSLength,SizeOf(word));

  lDx576.INPAR:=fActualInputParameters;
  fDeviceDataStream.Write(lDx576.INPAR,SizeOf(THInputParameterArray_x576));

  lDx576.OUTPAR:=fActualOutputParameters;
  fDeviceDataStream.Write(lDx576.OUTPAR,SizeOf(THOutputParameterArray_x576));

  lDx576.COMMPAR:=fActualCommonParameters;
  fDeviceDataStream.Write(lDx576.COMMPAR,SizeOf(THCommonParameters_x576));

  lDx576.Parity:=GetDataStreamParityByte(lDx576.INPAR,SizeOf(THInputParameterArray_x576)) xor
                 GetDataStreamParityByte(lDx576.OUTPAR,SizeOf(THOutputParameterArray_x576)) xor
                 GetDataStreamParityByte(lDx576.COMMPAR,SizeOf(THCommonParameters_x576));

  fDeviceDataStream.Write(lDx576.Parity,SizeOf(byte));
end;

function THartModule_x576.GetModuleSpecificDeviceDataAvailableStatus: boolean;
begin
  Result := inherited GetModuleSpecificDeviceDataAvailableStatus and
            fx576ParametersAvailable;
end;

procedure THartModule_x576.ClearModuleSpecificDeviceDataAvailableStatus;
begin
  inherited ClearModuleSpecificDeviceDataAvailableStatus;

  fx576ParametersAvailable:=false;
end;

procedure THartModule_x576.SetModuleSpecificDeviceDataAvailableStatus;
begin
  inherited SetModuleSpecificDeviceDataAvailableStatus;

  fx576ParametersAvailable:=true;
end;

// public

constructor THartModule_x576.Create;
begin
  inherited Create;

  fMaxChannels:=2;

  fx576ParametersAvailable:=false;
end;

destructor THartModule_x576.Destroy;
begin

  inherited Destroy;
end;

function THartModule_x576.GeTHDynamicVariables(var Target: THDynamicVariables): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Read_Channel_DynamicVariables(lHartMaster^,  fActualChannel, Target, 250, true)
  else Result:=false;
end;

function THartModule_x576.SetOutputCurrent(Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=WriteRead_Channel_Current_Value(lHartMaster^, fActualChannel, Value, true)
  else Result:=false;
end;

function THartModule_x576.CalibratePoint4mA(Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=WriteRead_Channel_TrimmDAC(lHartMaster^, fActualChannel, 4, Value, true)
  else Result:=false;
end;

function THartModule_x576.CalibratePoint20mA(Value: single): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=WriteRead_Channel_TrimmDAC(lHartMaster^, fActualChannel, 20, Value, true)
  else Result:=false;
end;

function THartModule_x576.GeTHTransmitterStatus(var Target: THTransmitterStatus): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Read_Channel_TransmitterStatus(lHartMaster^, fActualChannel, Target, true)
  else Result:=false;
end;

function THartModule_x576.DownloadUserSensorTableCoefficients(var Target: THUserSensorTable): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=Read_UserSensorTable_x576(lHartMaster^,Target,true)
  else Result:=false;
end;

function THartModule_x576.UploadUserSensorData(Source: THUserSensorData_x576): boolean;
begin
  if (lHartMaster<>nil) then
     Result:=WriteVerify_UserSensorData_x576(lHartMaster^,Source,true)
  else Result:=false;
end;

end.
