//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//**
//**	ORDER DESCRIPTION - CONFIGURATION SOFTWARE WITH HART COMMUNICATION (PCS45)
//**	HARTMODULES.PAS
//**
//**	Registry and other settings read and write procedures for PCS45
//**  (derived from the unit "Settings.pas")
//**
//**	Autor : Georg  Hahn
//**	Datum :  xx.yy.96 / 11.08.05 / 23.08.06 / 24.10.06 / 28.03.07 / 14.06.14
//**
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
unit SystemProcs;

interface

uses Windows, SysUtils, Types, DataTypes;

  function LoadNewResourceModule(DLL_Name: string): LongInt;

  procedure ReadAvailableLanguageInfos(var Target: TLanguageInfos);
  function GetLocaleFileSuffix(Source: TFileName): string;
  procedure ReadLanguageStrings(var Target: TLabelInfos);
  procedure ReadLanguageSystemStrings;

  procedure SavePasswordToRegistry (Value: string);
  procedure SaveLanguageToRegistry (Value: LANGID);
  procedure SavePortToRegistry (Value: integer);

  procedure SaveErrorIndicationToRegistry (Value: TTrippleBoolArray);
  procedure SaveDeviceTypeToRegistry (Value: TDeviceType);
  procedure SaveLastPathsToRegistry (LastSubKey,Value: string);

  function LoadPasswordFromRegistry: string;
  function LoadLanguageFromRegistry: LANGID;
  function LoadPortFromRegistry: integer;

  function LoadErrorIndicationFromRegistry: TTrippleBoolArray;
  function LoadDeviceTypeFromRegistry: TDeviceType;
  function LoadLastPathsFromRegistry(LastSubKey: string): string;

  procedure ArrangeModeAndConfiguration;


implementation

uses Dialogs, Consts, SHFolder, StringUtils, RegUtils, ProcessUtils, DataInstances, DataConstants,
     LabelNumbers, LanguageStringProcs, ConversionProcs, HelpContextNumbers;

const
  CMTL_Group_Software_Path='\Software\'+CMtlManufacturer;

  CAutoStartInfo = 'Pcs45.ConfigurationFiles';
  CAutoFileDescription ='Pcs45 data file';

  COptionsKey     = 'Options';
  CSettingsKey    = 'Settings';
  CPassword       = 'Password';
  CLanguage       = 'Language';
  CComPort        = 'Port';
  CErrors         = 'ErrorIndication';
  CTypeOfDevice   = 'Type Of Device';
  CLastD45Path    = 'Last Data Path';
  CLastImpexPath  = 'Last Impex Path';
  CLastLoggerPath = 'Last Logger Path';

type
  TFTDeviceBuffer = array [1..50] of char;

procedure UnloadResourceModule;
begin
  if ActiveLibrary<>0 then
     begin
     FreeLibrary(ActiveLibrary);
     @LoadDllStr:=nil;
     ActiveLibrary:=0;
     end;
end;

function LoadNewResourceModule(DLL_Name: string): LongInt;
var FileName: array [0..260] of char;
begin
  if (DLL_Name<>'') and FileExists(DLL_Name) then
     begin
     UnloadResourceModule;

     StrPCopy(FileName,DLL_Name);
     ActiveLibrary:=LoadLibrary(FileName);

     if ActiveLibrary<>0 then
        @LoadDllStr:=GetProcAddress(ActiveLibrary,'LoadDllStr')
     else @LoadDllStr:=nil;
     end;
  Result:=ActiveLibrary;
end;

function ReadLanguageInfosFromFile(Source: TSearchRec): TLanguageInfo;
var lDllVersion, lLangName: string;
    NewLangaugeInfo: TLanguageInfo;
    lIALabels: TImportExportLabels;
    lLangNumber: integer;
begin
  LoadNewResourceModule(Source.Name);
  lDllVersion:=LoadDllLanguageString(lhDllVersion);
  lLangNumber:=StrToIntDef(LoadDllLanguageString(lhDllLanguageNumber),0);
  lLangName:=LoadDllLanguageString(lhDllLanguageName);
  if (lDllVersion=CPcs45Version) then
     begin
     if (lLangNumber>0) and (lLangName<>'') then
        begin
        NewLangaugeInfo.LangNumber:=lLangNumber;
        NewLangaugeInfo.LangName:=lLangName;
        NewLangaugeInfo.FileName:=Source.Name;
        NewLangaugeInfo.ChangeWarning:=RestoreRessourceStringControls(LoadDllLanguageString(lhLanguageWarning));
        with lIALabels do
             begin
             LabelOn:=LoadDllLanguageString(flOn);
             LabelOff:=LoadDllLanguageString(flOff);
             AlarmLow:=LoadDllLanguageString(flLowAlarm);
             AlarmHigh:=LoadDllLanguageString(flHighAlarm);
             AlarmOff:=LoadDllLanguageString(flAlarmOff);
             Typ:=LoadDllLanguageString(flType);
             Resist:=LoadDllLanguageString(flResistor);
             Slow:=LoadDllLanguageString(flSlow);
             Medium:=LoadDllLanguageString(flMedium);
             Fast:=LoadDllLanguageString(flFast);
             EBD:=LoadDllLanguageString(flEBD);
             Trip:=LoadDllLanguageString(flTrip);
             User:=LoadDllLanguageString(flUser);
             in2out2:=LoadDllLanguageString(fl2in2out);
             in1out2:=LoadDllLanguageString(fl1in2out);
             SwitchProxy:=LoadDllLanguageString(flSwitchProximity);
             CurrentVoltage:=LoadDllLanguageString(flCurrentOrVoltage);
             Toggle:=LoadDllLanguageString(flToggleOutput);
             FixedPulses:=LoadDllLanguageString(flFixedPulse);
             LFD:=LoadDllLanguageString(flLFDShort);
             SimpleTrip:=LoadDllLanguageString(flSimpleTrip);
             WindowTrip:=LoadDllLanguageString(flWindowTrip);
             SoftAdjust:=LanguageStrings[flSoftAdjust];
             HardAdjust:=LanguageStrings[flHardAdjust];
             normal:=LanguageStrings[flNormalShort];
             inverted:=LanguageStrings[flInvertedShort];
             end;
        NewLangaugeInfo.IELabels:=lIALabels;
        end
     else // the language DLL file contents is anyway incorrect
        begin
        FmtStr(lDllVersion,'Corrupted language DLL file "%s"',[Source.Name]);
        MessageDlg(lDllVersion,mtError,[mbCancel],hcNoHelpTopic);
        NewLangaugeInfo.LangNumber:=0;
        end;
     end
  else // the language DLL file version is different than the application version
     begin
     FmtStr(lDllVersion,'Dissenting language DLL file "%s" version number.',[Source.Name]);
     MessageDlg(lDllVersion,mtError,[mbCancel],hcNoHelpTopic);
     NewLangaugeInfo.LangNumber:=0;
     end;
  Result:=NewLangaugeInfo;
end;

function AppendElementToArray(Element: TLanguageInfo; var TargetArray: TLanguageInfos): integer;
var L: integer;
begin
  L:=System.Length(TargetArray);
  SetLength(TargetArray,L+1);
  TargetArray[L]:=Element;
  Result:=L+1;
end;

procedure ReadAvailableLanguageInfos(var Target: TLanguageInfos);
var NewLangaugeInfo: TLanguageInfo;
    LangFilesPatternMask: string;
    SR: TSearchRec;
    TL: integer;
begin
  TL:=0; SetLength(Target,TL);
  LangFilesPatternMask:=OwnPath+OwnName+'???.dll';
  if FindFirst(LangFilesPatternMask,faAnyFile,SR)=0 then
     begin
     NewLangaugeInfo:=ReadLanguageInfosFromFile(SR);
     if (NewLangaugeInfo.LangNumber>0) then
        AppendElementToArray(NewLangaugeInfo,Target);
     while FindNext(SR)=0 do
           begin
           NewLangaugeInfo:=ReadLanguageInfosFromFile(SR);
           if (NewLangaugeInfo.LangNumber>0) then
              AppendElementToArray(NewLangaugeInfo,Target);
           end;
     end;
  FindClose(SR);
end;

function GetLocaleFileSuffix(Source: TFileName): string;
var P: integer;
    S: string;
begin
  S:=LowerCase(ExtractFileName(Source));
  if (System.Pos(LowerCase(OwnName),S)=1) then System.Delete(S,1,System.Length(OwnName));
  P:=System.Pos('.',S);
  if (P>0) then System.Delete(S,P,System.Length(S));
  if (System.Length(S)=3) then Result:=LowerCase(S)
  else Result:='';
end;

procedure ReadLanguageStrings(var Target: TLabelInfos);
var lLanguageString: string;
    I: integer;
begin
  for I:=flFirstLabelNumber to flLastLabelNumber do
      begin
      lLanguageString:=LoadDllLanguageString(I);

      if lLanguageString<>'' then
         Target[I]:=lLanguageString;
      end;
end;

procedure ReadLanguageSystemStrings;
begin
  HookResourceString(@SMsgDlgWarning, pchar(LanguageStrings[dlMsgWarning]));
  HookResourceString(@SMsgDlgError, pchar(LanguageStrings[dlMsgError]));
  HookResourceString(@SMsgDlgInformation, pchar(LanguageStrings[dlMsgInformation]));
  HookResourceString(@SMsgDlgConfirm, pchar(LanguageStrings[dlMsgConfirm]));
  HookResourceString(@SMsgDlgYes, pchar(LanguageStrings[dlMsgYes]));
  HookResourceString(@SMsgDlgNo, pchar(LanguageStrings[dlMsgNo]));
  HookResourceString(@SMsgDlgOK, pchar(LanguageStrings[dlMsgOK]));
  HookResourceString(@SMsgDlgCancel, pchar(LanguageStrings[dlMsgCancel]));
  HookResourceString(@SMsgDlgHelp, pchar(LanguageStrings[dlMsgHelp]));
  HookResourceString(@SMsgDlgAbort, pchar(LanguageStrings[dlMsgAbort]));
  HookResourceString(@SMsgDlgRetry, pchar(LanguageStrings[dlMsgRetry]));
  HookResourceString(@SMsgDlgIgnore, pchar(LanguageStrings[dlMsgIgnore]));
end;

function BuildActualRegistryMainKey: string;
begin
  Result:=CMTL_Group_Software_Path+'\'+OwnName;
end;

function BuildActualRegistrySubKey(SubKey: string): string;
begin
  Result:=BuildActualRegistryMainKey+'\'+CPcs45Version+'\'+SubKey;
end;

procedure SavePasswordToRegistry (Value: string);
begin
  SaveStringToRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CPassword,CodeString(Value));
end;

procedure SaveLanguageToRegistry (Value: LANGID);
begin
  SaveIntegerToRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CLanguage,Value and $1FF);
end;

procedure SavePortToRegistry (Value: integer);
begin
  SaveIntegerToRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CComPort,Value);
end;

procedure SaveErrorIndicationToRegistry (Value: TTrippleBoolArray);
begin
  SaveBinaryDataToRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CErrors,3,Value);
end;

procedure SaveDeviceTypeToRegistry (Value: TDeviceType);
begin
  SaveIntegerToRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(CSettingsKey),CTypeOfDevice,integer(Value));
end;

procedure SaveLastPathsToRegistry (LastSubKey,Value: string);
begin
  SaveStringToRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(CSettingsKey),LastSubKey,Value);
end;

function LoadPasswordFromRegistry: string;
begin
  Result:=LoadStringFromRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CPassword,'');
end;

function LoadLanguageFromRegistry: LANGID;
begin
  Result:=(LoadIntegerFromRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CLanguage,CDefaultLanguage) and $1FF);
end;

function LoadPortFromRegistry: integer;
begin
  Result:=LoadIntegerFromRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CComPort,0);
end;

function LoadErrorIndicationFromRegistry: TTrippleBoolArray;
var ErrorIndication: TTrippleBoolArray;
begin
  if not(LoadBinaryDataToRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(COptionsKey),CErrors,3,ErrorIndication))
  then ErrorIndication:=CDef_CFG.ErrorIndication;
  Result:=ErrorIndication;
end;

function LoadDeviceTypeFromRegistry: TDeviceType;
begin
  Result:=TDeviceType(LoadIntegerFromRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(CSettingsKey),CTypeOfDevice,0));
end;

function LoadLastPathsFromRegistry(LastSubKey: string): string;
begin
  Result:=LoadStringFromRegistry(HKEY_CURRENT_USER,BuildActualRegistrySubKey(CSettingsKey),LastSubKey,'');
end;

function GetSpecialFolderInfo(Selector: integer) : string;
const
   SHGFP_TYPE_CURRENT = 0;
   SHGFP_TYPE_DEFAULT = 1;
var FolderString: array [0..MaxChar] of char;
begin
  SHGetFolderPath(0,Selector,0,SHGFP_TYPE_CURRENT,@FolderString[0]) ;
  Result:= FolderString;
end;

function CreateSubdirectory: TFileName;
var lDocumentsFolder: string;
begin
  lDocumentsFolder:=GetMyDocumentsFolder+'\'+CMtlManufacturer;

  if not(SysUtils.DirectoryExists(lDocumentsFolder)) then
     SysUtils.CreateDir(lDocumentsFolder);

  lDocumentsFolder:=lDocumentsFolder+'\'+CApplicationName;

  if not(SysUtils.DirectoryExists(lDocumentsFolder)) then
     SysUtils.CreateDir(lDocumentsFolder);

  Result:=lDocumentsFolder;
end;

procedure CopyTemplateFiles(Target: TFileName);
const CTemplates = 'Templates';
var lTemplateSource,
    lTemplateTarget: TFileName;
begin
  lTemplateSource:=DeleteTrailingCharsUntil(DeleteTrailingChars(ExtractFilePath(GetApplicationFullFileName),'\'),'\')+CTemplates+'\*.*';
  lTemplateTarget:=Target+'\'+CTemplates;

  if not(SysUtils.DirectoryExists(lTemplateTarget)) then
     SysUtils.CreateDir(lTemplateTarget);

  CopyFilesToFolder(lTemplateSource, lTemplateTarget);
end;

procedure PerformFirstStartActions;
begin
  CleanRegistryKey(HKEY_CURRENT_USER,BuildActualRegistryMainKey);

  CFG45.Password:=CDef_CFG.Password;
  SavePasswordToRegistry(CFG45.Password);

  CFG45.Language:=CDefaultLanguage;
  SaveLanguageToRegistry(CFG45.Language);

  CFG45.PortNumber:=0;
  SavePortToRegistry(CFG45.PortNumber);

  CFG45.HighSpeed:=false;

  CFG45.ErrorIndication:=CDef_CFG.ErrorIndication;
  SaveErrorIndicationToRegistry(CFG45.ErrorIndication);

  MOD45.DevType:=CDefaultNewDevice;
  SaveDeviceTypeToRegistry(MOD45.DevType);

  MOD45.LastD45Path:=CreateSubdirectory;
  SaveLastPathsToRegistry(CLastD45Path,MOD45.LastD45Path);
  CopyTemplateFiles(MOD45.LastD45Path);

  CreateRegistryEntryForAutoStart(LowerCase(CDef_Pcs45_Ext),CAutoStartInfo,CAutoFileDescription,GetApplicationFullFileName);
end;

procedure ReadModeAndConfigurationFromRegistry;
var RegString: string;
    RegData: integer;
begin
  RegString:=LoadPasswordFromRegistry;
  if (RegString<>'') then CFG45.Password:=CodeString(RegString)
  else
     begin
     CFG45.Password:=CDef_CFG.Password;
     SavePasswordToRegistry(CFG45.Password);
     end;

  CFG45.Language:=LoadLanguageFromRegistry;
  if CFG45.Language=0 then
     begin
     CFG45.Language:=CDefaultLanguage;
     SaveLanguageToRegistry(CFG45.Language);
     end;

  RegData:=LoadPortFromRegistry;
  if not(RegData<-1) and not(RegData>8) then
     begin
     CFG45.PortNumber:=RegData;
     CFG45.PCL45USBSelected:=(RegData=-1)
     end
  else
     begin
     CFG45.PortNumber:=0;
     SavePortToRegistry(CFG45.PortNumber);
     end;

  CFG45.ErrorIndication:=LoadErrorIndicationFromRegistry;

  RegString:=LoadLastPathsFromRegistry(CLastD45Path);
  if (RegString<>'') then MOD45.LastD45Path:=RegString
  else
     begin
     MOD45.LastD45Path:=CreateSubdirectory;
     SaveLastPathsToRegistry(CLastD45Path, MOD45.LastD45Path);
     // CopyTemplateFiles(MOD45.LastD45Path);
     end;

  RegString:=LoadLastPathsFromRegistry(CLastImpexPath);
  if (RegString<>'') then MOD45.LastImpexPath:=RegString
  else MOD45.LastImpexPath:=GetMyDocumentsFolder+'\'+CMtlManufacturer+'\'+CApplicationName;

  RegString:=LoadLastPathsFromRegistry(CLastLoggerPath);
  if (RegString<>'') then MOD45.LastLoggerPath:=RegString
  else MOD45.LastLoggerPath:=GetMyDocumentsFolder+'\'+CMtlManufacturer+'\'+CApplicationName;
end;

procedure ArrangeModeAndConfiguration;
begin
  if not(RegistryKeyExists(HKEY_CURRENT_USER,BuildActualRegistryMainKey+'\'+CPcs45Version)) then
     PerformFirstStartActions
  else ReadModeAndConfigurationFromRegistry;
end;

end.
